#ifndef _IMALG_RESAMPLING_H 
#define _IMALG_RESAMPLING_H 
////////////////////////////////////////////////////////////////////////////////////////////////
#define _USE_MATH_DEFINES
#include <math.h>
#include "aimage.h"
#include "img_trans_models.h"
////////////////////////////////////////////////////////////////////////////////////////////////
namespace Vtl 
{
//! \brief     
namespace ImAlg 
{

/*!
 * \brief C   ,    
 *  ,   ( ). 
 * 
 *    :
 * 1)     
 * 2)     .
 *
 *****************************************************************************************************/
namespace Resamp 
{
////////////////////////////////////////////////////////////////////////////////////////////////
	/*! \brief        (x,y).  x,y  
	 *       .
     *
	 *	        ( 2x2)  ,  
	 *            . 
	 *             ( 
	 *   ),           
	 *   : 
	 *  \image html 1_1.eps 
	 *  
	 *  \f$  I_1=J_{11}(1-\delta x) + J_{12} \delta x  \f$
	 *  \f$  I_2=J_{21}(1-\delta x) + J_{22} \delta x  \f$
	 *  \f$  I_0=I_1(1-\delta y) + I_2 \delta y   \f$
	 *    \f$ I_{ij} \f$ -   .
	 *  
	 *   ,        \f$ \delta x \f$   \f$ \delta y \f$ . 
	 *        \f$ I_0 \f$ :
	 *  
	 *  \f$  I_0=I_{11} (1- \delta x - \delta y ) + (I_{11}+I_{22}-I_{12}-I{21}) \delta x \delta y  + (I_{12} \delta x + I_{21} \delta y ) \f$
	 *  
	 *  ,  ,         ,     , 
	 *              \f$ I_{ij} I_{ij} \f$ .   
	 *       ,       . 
	 *         -   x    y  ,   .
	 *
	 *           ,     
	 *   ,    .         
	 *  (7 /  3 ).
	 *
	 *  \f$  I_0 = I_{11} - (I_{11}-I_{12})\delta x - (I_{11}-I_{21})\delta y + (I_{11}+I_{22}-I_{12}-I{21}) \delta x \delta y \f$
	 *  \f$  I_0 = I_{11} + (I_{21}-I_{11})\delta y + ((I_{22}-I{21}-b) \delta y + b )\delta x  \f$ ,
	 *  \f$  b= I_{12}-I_{11} \f$  
	 *  
     *********************************************************************************************************/ 
	template<typename Pixel,unsigned stride=1>
	class InterpBilinear
	{
		const Pixel* data;
		const size_t nls;
		const size_t w;
		const size_t h;
		const Pixel blank;
	public:
		typedef Pixel result_type;
	//!   .    Bilinear
		InterpBilinear(const Pixel* data_ptr,size_t width,size_t height,size_t next_line_shift,Pixel fill)
			: data(data_ptr),nls(next_line_shift),w(width),h(height),blank(fill)
		{
		}
		size_t border() const { return 1; }
		Pixel at(double x, double y) const
		{
			if ( (x < 0) || (y < 0) || ( x > (w-1) ) || ( y > (h-1) ) ) return blank;
			return this->operator()(x,y);
		}

		Pixel operator()(double x, double y) const
		{
			int ix=int(x);
			int iy=int(y);
			const Pixel* p=data+iy*nls+ix*stride;
			double dx=x-ix;
			double dy=y-iy;
			double I1=*p;
			I1+=dx*( *(p+stride) - I1);
			double I2=*(p+nls);
			I2+=dx*( *(p+nls+stride) - I2);
			return Pixel( I1 + (I2 - I1)*dy );
		}
		size_t W() const {return w;}
		size_t H() const {return h;}
		size_t Nls() const {return nls;}
	};
////////////////////////////////////////////////////////////////////////////////////////////////
	/*!
	 * 
	 * \brief    value_min , value_max     N=N_grid_cells
	 * .   0   value_min,   (N-1) -  value_max.
	 *   i    value_min+step*i ,
	 *    : step=(value_max-value_min)/(N-1) .
	 *  
	 ************************************************************************************************/
	template<typename T>
	class LinGrid
	{
		T value_min;
		T value_max;
		T step;
		size_t n;
	public:
		LinGrid(T value_min_,T value_max_,size_t N_grid_cells)
			: value_min(value_min_),value_max(value_max_),n(N_grid_cells)
		{
			step=(value_max-value_min)/(n-1);
		}
		T value(size_t i) const { return value_min+step*i; }
		size_t N() const {return n;}
		template<typename I>
		T values_shift(I i_distance=1) const { return T(i_distance*step); }
	};
	/*!
	 *
	 * \brief    value_min , value_max     N=N_grid_cells
	 * .   0   value_min,   (N-1) -  value_max.
	 *   i    value_min*exp(step*i) ,
	 *     : step=(log(value_max)-log(value_min))/(N-1) .
	 *
	 *  log_rho_step*width=log(rho_max)-log(rho_min);
	 *  log_rho_step=(log(rho_max)-log(rho_min))/width;
	 *  log_rho_step*i+log(rho_min)=log(rho);
	 *  rho=exp(log_rho_step*i+log(rho_min))=rho_min*exp(log_rho_step*i);
	 *  
	 *  rho1=rho_min*exp(log_rho_step*i);
	 *  rho2=rho_min*exp(log_rho_step*(i+1));
	 *  alpha=rho2/rho1=exp(log_rho_step)
	 *  log_rho_step=log(N/(N-1))
	 *  
	 ************************************************************************************************/
	template<typename T>
	class LogGrid
	{
		T value_min;
		T value_max;
		T log_value_step;
		size_t n;
	public:
		///Only for containers and class member initialization, senseless.
		LogGrid() 
		{
			value_min=value_max=log_value_step=T(0);
			n=0;
		}
		LogGrid(T value_min_,T value_max_,size_t N_grid_cells)
			: value_min(value_min_),value_max(value_max_),n(N_grid_cells)
		{
			log_value_step=(std::log(value_max)-std::log(value_min))/(n-1);
		}
		T value(size_t i) const { return value_min*std::exp(log_value_step*i); }
		size_t N() const {return n;}
		template<typename I>
		T values_ratio(I i_distance=1) const { return T(std::exp(i_distance*log_value_step)); }
	};
	template<typename T>
	class PhiGrid
	{
		T from;
		T step;
		size_t n;
	public:
		PhiGrid(T from_,T to_,size_t N_grid_cells)
			: from(from_),step(T((to_-from_)/(N_grid_cells-1))),n(N_grid_cells)
		{
		}
		T value(size_t i) const {return from+i*step;}
		size_t N() const {return n;}
	};
	/*!
	 *   ()        
	 * ( -).  ,    \f$ \rho \f$    \f$ \phi \f$ 
	 *   RadGrid  PhiGrid     .    
	 *  (   RadGrid  LogGrid)-     .
	 *            -     RadGrid  PhiGrid, 
	 *          0-2*Pi,      
	 *   .
	 * x0, y0 -              
	 *  polar,width,height,nls -   ()  .
	 *
	 *
	 *************************************************************************************************/
	template<typename T,class Interpolator,class RadGrid,class PhiGrid> //RadGrid==LogGrid<float>
	void CartToPolar(Interpolator& interp,float x0,float y0,T * polar,int width,int height,int nls, RadGrid& lsg,PhiGrid phiGrid)
	{
//		const double rho_max=interp.W()/2;

		// x, y -   
		// phi, rho  -   
		// i, j -     polar
		for (int j = 0; j < height; ++j )
		{
			double phi=phiGrid.value(j);
			double c = cos(phi);
			double s = sin(phi);
			T *r = polar;
			for (int i = 0 ; i < width; ++i)
			{
				double rho=lsg.value(i);
				double x = rho*c;
				double y = rho*s;
				*r++ = interp.at(x0+x,y0+y);
			}
			polar += nls;
		}
	}
////////////////////////////////////////////////////////////////////////////////////////////////
	template<typename T,class Interpolator>
	void RotateScale(Interpolator& interp,float x0,float y0,float phi,float scale,T * result,int width,int height,int nls)
	{
		float matrix[]={scale*cos(phi),scale*sin(phi),scale*sin(-phi),scale*cos(phi)};
		Affine(interp,x0,y0,x0,y0,matrix,result,width,height,nls);
	}
	template<typename T,class Interpolator>
	void Affine(Interpolator& interp,float x0_in,float y0_in,float x0_out,float y0_out,float matrix_2x2[],T * result,int width,int height,int nls)
	{

		// x, y -      
		// phi, rho  -   
		// i, j -     ,   
		float a=matrix_2x2[0];
		float b=matrix_2x2[1];
		float c=matrix_2x2[2];
		float d=matrix_2x2[3];
		for (float j = -y0_out, jmax=height-y0_out ; j < jmax; ++j)
		{
			T *r = result;
			for (float i = -x0_out, imax=width-x0_out ; i < imax; ++i)
			{
				double x = a*i+b*j;
				double y = c*i+d*j;
				//std::cout << "interp" << x0_in+x << " " << y0_in+y <<std::endl;
				*r++ = interp.at(x0_in+x,y0_in+y);
			}
			result += nls;
		}
	}
	/*!
	 * \brief   ()     ,
	 *   Transformer (.   namespace Vtl::ImAlg::Transf). 
	 *   Transformer     Vtl::ImAlg::Model. 
	 *     Interpolator,
	 * .  class InterpBilinear.    
	 * result,width,height,nls.
	 *
	 ****************************************************************************************/
	template<typename T,class Interpolator,class Transformer>
	void Transform(const Interpolator& interp,const Transformer& transf,T * result,int width,int height,int nls)
	{
/*
		size_t bord=interp.border();
		Vector2 c[8]; 
		c[0]=transf(0,0);
		c[1]=transf(w-1,0);
		c[2]=transf(w-1,h-1);
		c[3]=transf(0,h-1);
		c[4]=transf(bord,bord);
		c[5]=transf(w-1-bord,bord);
		c[6]=transf(w-1-bord,h-1-bord);
		c[7]=transf(bord,h-1-bord);
				std::cout << c[0] << std::endl;
				std::cout << c[1] << std::endl;
				std::cout << c[2] << std::endl;
				std::cout << c[3] << std::endl;
*/
		for (float j = 0; j < height; ++j)
		{
			T *r = result;
			for (float i = 0; i < width; ++i)
			{
				//typename Model::Transf::value_type x;
				//typename Model::Transf::value_type y;
				typename Transformer::value_type x;
				typename Transformer::value_type y;
				transf(i,j,x,y);
				*r++ = interp.at(x,y);
			}
			result += nls;
		}
	}
////////////////////////////////////////////////////////////////////////////////////////////////
	//!     'P'
	template<typename T,typename TT,class Interpolator>
	void ProjectiveTransform(Interpolator& interp,TT* projective_mat_3x3,double x0_in,double y0_in,double x0_out,double y0_out,T* trans,int width,int height,int nls)
	{
		nls-=width;
		for (int i=0; i<height; i++)
		{
			for (int j=0; j<width; j++)
			{
				double x=j-x0_out;
				double y=i-y0_out;
				double u=projective_mat_3x3[0]*x+projective_mat_3x3[1]*y+projective_mat_3x3[2];
				double v=projective_mat_3x3[3]*x+projective_mat_3x3[4]*y+projective_mat_3x3[5];
				double w=projective_mat_3x3[6]*x+projective_mat_3x3[7]*y+projective_mat_3x3[8];
				x=u/w;
				y=v/w;
				*trans++ = interp.at(x+x0_in,y+y0_in);
			}
			trans+=nls;
		}
	}
////////////////////////////////////////////////////////////////////////////////////////////////
#if 0
	template<typename T>
	class PolarGrid
	{
		double  phi;
		double dphi;// = M_PI/hp;//   
		double c;
		double s;

		double rhomin;
		double drho;// = 1;//   
	public:
		PolarGrid(size_t phi_steps,T phi_from,T phi_to
			     ,size_t r_steps,T r_from,T r_to)
			:phi(phi_from),dphi((phi_to-phi_from)/phi_steps)
			,rhomin(r_from),drho((r_to-r_from)/r_steps)
		{
			c = cos(phi);
			s = sin(phi);
		}
		void nextPhi()
		{
			phi+=dphi;
			c = cos(phi);
			s = sin(phi);
		}
		void startRho()
		{
			rho = rhomin
		}
		void nextRho()
		{
			rho += drho
			double x = rho*c;
			double y = rho*s;
		}
//			for (int j = 0,double rho = rhomin; j < wp; j++, rho += drho)
	};
	template<class InsideInterp,class BoundInterp,class OutsideInterp>
	class PolarIterator
	{

		float psi_0;
		float psi_step;
		float psi_end;
		float rho_0;
		float rho_step;
		float rho_end;
	public:
		typedef typename InsideInterp::result_type result_type;
		class HIter
		{
			const PolarIterator& pi;
			double x;
			double y;
			int count;
			const double dx;
			const double dy;
		private:
			HIter(const PolarIterator& pi_,double x_,double y_,double dx_,double dy_,int count_)
				:pi(pi_),x(x_),y(y_),dx(dx_),dy(dy_),count(count_)
			{
			}
		public:
//			for (int j = 0,double rho = rhomin; j < wp; j++, rho += drho)
			HIter operator++()
			{
				//double x = rho*c;
				//double y = rho*s;
				x+=dx;
				y+=dy;
				++count;
			}
			bool operator!=(const HIter& h)	{ return count!=h.count;}
			result_type operator*()	{return pi.interp.at(x,y);}
		};
		class VIter
		{
			const PolarIterator& pi;

			float psi;
			float psi_step;

			int count;

			float c;
			float s;
			HIter begin_;
			HIter end_;
		private:
			VIter(const PolarIterator& pi_,float x0,float y0,float psi_from,float psi_to,int psi_N_step)
				: pi(pi_), psi(psi_from),psi_step((psi_to-psi_from)/psi_N_step)
			{
			}
		public:
			VIter& operator++() 
			{
				psi+=psi_step;
				c=std::cos(psi);
				s=std::sin(psi);

				double x = x0 + pi.rho_0 * c;
				double y = y0 + pi.rho_0 * s;
				double dx = pi.rho_step * c;
				double dy = pi.rho_step * s;
				begin_=HIter(pi,x,y,dx,dy,0);
				double x = x0 + pi.rho_0*c;
				double y = y0 + pi.rho_0*s;
				begin_=HIter(pi,x,y,dx,dy,0);

				return *this;
			}
			//HIter operator*() {}
			HIter begin() const { return begin_;}
			HIter end()   const { return end_;  } 

		};
	public:
		PolarIterator(int width,int height,float x0,float y0,float psi_from,float psi_to,int psi_N,bool include_psi_to,float rho_from,float rho_to,int rho_N,bool include_rho_to)
		{
		}
		VIter begin()
		{
		}
		VIter end()
		{
		}
	};
	//! x0, y0 -              
	//!  phi_from,phi_to
	template<typename T,class Interpolator>
	void CartToPolar(Interpolator& interp,float x0,float y0,T * polar,int width,int height,int nls)
	{
		double dphi = M_PI/height;//   
		double drho = 0.708;//   

		const double rhomin = 0;//  rho

		// x, y -   
		// phi, rho  -   
		// i, j -     polar
		double  phi = 0;
//		double  phi = M_PI;
		for (int j = 0; j < height; ++j, phi += dphi)
		{
			double c = cos(phi);
			double s = sin(phi);
			double x = x0+rhomin*c;
			double y = y0+rhomin*s;
			double dx = drho*c;
			double dy = drho*s;
			T *r = polar;
			for (int i = 0 ; i < width; ++i)
			{
				*r++ = interp.at(x,y);
				x+=dx;
				y+=dy;
			}
			polar += nls;
		}
	}
	//!     
	/*!
	 * \brief     
	 * 
	 *   cart   :   .
	 *  :    ,
	  ,   . 
	 :    ,   
	  .
	   polar      cart.
	       polar.
	       polar.
	*/
	inline void CartToPolar(AImageMemFloat* cart, AImageMemFloat* polar)
	{
		int wc = cart->Width();
		int hc = cart->Height();
		InterpBilinear<AImageMemFloat::Pixel> interp((*cart)[0],wc,hc,cart->NextLineShift());
		float x0=float(wc/2);
		float y0=float(hc/2);
		CartToPolar(interp,x0,y0,(*polar)[0],polar->Width(),polar->Height(),polar->NextLineShift());
	}
#endif 
////////////////////////////////////////////////////////////////////////////////////////////////
}//namespace Resamp
}//namespace ImAlg 
}//namespace Vtl
////////////////////////////////////////////////////////////////////////////////////////////////
#endif //_IMALG_RESAMPLING_H 
