#ifndef EDT_MEIJSTER_2000_H_
#define EDT_MEIJSTER_2000_H_
//////////////////////////////////////////////////////////////////////////////////////
#include "ytypes.h"

namespace Vtl { namespace ImAlg { namespace Edt {


	template<typename T>
	struct isEdge
	{
		T thresh;
		isEdge(T threshold=T(0))
			: thresh(threshold)
		{
		}
		bool operator()(T value)
		{
			return value > thresh;
		}
	};
	/// Calculates Euclidean Distance Transform using algorithm
	/// Meijster, A., Roerdink, J., Hesselink,W. 2000. A general algorithm for computing distance transforms
	/// in linear time. In Proceedings of the 5th International Conference on Mathematical Morphology and its
	/// Applications to Image and Signal Processing.
    class Meijster
    {
		enum {MaxValue=INT32_MAX};
	public:
		template<typename T1,typename T2,class Functor>
        static void edt(T1* src,int w,int h,int nls_s, T2* res,int nls_d,Functor isEdgeFunc)
        {
            //const int n0 = w;
			//const int n1 = h;

            //First Step

            std::vector<int> t1(w);
			std::vector<int> t2(h);

            for (int i = 0; i < w; ++i)
            {
				T1* col=src+i;
                int j = 0;
                for( ; j < h && !isEdgeFunc(*col) ; ++j, col+=nls_s )
                    t1[j] = MaxValue;
                for( ; j < h ; ++j, col+=nls_s )
                {
                    if( isEdgeFunc(*col) )
                        t1[j] = 0;
                    else
                        t1[j] = t1[j - 1] + 1;
                }

                //j = h - 1;
				//col=src+j*nls_s;
				--j;
				col-=nls_s;
                for( ; j >= 0 && !isEdgeFunc(*col) ; --j, col-=nls_s )
                    t2[j] = MaxValue;
                for( ; j >= 0 ; --j, col-=nls_s )
                {
                    if( isEdgeFunc(*col) )
                        t2[j] = 0;
                    else
                        t2[j] = t2[j + 1] + 1;
                }

				T2* r=res+i;
                for(j = 0 ; j < h; ++j , r+=nls_d )
					*r = std::min(t1[j], t2[j]);
            }


            //Second Step

			T2* r=res;
            std::vector<int> dist(w);

            for(int i = 0; i < h ; ++i, r+=nls_d)
            {
                for(int j = 0; j < w; ++j)
                    dist[j] = r[j];
				edt1D(&(dist[0]),w);
                for(int j = 0; j < w; ++j)
                    r[j] = dist[j];
            }
        }

	private:
		static void edt1D(int* dist,int N)
        {
            std::vector<int> prev(N);
            std::vector<int> next(N);

            for(int i=0 ; i < N ; ++i )
            {
                prev[i] = i - 1;
                next[i] = i + 1;
            }

            // Non-existent Points Removal

            int first = 0;
			int last = N-1;

            while(dist[first] == MaxValue && first < last)
                ++first;

            while (dist[last] == MaxValue && first < last)
                --last;

			int cur;
            if( first < last )
            {
                cur = next[first];

                while( cur < last )
                {
                    if( dist[cur] == MaxValue )
                    {
                        next[prev[cur]] = next[cur];
                        prev[next[cur]] = prev[cur];
                    }
                    cur = next[cur];
                }
            }

            //Bad Point Test
            if( first < last )
            {
                cur = next[first];

                while( cur < last )
                {
                    int pcur = prev[cur];
					int ncur = next[cur];
                    float A = float(ncur - pcur);
                    float B = float(dist[ncur] - dist[pcur]);
                    float X = 0.5f * (ncur + pcur + B / A * (dist[ncur] + dist[pcur]));
                    float D1 = (X - pcur) * (X - pcur) + dist[pcur] * dist[pcur];
                    float D2 = (X - cur) * (X - cur) + dist[cur] * dist[cur];
                    if( D1 <= D2 )
                    {
                        next[pcur] = ncur;
                        prev[ncur] = pcur;
                        if (pcur == first)
                            cur = ncur;
                        else
                            cur = pcur;
                    }
                    else
                        cur = ncur;
                }
            }

			//End of Bad Point Test
            std::vector<int> res(N);
            cur = first;

            for( int i=0 ; i < N ; ++i )
            {
                int D = (cur - i) * (cur - i) + dist[cur] * dist[cur];

                while( cur < last )
                {
                    int ncur = next[cur];
                    int D2 = (ncur - i) * (ncur - i) + dist[ncur] * dist[ncur];
                    if( D < D2 )
                        break;
                    else
                    {
                        cur = next[cur];
                        D = D2;
                    }
                }

                res[i] = D;
            }
			::memcpy(dist,&(res[0]),N*sizeof(dist[0]));
        }
    };

}}}///namespace Vtl { namespace ImAlg { namespace Edt {
//////////////////////////////////////////////////////////////////////////////////////
#endif //EDT_MEIJSTER_2000_H_
