#ifndef MODEL_H_
#define MODEL_H_

//library includings
#include "num_linalg.h"

//own includings
#include "Point2D.h"
#include "Points2DSet.h"
#include "Points2DParamSet.h"

namespace Basics3D
{
	template<class Operations>
	class T_Model
	{
	private:
		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> m_T;			// Normalization
		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> m_T1_1;   //           matrices
		T_Points2DSet<C_Point2D> m_PointsSet;
    /////////////////////////////////////////////////////////////////////////////////////
	private:
		void Denormalize(std::vector<Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3>>
			                                                                     &out_Matrices,
										 const double in_err)
		{
      for (int i = 0; i < (int)out_Matrices.size(); i++)
			{
				T_MatrixOperations::Normalize(out_Matrices[i], out_Matrices[i], m_T1_1, m_T);				
			}
		}

  public:
    /////////////////////////////////////////////////////////////////////////////////////
		T_Model(){}

    /////////////////////////////////////////////////////////////////////////////////////
		T_Model(T_Points2DSet<C_Point2D> in_PointsSet)
		{
		  m_PointsSet = in_PointsSet;
		}

    /////////////////////////////////////////////////////////////////////////////////////
		void Transform(Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 1> &out_Point,
			             const Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &in_Matrix,
			             const int i)
		{
		  m_PointsSet[i].Matrix(&out_Point);
			out_Point = boost::numeric::ublas::prod(in_Matrix, out_Point);
			out_Point(0, 0) = out_Point(0, 0) / out_Point(2, 0);
			out_Point(1, 0) = out_Point(1, 0) / out_Point(2, 0);
			out_Point(2, 0) = 1;
		}

    /////////////////////////////////////////////////////////////////////////////////////
		void Transform(Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 1> &out_Point,
			             const Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &in_Matrix,
			             const C_Point2D &in_point)
		{
		  in_point.Matrix(&out_Point);
			out_Point = boost::numeric::ublas::prod(in_Matrix, out_Point);
			out_Point(0, 0) = out_Point(0, 0) / out_Point(2, 0);
			out_Point(1, 0) = out_Point(1, 0) / out_Point(2, 0);
			out_Point(2, 0) = 1;
		}

    /////////////////////////////////////////////////////////////////////////////////////
		void GetSample(T_Points2DSet<C_Point2D> &out_sample)
		{
			m_PointsSet.RetPoints(out_sample, Operations::SampleSize());
		}

    /////////////////////////////////////////////////////////////////////////////////////
		void ComputeMatrices
			       (std::vector<Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3>> &out_Matrices,
							T_Points2DSet<C_Point2D> &in_sample,
							const double in_sigma)
		{
			Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> T1;
			double err = in_sample.Normalize(m_T, T1, in_sigma);
			Operations::NormMatrix(m_T1_1, T1, err);
		  Vtl::NumAlg::Lin::ublas::matrix<double> system;
			Operations::CreateSystem(system, in_sample);
			Operations::Find(out_Matrices, system);
			Denormalize(out_Matrices, err);
		}

    /////////////////////////////////////////////////////////////////////////////////////
		int IsInlier(const Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &in_Matrix,
			           const int i,
								 const double in_sigma)
		{
			if (m_PointsSet[2 * i].Numb() < 0)
			{
			  return 0;
			}

			Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 1> Point(3, 1);
		  Transform(Point, in_Matrix, 2 * i);

			double d = Operations::Dist(Point, m_PointsSet[2 * i + 1]);

			if (d < Operations::DistThresh(in_sigma))
			{
				return 1;
			}
			else
			{
			  return 0;
			}
		}

		/////////////////////////////////////////////////////////////////////////////////////
		void CountInliers(T_Points2DSet<C_Point2D> &out_inl, 
			               const C_Points2DParamSet &in_points1,
										 const C_Points2DParamSet &in_points2,
			               const Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &in_matrix,
			               const double in_sigma,
										 const double in_t)
		{
			int size1 = in_points1.Size();
			int size2 = in_points2.Size();

			// for each point from the 1st set
			for (int i = 0; i < size1; i++)
			{
				const C_Point2DParam *p_Point1 = in_points1[i];

				// points, which are not to be operated on, have number -1
				if (p_Point1->Numb() < 0)
				{
				  continue;
				}

				Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 1> PointTrM(3, 1);
				Transform(PointTrM, in_matrix, 
					        C_Point2D(p_Point1->X(), p_Point1->Y(), p_Point1->Numb()));
				
				C_Point2D prevPoint(-1, -1, -1);
				C_Point2D bestPoint(-1, -1, -1);
				double prevDist = 1;
				double bestDist = 1000;

				// for each point of the 2nd set
				for (int j = 0; j < size2; j++)
				{
					const C_Point2DParam *p_Point2 = in_points2[j];

					// points, which are not to be operated on, have number -1
					if (p_Point2->Numb() < 0)
					{
						continue;
					}

					// if PointTrM is close enought to p_Point2
					if (Operations::Dist(PointTrM, *p_Point2) < Operations::DistThresh(in_sigma))
					{
						// distance between parameter vectors of p_Point1 and p_Point2
						double dist = p_Point1->Distance(*p_Point2);

						if (dist < bestDist)
						{
							prevDist = bestDist;
							prevPoint = bestPoint;
							bestDist = dist;
							bestPoint = *p_Point2;
						}
					}
				}

				// if difference between the best corresponding point and 
				// the next cor. point is big enought
				if (prevDist / bestDist > in_t)
				{
					std::vector<C_Point2D> P;
					P.push_back(C_Point2D(p_Point1->X(), p_Point1->Y(), p_Point1->Numb()));
					P.push_back(C_Point2D(bestPoint.X(), bestPoint.Y(), bestPoint.Numb()));
					out_inl.Add(P);
				}
			}
		}
	  

	};
}
#endif //MODEL_H_
