#ifndef RANSAC__H_
#define RANSAC__H_

//library includings
#include "num_linalg.h"

// own includings
#include "Model.h"
#include "Point2D.h"
#include "Points2DSet.h"

namespace Basics3D
{
	// in_p is a probability, that at least one of the random 
	//      samples of m_samp points is free from outliers
	// in_t is the threshold for distance between pixels
	template<class Operations>
	void Ransac(Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &out,
		         T_Points2DSet<C_Point2D> &out_Inliers,
		         T_Model<Operations> &in_Model,
						 T_Points2DSet<C_Point2D> &in_PointsSet,
						 int in_size,
						 double in_sigma,
					   double in_p = 0.99)
  {
		double N = 100000;              // number of selections
		int maxInl = 0;					          // max number of inliers
		int size = in_PointsSet.Size();	// number of pairs in input set
				
		for (int sampleCount = 0; sampleCount < N; sampleCount++)
		{
			// Get normalized sample and normalization matrices
			T_Points2DSet<C_Point2D> sample(2);
			in_Model.GetSample(sample);
			
			// Compute matrix
			std::vector<Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3>> Matrices;
			in_Model.ComputeMatrices(Matrices, sample, in_sigma);
			
			// for homography matrix Matrices.size = 1,
			// for fundamental matrix Matrices.size = 1 or 3.
			for (int j = 0; j < (int)Matrices.size(); j++)
			{
				Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> M(3, 3);
				M = Matrices[j];
				
				// Count the number of inliers
				int numbInl = 0;
				std::vector<C_Point2D> inlPoints;

				for (int i = 0; i < size; i ++)
				{
					if (in_Model.IsInlier(M, i, in_sigma))
					{
						numbInl++;

						inlPoints.push_back(in_PointsSet[2 * i]);
						inlPoints.push_back(in_PointsSet[2 * i + 1]);
					}
				}

				// if new the best appropriate model is found
				if (numbInl > maxInl)
				{
					maxInl = numbInl;
					out = M;

					out_Inliers.Clear();
					out_Inliers.Add(inlPoints);
					
					/**/std::cout << std::endl;
					/**/T_MatrixOperations::Scale(M, M);
					/**/T_MatrixOperations::Print(M);
					/**/std::cout << std::endl;
					std::cout << std::endl << "inl = " << numbInl << "   ";
				}

				// recompute the number of selections
				double eps = 1 - (double)numbInl / in_size; // probability that any selected data
																								 // point is outlier
				if (eps <= 0.999)
				{
					double newN = log10(1 - in_p) / log10(1 - pow(1 - eps, 4.0));

					if (N > newN)
					{
						N = newN;
						std::cout << "N = " << N << std::endl;
					}
				}
				///**/std::cout << "i = " << sampleCount << "    N = " << N << "\n";
			}
		}
		
		/////////////////////////////////////////////////////////////////////////////////////////*/
		/**/std::cout << "\n\n__________________________________________________________\n" <<
		/**/						 "inl = " << maxInl << "   N = " << N << "\n\n";
		/**/T_MatrixOperations::Scale(out, out);
		/**/T_MatrixOperations::Print(out); 
		/**/std::cout << "\n\n";
	  
		// Recalculate with all the inliers
		if (maxInl > Operations::SampleSize())
		{
			Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> prevResult(0, 0);
			int prevNumbInl = maxInl - 1;
			int numbInl = maxInl;
			
			while ( numbInl > prevNumbInl)
			{
				prevNumbInl = numbInl;
				prevResult = out;
				
				std::vector<Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3>> vect;
/*?*/		in_Model.ComputeMatrices(vect, out_Inliers, in_sigma);
				
				std::vector<C_Point2D> newInliers;

				if (vect.size() > 0)
				{
					out = vect[0];
					
					numbInl = 0;

					for (int i = 0; i < size; i++)
					{
						if (in_Model.IsInlier(out, i, in_sigma))
						{
							numbInl++;
							
							newInliers.push_back(in_PointsSet[2 * i]);
							newInliers.push_back(in_PointsSet[2 * i + 1]);
						}
					}

					T_MatrixOperations::Scale(out, out);
					T_MatrixOperations::Print(out);
					std::cout << std::endl << std::endl;

					std::cout << std::endl << " inl = " << out_Inliers.Size() << std::endl;
				}
				if (numbInl > prevNumbInl)
				{
					out_Inliers.Clear();
					out_Inliers.Add(newInliers);
				}
			}

			if (prevNumbInl > numbInl)
			{
				out = prevResult;
			}

			std::cout << std::endl << "result" << std::endl;
			T_MatrixOperations::Print(out);
			std::cout << std::endl << std::endl;

			std::cout << std::endl << " inl = " << out_Inliers.Size() << std::endl;
		}
	}
}

#endif // RANSAC__H_
