#ifndef MARK_CENTERS_H
#define MARK_CENTERS_H

//system includings
#include <math.h>

// library includings
#include "aimage.h"
#include "curvature/CurvatureFunc.h"

// own includings
#include "KMeans.h"
#include "CurveDetection/CenterClusterization.h"
#include "ClustViaAccum.h"

namespace Vtl
{
namespace CurveDetection
{
  template <typename IterType, typename T>
	T MarkCenters(IterType dst,
                std::vector< CircleCenter<T> > &centers,
		            IterType curvature,
								IterType gradDirX,
								IterType gradDirY,
								int w,
								int h,
								int nls,
								T threshold = T(1))
	{
    centers.clear();

    // initialize destination with 0
    IterType begin = dst;

    for (int y = 0; y < h; y++)
    {
      for (int x = 0; x < w; x++)
      {
        (*dst) = T(0);
        dst++;
      }
    }

    // build centers accumulator
    int shiftX = w / 3;
    int shiftY = h / 3;

    T max = T(0);

		for (int y = 0; y < h; y++)
    {
			for (int x = 0; x < w; x++)
			{
				if (fabs(*curvature) > threshold)
				{
          //std::cout << std::endl << "curv > thresh  " << x << "  " << y;
          T radius = Vtl::Curvature::CurvatureScale<T>::value_to_radius(*curvature);
					T modGrad = sqrt((*gradDirX) * (*gradDirX) + (*gradDirY) * (*gradDirY));
					int c_x = int(T(x) + radius * (*gradDirX) / modGrad + 0.5);
					int c_y = int(T(y) + radius * (*gradDirY) / modGrad + 0.5);
          c_x = c_x / 3 + shiftX;
          c_y = c_y / 3 + shiftY;

					if (c_x < w && c_x >= 0 && c_y < h && c_y >= 0)
					{
            IterType iter = begin + c_y * nls + c_x;

						*(iter) += T(1);

            if (*(iter) > max)
            {
              max = *(iter);
              //std::cout << std::endl << std::endl << "new max = " << max;
              //std::cout << "";
            }

            bool exists = false;
            int i = 0;

            //std::cout << std::endl << "centers.size = " << centers.size()
            //  << "   " << c_x << " " << c_y << "   exists = " << exists;

            for (; i < (int)centers.size(); i++)
            {
              //std::cout << std::endl << centers[i].x << "  " << centers[i].y;
              if (centers[i].x == c_x && centers[i].y == c_y)
              {
                exists = true;
                break;
              }
            }

            //std::cout << std::endl << "exists = " << exists << "   i = " << i;

            if (!exists)
            {
              centers.push_back(CircleCenter<T>(c_x, c_y));
            }

            //std::cout << std::endl << "centers.size = " << centers.size();

            centers[i].points.push_back(DatumR<T>(x, y, (*curvature)));
          }
				}
				
				curvature++;
				gradDirX++;
				gradDirY++;
			}
		}
    return max;
	}

  //---------------------------------------------------------------------------
  template <typename IterType, typename T>
  void BuildCenters(std::vector< DatumXY<T> > &centers,
                    std::vector<const DatumR<T> *> &points,
										IterType GradX,
										IterType GradY,
										int nls)
  {
    centers.clear();

	  int size = points.size();
	  for (int i = 0; i < size; i++)
	  {
		  DatumR<T> p = *(points[i]);
			int x = p.x;
			int y = p.y;

		  T radius = CurvatureScale<T>::value_to_radius(p.val * p.sgn);
			T modGrad = sqrt((*(GradX + y*nls + x)) * (*(GradX + y*nls + x))
			               + (*(GradY + y*nls + x)) * (*(GradY + y*nls + x)));
			T c_x = T(x) + radius * (*(GradX + y*nls + x)) / modGrad;
			T c_y = T(y) + radius * (*(GradY + y*nls + x)) / modGrad;

      centers.push_back(DatumXY<T>(c_x, c_y, points[i]));
		}
	}

  /////////////////////////////////////////////////////////////////////////////
  template <typename T>
  T Gauss(T sigma,
          double x)
  {
    return T(1.0 / pow(2.0 * M_PI, 0.5) / sigma * 
             exp(- x * x / (2.0 * sigma * sigma)));
  }

  /////////////////////////////////////////////////////////////////////////////
  template <typename IterType, typename T>
  void CircleParams(int *c_x,
                    int *c_y,
                    T *r,
                    DatumR<T> &point,
                    IterType gradX,
                    IterType gradY,
                    int nls)
  {
    int x = point.x;
    int y = point.y;
    T curv = point.val * point.sgn;
    T dx = *(gradX + y * nls + x);
    T dy = *(gradY + y * nls + x);
    T modGrad = sqrt(dx * dx + dy * dy);

    // calculate parameters of probable circle
    *r = Vtl::Curvature::CurvatureScale<T>::value_to_radius(curv);
    *c_x = int(x + (*r) * dx / modGrad + 0.5);
		*c_y = int(y + (*r) * dy / modGrad + 0.5);
  }

  /////////////////////////////////////////////////////////////////////////////
  int FindCell(std::vector<Vtl::CurveDetection::Circle> &centers,
               int x,
               int y)
  {
    int n = centers.size();

    for (int i = 0; i < n; i++)
    {
      if (centers[i].x == x)
      {
        if (centers[i].y == y)
        {
          return i;
        }
      }
    }
    return (-1);
  }

  /////////////////////////////////////////////////////////////////////////////
  template <typename T>
  class RadiusMesh
  {
  private:
    const T r0;
    const T rMax;
    const T step;

    std::vector<T> r;

  public:
    RadiusMesh(T in_r0,
               T in_rMax,
               T in_step)
               : r0(in_r0), rMax(in_rMax), step(in_step)
    {
      for (float curR = r0; curR <= rMax; curR *= (T(1) + step))
      {
        r.push_back(curR);
      }
    }

    //-------------------------------------------------------------------------
    int size() const
    {
      return r.size();
    }

    //-------------------------------------------------------------------------
    T operator() (int i) const
    {
      if (i <= 0)
      {
        return r0;
      }

      if (i >= int(r.size()))
      {
        return rMax;
      }

      return r[i];
    }

    //-------------------------------------------------------------------------
    int prevNumb(T rad) const
    {
      return int(log(rad / r0) / log(1.0 + step));
    }
  };
  
  /////////////////////////////////////////////////////////////////////////////
  template <typename IterType, typename T>
  T FillCircleAccum(ImageContainer< std::vector<T> > &accum,
                    std::vector< std::vector<Circle> > &centers,
                    std::vector< DatumR<T> > &points,
                    IterType gradDirX,
                    IterType gradDirY,
                    const int w,
                    const int h,
                    const int nls,
                    const RadiusMesh<T> &rMesh,
                    const T sigma)
  {
    int rNumb = accum.size();   // number of radius layers
    int pNumb = points.size();  // number of input edge points

    int shiftX = 2 * w / 5;
    int shiftY = 2 * h / 5;

    T max = T(0);   // max value in the accumulator to be returned

    // for each edge point
    for (int p = 0; p < pNumb; p++)
    {
      int c_x;
      int c_y;
      T radius;
      CircleParams(&c_x, &c_y, &radius, points[p], gradDirX, gradDirY, nls); 

      c_x = c_x / 5 + shiftX;
      c_y = c_y / 5 + shiftY;

      if (c_x < w && c_x >= 0 && c_y < h && c_y >= 0)
      {
        radius *= points[p].sgn; // r = fabs(radius);
        int numb = rMesh.prevNumb(radius); // number of previous radius layer

        std::vector<double> dist;
        int minI;
        int maxI;

        if (numb < rNumb)
        {
          double r_1 = rMesh(numb);
          double r1 = rMesh(numb + 1);
          double d = r1 - r_1;

          // distances between curvature radius and radii on neighbour layers
          dist.resize(4);
          dist[1] = (radius - r_1 ) / d;
          dist[0] = dist[1] + 1.0f;
          dist[2] = (r1 - radius ) / d;
          dist[3] = dist[2] + 1.0f;
          
          minI = std::max(numb - 1, 0);
          maxI = std::min(numb + 2, rNumb - 1);
        }
        else
        {
          dist.push_back(0.0);
          numb = rNumb - 1;
          minI = maxI = numb - 1;
        }

        for (int i = minI; i <= maxI; i++)
        {
          // increase point in accumulator
          T coef = Gauss(sigma, dist[i - numb  + 1]);

          std::vector<T> *_radLayer = accum[i];
          T *iter = &((*_radLayer)[0]);
          iter += c_y * nls + c_x;
          *(iter) += coef;

          if (*(iter) > max)
          {
            max = *(iter);
          }

          // add source point to the corresponding center cell
          int cellNumb = FindCell(centers[i], c_x, c_y);

          if (cellNumb >= 0)
          {
            (centers[i])[cellNumb].pointsNumbs.push_back(p);
          }
          else
          {
            centers[i].push_back(Circle(c_x, c_y, p));
          }
        }
      }
    }  
   return max;
  }
}
}


#endif // MARK_CENTERS_H