#ifndef CORRELATION_MATRIX_H
#define CORRELATION_MATRIX_H

// library includings
#include "plf_functors.h"
#include "imalg_tools.h"

template< typename T >
class CorrelationMatrix
{
public:
  std::vector<T> LxLx;
  std::vector<T> LxLy;
  std::vector<T> LyLy;

private:
  int w;
  int h;
  int nls;

  /////////////////////////////////////////////////////////////////////////////
public:
  void BuildGray(T* Lx,
                 T* Ly,
                 int in_w,
                 int in_h,
                 int in_nls)
  {
    Init(in_w, in_h, in_nls);
    BuildForColorPlane(Lx, Ly, (*LxLx)[0], (*LxLy)[0], (*LyLy)[0]);
  }

  /////////////////////////////////////////////////////////////////////////////
  void BuildColor(T* Rx,
                  T* Ry,
                  T* Gx,
                  T* Gy,
                  T* Bx,
                  T* By,
                  int in_w,
                  int in_h,
                  int in_nls)
  {
    Init(in_w, in_h, in_nls);

    int size = w * h;

    std::vector<T> RxRx(size);
    std::vector<T> RxRy(size);
    std::vector<T> RyRy(size);
    BuildForColorPlane(Rx, Ry, &(RxRx[0]), &(RxRy[0]), &(RyRy[0]));

    std::vector<T> GxGx(size);
    std::vector<T> GxGy(size);
    std::vector<T> GyGy(size);
    BuildForColorPlane(Gx, Gy, &(GxGx[0]), &(GxGy[0]), &(GyGy[0]));

    std::vector<T> BxBx(size);
    std::vector<T> BxBy(size);
    std::vector<T> ByBy(size);
    BuildForColorPlane(Bx, By, &(BxBx[0]), &(BxBy[0]), &(ByBy[0]));

    Vtl::Plf::F31::Sum3<T> sumFunctor;
    Vtl::ImAlg::Tools::copy_3_1(&(RxRx[0]), w, h, nls, &(GxGx[0]), nls,
                                &(BxBx[0]), nls, &(LxLx[0]), nls, sumFunctor);
    Vtl::ImAlg::Tools::copy_3_1(&(RxRy[0]), w, h, nls, &(GxGy[0]), nls,
                                &(BxBy[0]), nls, &(LxLy[0]), nls, sumFunctor);
    Vtl::ImAlg::Tools::copy_3_1(&(RyRy[0]), w, h, nls, &(GyGy[0]), nls,
                                &(ByBy[0]), nls, &(LyLy[0]), nls, sumFunctor);
  }

  /////////////////////////////////////////////////////////////////////////////
  void MaxEigenValueAndVector(T* val,
                              T* vectX,
                              T* vectY)
  {
    Vtl::Plf::F33::DiZenzo<T> func;

    Vtl::ImAlg::Tools::copy_3_3(&(LxLx[0]), w, h, nls, &(LxLy[0]), nls, 
    &(LyLy[0]), nls, val, nls, vectX, nls, vectY, nls, func);
  }

  /////////////////////////////////////////////////////////////////////////////
private:
  void Init(int in_w,
            int in_h,
            int in_nls)
  {
    w = in_w;
    h = in_h;
    nls = in_nls;

    int size = w * h;

    LxLx.resize(size);
    LxLy.resize(size);
    LyLy.resize(size);
  }

  /////////////////////////////////////////////////////////////////////////////
  void BuildForColorPlane(T* pLx,
                          T* pLy,
                          T* pLxLx,
                          T* pLxLy,
                          T* pLyLy)
  {
    Vtl::Plf::F23::CorrMatrix<T, T&> functor;

    for (int y = 0; y < h; ++y)
    {
      for (int x = 0; x < w; ++x)
      {
        /*std::cout << std::endl << "y = " << y << "  x = " << x << " Lx = "
          << pLx[x] << "  Ly = " << pLy[x];
        std::cout << "  LxLx = " << *(pLxLx + x);
        std::cout << "  LxLy = " << pLxLy[x];
        std::cout << "  LyLy = " << pLyLy[x];*/
        functor(pLx[x], pLy[x], pLxLx[x], pLxLy[x], pLyLy[x]);
      }

      pLx += nls;
      pLy += nls;
      pLxLx += nls;
      pLxLy += nls;
      pLyLy += nls;
    }
  }
};

#endif CORRELATION_MATRIX_H