
#ifndef DIZENZO_H
#define DIZENZO_H

// std includings
#include <vector>

// library includings
#include "aimage.h"
#include "plf_functors.h"
#include "consolap.h"
#include "imalg_tools.h"

// own includings
#include "SaveImageArray.h"
#include "GradientOperations.h"
#include "CorrelationMatrix.h"
#include "Hysteresis.h"

template< typename T >
class DiZenzoEdgeDetector
{
private:
  int w;
  int h;
  int size;
  int nls;

  /////////////////////////////////////////////////////////////////////////////
public:
  void Apply(AImageMemRgb8* img,
             T* edges,
             T* xDir,
             T* yDir,
             T sigma,
             T nSigmasEnv,
             T hisThStart,
             T hisThEnd)
  {
    Init(img);

    // get color planes
    std::vector<T> r(size);
    std::vector<T> g(size);
    std::vector<T> b(size);

    GetColorPlanes(img, &(r[0]), &(g[0]), &(b[0]));

    // calculate partial derivatives
    std::vector<T> rx(size);
    std::vector<T> ry(size);
    std::vector<T> gx(size);
    std::vector<T> gy(size);
    std::vector<T> bx(size);
    std::vector<T> by(size);

    GradientComponents(&(rx[0]), &(ry[0]), &(r[0]), w, h, sigma, nSigmasEnv);
    GradientComponents(&(gx[0]), &(gy[0]), &(g[0]), w, h, sigma, nSigmasEnv);
    GradientComponents(&(bx[0]), &(by[0]), &(b[0]), w, h, sigma, nSigmasEnv);

    // build correlation matrix
    CorrelationMatrix<T> corrMatrix;
    corrMatrix.BuildColor(&(rx[0]), &(ry[0]), &(gx[0]), &(gy[0]), &(bx[0]), 
                                                        &(by[0]), w, h, nls);
    // find maximum eigen value of the correlation matrix
    // and corresponding eigen vector in each image point
		std::vector<T> edgesStrength(size);
    corrMatrix.MaxEigenValueAndVector(&(edgesStrength[0]), xDir, yDir);
		Tests::SaveNormImageArray(&(edgesStrength[0]), w, h, "Button_", "ColorGrad");
    
    // non maxima supression
    Vtl::Plf::Fpie::CannyNMS<T> nms(T(0));
    Vtl::Plf::Extent ext = nms.ext();
    Vtl::ImAlg::Tools::envTransf_3e1_1(ext, &(edgesStrength[0]), w, h, w, xDir, w,  
                                             yDir, w, edges, w, nms);
    // hysteresis
    Hysteresis<T> hyst(edges, w, h);
    hyst.Apply(hisThStart, hisThEnd, edges);
  }

  /////////////////////////////////////////////////////////////////////////////
private:
  void Init(AImageMemRgb8* img)
  {
    w = img->Width();
    h = img->Height();
    size = w * h;
    nls = img->GetPlane8(0)->NextLineShift();
  }

  /////////////////////////////////////////////////////////////////////////////
  void GetColorPlanes(AImageMemRgb8* img,
                      T* r,
                      T* g,
                      T* b)
  {
    AImageMem8* rPlane = img->GetPlane8(0);
    AImageMem8* gPlane = img->GetPlane8(1);
    AImageMem8* bPlane = img->GetPlane8(2);
    
    Vtl::Plf::F11::Convert<T> conv2T;
	  Vtl::ImAlg::Tools::copy_1_1((*rPlane)[0], w, h, nls, r, w, conv2T);
    Vtl::ImAlg::Tools::copy_1_1((*gPlane)[0], w, h, nls, g, w, conv2T);
    Vtl::ImAlg::Tools::copy_1_1((*bPlane)[0], w, h, nls, b, w, conv2T);
  }
};

#endif DIZENZO_H