#ifndef GRAD_MOD_H
#define GRAD_MOD_H

// system includings
#define _USE_MATH_DEFINES
#include <cmath>

// library includings
#include "convolve.h"
#include "plf_kernels_1d.h"

template< typename T >
void GradientComponents(T *Lx,
                        T *Ly,
                        const T *src,
                        const int w,
                        const int h,
                        float sigma,
                        float nSigmasEnv)
{
  const int sz = w * h;

	// copy source data
	memcpy(Lx, src, sz * sizeof(T));
	memcpy(Ly, src, sz * sizeof(T));

  // convolve with gaussian function and its derivative
	typedef Vtl::ImAlg::WndFun::Gauss<float,0, false> FF0; 
	FF0 f0(sigma, nSigmasEnv);
	typedef Vtl::Plf::Kernel::FKern<float,Vtl::Plf::Kernel::KT_Conv,true> K0;
	K0 ker0(f0);
	Vtl::ImAlg::Conv::ConvInplace<float,K0> g0(ker0);
	
	typedef Vtl::ImAlg::WndFun::Gauss<float,1,false> FF1; 
	FF1 f1(sigma, nSigmasEnv);
	typedef Vtl::Plf::Kernel::FKern<float,Vtl::Plf::Kernel::KT_Conv,true> K1;
	K1 ker1(f1);
	Vtl::ImAlg::Conv::ConvInplace<float,K1> g1(ker1);

  // Image derivative with respect to x
	Vtl::ImAlg::Conv::Conv2D::conv2DInplace(Lx, w, h, w, g1, g0);
	// Image derivative with respect to y
	Vtl::ImAlg::Conv::Conv2D::conv2DInplace(Ly, w, h, w, g0, g1);
}

///////////////////////////////////////////////////////////////////////////////
template< typename T >
void GradientAbsValue(T *grMod,
                      T *Lx,
                      T *Ly,
                      int size,
                      float sigma)
{
  for(int i = 0; i < size; i++)
  {
	  grMod[i] = T(sqrt(2 * M_PI) * sigma * sqrt(Lx[i] * Lx[i] + Ly[i] * Ly[i]));
  }
}

///////////////////////////////////////////////////////////////////////////////
template<typename T>
void Gradient(T *grMod,
              T *Lx,
              T *Ly,
              const T *src,
              const int w,
              const int h,
              float sigma,
              float nSigmasEnv)
{
  GradientComponents(Lx, Ly, src, w, h, sigma, nSigmasEnv);
  GradientAbsValue(grMod, Lx, Ly, w * h, sigma);
}

///////////////////////////////////////////////////////////////////////////////
template< typename T >
void EdgeComponents(T *edges,
                    T *gradMod,
                    T *Lx,
                    T *Ly,
                    int sz)
{
	for (int i = 0; i < sz; i++)
	{
	  T coef;

	  if ((*gradMod) == 0)
		  {
			coef = 1.0f;
		  }
		  else
		  {
			coef = (*edges) / (*gradMod);
		  }

	  *Lx *= coef;
	  *Ly *= coef;

	  edges++;
	  gradMod++;
	  Lx++;
	  Ly++;
	}
}

///////////////////////////////////////////////////////////////////////////////
template< typename T >
void Levelling(T *img,
               int sz,
               T level)
{
	for (int i = 0; i < sz; i++)
	{
	  if (*img > 0)
	  {
		*img = level;
	  }

	  img++;
	}
}


#endif // GRAD_MOD_H