#ifndef _AIMAGE_H
#define _AIMAGE_H
#include <string.h>
#include <assert.h>
#include "ytypes.h"
#include "progrcnt.h"

/** @defgroup AbstractImage  Handling with images 
 *  @{
 */
typedef uint8_t    Pixel8;
typedef Pixel8 *  Pixel8Ptr;
typedef uint16_t     Pixel16;
typedef Pixel16 * Pixel16Ptr;
typedef float     PixelFloat;
typedef PixelFloat * PixelFloatPtr;


//////////////////////////////////////////////////////////////////////
class AImage;
typedef AImage*         AImagePtr;
class AImageMem8;
typedef AImageMem8*     AImageMem8Ptr;
class AImageMem16;
typedef AImageMem16*    AImageMem16Ptr;
class AImageMemFloat;
typedef AImageMemFloat* AImageMemFloatPtr;
class AImageMemRgb8;
typedef AImageMemRgb8*  AImageMemRgb8Ptr;
class AImageMemRgb16;
typedef AImageMemRgb16* AImageMemRgb16Ptr;

/*! 
 @brief     .

 @par fname -  -
 @par indicator -   , . class ProgressCounter 
           indicator,    consolap.h
 @par img -   ,   AImage  0.   :
 @par replace -  false,       
       ,      img.   ,
       img      AImageMem16,  
        TrueColor,      . \n
       replace==true,        ,    
        . 
 @return        ,    . 
  0   .         exception,
     .\n
 <tt>   !</tt>    \c delete,   - 
    std::auto_ptr<AImage>,  \c       
  .\n
 @see AImageLoad, AImageSave, AImageCreate\n
   img   AImageLoad, AImageCreate   ,  
     .   ,    
  Windows       :
 @code
 class MyWindow 
	: public Window //  ,    ,  CMdiChild,  MFC
{
    .....
	AImage* image;
	VirtDib draw_obj;
	ZoomFactor2D zoom;
	BYTE *map_table;
    .....
public:
	MyWindow()
	{
		image=0; //   ,  ,     
		         // (\c delete) ,    .
		map_table=0; //         ,
		             //        ,
		             //     image  <tt>   </tt>!
		             //   ,     
		             //     ,    
		             //   . 
	}
	virtual ~MyWindow() {Clear();} //  
	void Clear()
	{
		if(image) 
		{
			delete image;
			image=0;
		}
	}
    .....
	// ,      WM_PAINT
	void Paint(HWND hWnd, HDC hdc)
	{
		///  ,      ,     .
		HBRUSH hBrush=CreateSolidBrush(...); 
		draw_obj.DefaultPaint(hWnd,hdc,hBrush,image,zoom,map_table);
		//  
	}
	//  - ,        
	// :       -    
	//   .
	void Load(const char* fname)
	{
		image=AImageLoad(image,true,fname,my_graphical_indicator_or_null);
	}
	void Save(const char* fname)
	{
		AImageSave(image,fname,my_graphical_indicator_or_null);	
	}
	void New()
	{
		image=AImageCreate(image,AImage::YImageClass::IMGC_GRAYSCALE,500,400,12);
	}
    .....
};
@endcode
*/
AImage * AImageLoad(AImage *img,bool replace,const char *fname,ProgressCounter* indicator=0);
/////////////////////////////////////////////////////////////////////////////////////////
/// @brief     .
///
/// @par fname -  -
/// @par indicator -   , . class ProgressCounter 
///           indicator,    consolap.h
/// @par img -    (),   AImage, 
///  .
///
/// !!!      .    
///      . ,   
///    *.bmp,    *.bmp   GRAYSCALE,  
///  TrueColor   ,       ! 
///      \b   ,     
///  ()  ,    \b    GrayScale.
/// @see AImageLoad, AImageSave, AImageCreate
/// 
void AImageSave(const AImage * img,const char *fname,ProgressCounter* indicator=0);
//////////////////////////////////////////////////////////////////////
// class AImage 
/// @brief  Base class for all images 
//////////////////////////////////////////////////////////////////////
class AImage
{
public:
//TYPES
	class Error
	:public ErrorStatusBase
	{
		char message[512];
	public:
		Error(const char *fname,int nline,const AImage *obj,const char *msgfmt,...);
	};
	struct Rgb
	{
		Pixel8 r;
		Pixel8 g;
		Pixel8 b;
	};
	struct Bgr
	{
		Pixel8 b;
		Pixel8 g;
		Pixel8 r;
	};
	/// \brief  ,   
	enum YImageClass
	{
		IMGC_UNKNOWN,
		IMGC_BILEVEL,
		IMGC_GRAYSCALE,
		IMGC_PALETTE,
		IMGC_RGB,
		IMGC_FLOAT
	};
	/// \brief      ,  
	///       .
	enum YImageStore
	{
		IMGS_Unknown,
		IMGS_AImageMem8,
		IMGS_AImageMem16,
		IMGS_AImageMemFloat,
		IMGS_AImageMemRgb8,
		IMGS_AImageMemRgb16
	};
	/// \brief          
	/// ,       
	///      .    
	///            
	///      .
	struct Info
	{
		int width;
		int height;
		int bits;
		YImageClass iclass;
		YImageStore istore;
	};
//object identification
	static  const char * ImageStoreName(YImageStore istore);
	static  const char * ImageClassName(YImageClass iclass);
  static  const char * ThisClassName();
  virtual const char * ClassName() const { return ThisClassName(); }
//store
	/// returns
	///   0 - if this is image container
	///   pointer - if this is reference to image container or reference to ...
	virtual AImage * GetFullImage() const { return 0;}
	virtual AImage::YImageStore GetImageStore() const=0;
//Creation/destroying	,load/Save, verification
	AImage() {}
	virtual ~AImage() {}
	virtual bool IsOk() const=0;
//  virtual void Load(const char *filename,ProgressCounter* indicator=0)
//		{::AImageLoad(this,false,filename,indicator);}
//  virtual void Save(const char *filename,ProgressCounter* indicator=0) const
//		{::AImageSave(this,filename,indicator);}
  virtual bool Create(short width,short height,short bitsPerPix,bool hasPalette=false)=0;
  virtual void Clear()=0;
// Image info
	virtual int Width()  const=0;
	virtual int Height() const=0;
  virtual int Bits()   const=0;
  virtual int NextLineShift() const=0;
	virtual int NumPlanes() const { return 1;}
  virtual AImage  * GetPlane(int plane) { return this;}
  virtual const AImage  * GetPlane(int plane) const { return this;}
	virtual YImageClass  ImageClass() const=0;
	virtual const char * ImageStore() const {	return AImage::ImageStoreName(GetImageStore()); }
	virtual const char * ImageClassName() const {	return AImage::ImageClassName(ImageClass()); }
	Info GetImageInfo() const 
	{
		Info inf;
		inf.width=Width();
		inf.height=Height();
		inf.bits=Bits();
		inf.iclass=ImageClass();
		inf.istore=GetImageStore();
		return inf;
	}

	static bool CopyPalette(AImage* to,const AImage* from); //if cant copy - returns false
	virtual int GetPaletteEntries() const { return 0;}
	virtual AImage::Rgb *Palette() const { return 0;}
	virtual void read(int x,int y,Pixel16 *data,int len) const=0;
///specific, not use without special permissions!
	virtual void ChangeBits(int bits) { } 
///specific, not use without special permissions!
	virtual void ChangeHeight(short new_height)
		{ Create(short(Width()),new_height,short(Bits()),ImageClass()==IMGC_PALETTE);	}
};

//////////////////////////////////////////////////////////////////////
bool AImageIsCompatible(AImage *img,AImage::YImageClass iclass,short width,short height,short bitsPerPix);
/////////////////////////////////////////////////////////////////////////////////////////
/// @brief    ().
///
/// @par img -   ,   AImage  0.    ,
///             . ,   
///             ,    
///               . 
///        img   AImageLoad, AImageCreate   ,  
///          .
///
/// @par img -    (),   AImage, 
///  .
/// @par iclass -   AImage::YImageClass,   . 
/// @par width  -    
/// @par height -    
/// @par bitsPerPix -    ,     -   
///                    <tt>  </tt>,    TrueColor 24 bkb 48,
///                     - 8  16 .    \c float - 32 .
/// @return        ,    . 
///  0   .         exception,
///     .\n
/// <tt>   !</tt>    \c delete,   - 
///    std::auto_ptr<AImage>,  \c       
///  .\n
///
/// @see AImageLoad, AImageSave, AImageCreate
AImage *AImageCreate(AImage *img,AImage::YImageClass iclass,short width,short height,short bitsPerPix);
inline AImage *AImageCreate(AImage *img,AImage::Info inf)
{
	return AImageCreate(img,inf.iclass,inf.width,inf.height,inf.bits);
}
//////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// class AImageMem8 
/// @brief  Interface for 8 bit per pixel container 
////////////////////////////////////////////////////////////////////////////////
class AImageMem8
: virtual public AImage
{
public:
	typedef Pixel8 Pixel;
  static const char *ThisClassName();
  virtual const char *ClassName() const { return ThisClassName(); }
	virtual AImage::YImageStore GetImageStore() const { return AImage::IMGS_AImageMem8; }
	AImageMem8() {}
  virtual ~AImageMem8() {}
  virtual Pixel8 * operator[](int y) const=0;
	static void SetTo(AImageMem8* img,Pixel8 color);
  virtual void SetTo(Pixel8 color) {SetTo(this,color);}
};
////////////////////////////////////////////////////////////////////////////////
// class AImageMem16 
/// @brief  interface for 16 bit per pixel container 
////////////////////////////////////////////////////////////////////////////////
class AImageMem16
: virtual public AImage
{
private:
	AImageMem16(const AImageMem16& im);
public:
	typedef Pixel16 Pixel;
  static const char *ThisClassName();
  virtual const char *ClassName() const { return ThisClassName(); }
	virtual AImage::YImageStore GetImageStore() const { return AImage::IMGS_AImageMem16; }
	AImageMem16() {}
	virtual ~AImageMem16() { }
  virtual Pixel16* operator[](int y) const=0;
	static void SetTo(AImageMem16* img,Pixel16 color);
  virtual void SetTo(Pixel16 color) {SetTo(this,color);}
};

////////////////////////////////////////////////////////////////////////////////
// class AImageMemFloat 
/// @brief  Interface for float per pixel container //////
////////////////////////////////////////////////////////////////////////////////
class AImageMemFloat
: virtual public AImage
{
public:
	typedef PixelFloat Pixel;
  static const char *ThisClassName();
  virtual const char *ClassName() const { return ThisClassName(); }
	virtual AImage::YImageStore GetImageStore() const { return AImage::IMGS_AImageMemFloat; }
	AImageMemFloat() {}
  virtual ~AImageMemFloat() {}
  virtual PixelFloat * operator[](int y) const=0;
  virtual void SetTo(PixelFloat color)=0;
};

////////////////////////////////////////////////////////////////////////////////
// class AImageMemRgb8
/// @brief  Interface for TrueColor 8 bit per each pixel sample container 
////////////////////////////////////////////////////////////////////////////////
class AImageMemRgb8
: virtual public AImage
{
public:
	typedef AImageMem8 PlaneImageType;
	typedef Pixel8 Pixel;
  static const char *ThisClassName();
  virtual const char *ClassName() const { return ThisClassName(); }
	virtual AImage::YImageStore GetImageStore() const { return AImage::IMGS_AImageMemRgb8; }
	AImageMemRgb8() {}
  virtual ~AImageMemRgb8() {}
	virtual int NumPlanes() const { return 3;}
  virtual AImageMem8  * GetPlane8(int plane) const=0;
  virtual AImage* GetPlane( int plane) { return GetPlane8(plane); }
  virtual const AImage* GetPlane( int plane) const { return GetPlane8(plane); }
	static void SetTo(AImageMemRgb8* img,Rgb color);
	virtual void SetTo(Rgb color) {SetTo(this,color);}
};

////////////////////////////////////////////////////////////////////////////////
// class AImageMemRgb16
/// @brief  Interface for TrueColor 16 bit per each pixel sample container 
////////////////////////////////////////////////////////////////////////////////
class AImageMemRgb16
: virtual public AImage
{
public:
	typedef AImageMem16 PlaneImageType;
	typedef Pixel16 Pixel;
  static const char *ThisClassName();
  virtual const char *ClassName() const { return ThisClassName(); }
	virtual AImage::YImageStore GetImageStore() const { return AImage::IMGS_AImageMemRgb16; }
	AImageMemRgb16() {}
  virtual ~AImageMemRgb16() {}
	virtual int NumPlanes() const { return 3;}
//  virtual AImageMem8  * operator[](int plane) const=0;
  virtual AImageMem16  * GetPlane16(int plane) const=0;
  virtual AImage* GetPlane( int plane) { return GetPlane16(plane); }
  virtual const AImage* GetPlane( int plane) const { return GetPlane16(plane); }
  virtual void SetTo(Rgb color)=0;
};

////////////////////////////////////////////////////////////////////////////////
/// Working with image references, use ONLY as base class
////////////////////////////////////////////////////////////////////////////////
template<class ImgT>
class tplAImageReference
: public ImgT
{
protected:
	ImgT *img;
public:
	virtual AImage * GetFullImage() const { return img;}
  virtual const char * ClassName() const { return (img)?img->ThisClassName():0; }
	virtual AImage::YImageStore GetImageStore() const {return (img)?img->GetImageStore(): AImage::IMGS_Unknown;}

	tplAImageReference() { img=0;}
	virtual ~tplAImageReference() {}
	virtual void Init(ImgT *_img) {img=_img;}

	virtual bool IsOk() const { return (img)?img->IsOk():false;}
  virtual void Load(const char *filename,ProgressCounter* indicator=0)
		{}
  virtual void Save(const char *filename,ProgressCounter* indicator=0) const
		{}
  virtual bool Create(short width,short height,short bitsPerPix,bool hasPalette=false)
		{return false;}
  virtual void Clear() {}
	virtual int Width()  const {return (img)?img->Width() :0;}
	virtual int Height() const {return (img)?img->Height() :0;}
  virtual int Bits()   const {return (img)?img->Bits():0;}
  virtual int NextLineShift() const {return (img)?img->NextLineShift():0;}
	virtual int NumPlanes() const {return (img)?img->NumPlanes():0;}
	virtual AImage::YImageClass  ImageClass() const {return (img)?img->ImageClass() : AImage::IMGC_UNKNOWN;}
	virtual int GetPaletteEntries() const {return (img)?img->GetPaletteEntries():0;}
	virtual AImage::Rgb *Palette() const {return (img)?img->Palette():0;}
private:
	virtual void ChangeHeight(short new_height) { }
	virtual void read(int x,int y,Pixel16 *data,int len) const {if(img) img->read(x,y,data,len); }
};
////////////////////////////////////////////////////////////////////////////////
/// working with image fragment, use ONLY as base class
////////////////////////////////////////////////////////////////////////////////
template<class ImgT>
class tplAImageFragment
: public tplAImageReference<ImgT>
{
protected:
	int fx;
	int fy;
	int fwidth;
	int fheight;
public:
//	AImageFragment() : tplAImageReference() {}
	tplAImageFragment():tplAImageReference<ImgT>() { _Init(0);}
	virtual ~tplAImageFragment() {}
	virtual void Init(ImgT *_img) {tplAImageReference<ImgT>::Init(_img);}
	virtual void Init(int x,int y,int width,int height) {	_Init(tplAImageReference<ImgT>::GetFullImage(),x,y,width,height);}
	virtual void Init(ImgT *_img,int x,int y,int width,int height) {Init(_img); Init(x,y,width,height); }
	virtual int Width()  const { return fwidth;}
	virtual int Height() const { return fheight;}
	virtual int Bits()   const { return (tplAImageReference<ImgT>::img)?tplAImageReference<ImgT>::img->Bits():0;}
	virtual int NextLineShift() const { return (tplAImageReference<ImgT>::img)?tplAImageReference<ImgT>::img->NextLineShift():0;}
private:
protected:
	void _Init(const AImage *img)
	{
		fx=fy=0;
		if(img && img->IsOk())
		{
			fwidth=img->Width();
			fheight=img->Height();
		}
		else
		{
			fwidth=fheight=0;
		}
	}
	void _Init(const AImage *img,int x,int y,int width,int height)
	{
		if(img)
		{
			fx=IN_BOUNDS(x,0,img->Width()-1);
			fy=IN_BOUNDS(y,0,img->Height()-1);
			fwidth =fx+MAX(width,0);
			fheight=fy+MAX(height,0);
			fwidth =IN_BOUNDS(fwidth ,0,img->Width()) -fx;
			fheight=IN_BOUNDS(fheight,0,img->Height())-fy;
		}
		else
		{
			fx=fy=fwidth=fheight=0;
		}
	}
};

////////////////////////////////////////////////////////////////////////////////
// class ImageFragment8 
/// @brief 
////////////////////////////////////////////////////////////////////////////////
class ImageFragment8
: public tplAImageFragment<AImageMem8>
{
public:
  static const char *ThisClassName();
  virtual const char *ClassName() const { return ThisClassName(); }
	ImageFragment8() {}
	virtual ~ImageFragment8() {}
	//
  virtual Pixel8 * operator[](int y) const { return (img)?((*img)[y+fy]+fx):0;}
  virtual void SetTo(Pixel8 color) {  AImageMem8::SetTo(this,color); }
	virtual void read(int x,int y,Pixel16 *data,int len) const {if(img) img->read(x+fx,y+fy,data,len); }
};

////////////////////////////////////////////////////////////////////////////////
// class ImageFragment16 
/// @brief 
////////////////////////////////////////////////////////////////////////////////
class ImageFragment16
: public tplAImageFragment<AImageMem16>
{
public:
  static const char *ThisClassName();
  virtual const char *ClassName() const { return ThisClassName(); }
	ImageFragment16() {}
	virtual ~ImageFragment16() {}
  virtual Pixel16 * operator[](int y) const { return (img)?((*img)[y+fy]+fx):0;}
  virtual void SetTo(Pixel16 color) {  AImageMem16::SetTo(this,color); }
	virtual void read(int x,int y,Pixel16 *data,int len) const {if(img) img->read(x+fx,y+fy,data,len); }
};
////////////////////////////////////////////////////////////////////////////////
/// class ImageFragmentRgb8 
/// @brief 
////////////////////////////////////////////////////////////////////////////////
class ImageFragmentRgb8
: public tplAImageFragment<AImageMemRgb8>
{
	ImageFragment8 _pl[3];
	AImageMem8 *pl[3];
public:
  static const char *ThisClassName();
  virtual const char *ClassName() const { return ThisClassName(); }
	ImageFragmentRgb8() {}
	virtual ~ImageFragmentRgb8() {}
	virtual void Init(AImageMemRgb8 *_img)
	{
		tplAImageReference<AImageMemRgb8>::Init(_img);
		_pl[0].Init(_img->GetPlane8(0)); pl[0]=&(_pl[0]);
		_pl[1].Init(_img->GetPlane8(1)); pl[1]=&(_pl[1]);
		_pl[2].Init(_img->GetPlane8(2)); pl[2]=&(_pl[2]);
	}
	virtual void Init(int x,int y,int width,int height)
	{
		_Init(GetFullImage(),x,y,width,height);
		_pl[0].Init(x,y,width,height);
		_pl[1].Init(x,y,width,height);
		_pl[2].Init(x,y,width,height);
	}
  virtual AImageMem8  * GetPlane8(int plane) const { return pl[plane];}
  virtual void SetTo(Rgb color) {  AImageMemRgb8::SetTo(this,color); }
	virtual void read(int x,int y,Pixel16 *data,int len) const {if(img) img->read(x+fx,y+fy,data,len); }
};

////////////////////////////////////////////////////////////////////////////////
/// class ImageMem16Stuff 
/// @brief 
////////////////////////////////////////////////////////////////////////////////
class ImageMem16Stuff
: public AImageMem16
{
	short wid;
  short len;
  short bits;
  Pixel16 *data;
private:
	ImageMem16Stuff(const ImageMem16Stuff& im);
public:
  static const char *ThisClassName();
  virtual const char *ClassName() const { return ThisClassName(); }

	ImageMem16Stuff(Pixel16 *_extern_data,short width,short height,short bitsPerPixel=10)
  {
		data=_extern_data;
    wid=width;
    len=height;
    bits=bitsPerPixel;
  }
	virtual ~ImageMem16Stuff() {}

	virtual bool IsOk() const { return (data!=0); }
  virtual void Load(const char *ozfilename,ProgressCounter* indicator=0);
//  virtual void Save(const char *ozfilename,ProgressCounter* indicator=0) const{}
  virtual bool Create(short width,short height,short bitsPerPix,bool hasPalette=false);
  virtual void Clear();
	virtual int Width()  const { return wid;}
	virtual int Height() const { return len;}
  virtual int Bits()   const { return bits;}
  virtual int NextLineShift() const { return wid;}
	virtual YImageClass  ImageClass() const { return (data)?IMGC_GRAYSCALE:IMGC_UNKNOWN;}
	//
  virtual Pixel16 * operator[](int y) const {return (data+long(wid)*y); }//{ return line[y]; } //@@need test bounds and NULL
	virtual void read(int x,int y,Pixel16 *data2,int len) const { ::memcpy(data2,operator[](y)+x,len); }
//Exceptions
  class Error
  : public ErrorStatusBase
	{
	public:
    enum ErrorStatus
    {
      errOK,
      errInsufficientMemory,
      errCantLoadObject,
      errCantCreateObject,
      errCantClearObject,
      errUnknown=0xFFFF
    };
  private:
    ErrorStatus err;
	public:
		Error(const char *fname,int nline,const ImageMem16Stuff *obj,bool del_msg,ErrorStatus id_err,const char *msgfmt,...);
		virtual ~Error() {}
    ErrorStatus   ErrorID() { return err;}
	};
};

#define IMAGE_MAX_FRAME	64

////////////////////////////////////////////////////////////////////////////////
// class YImageFrame
/// 
/// @brief 
/// 
////////////////////////////////////////////////////////////////////////////////
class YImageFrame
{
public:
	typedef uint16_t Pixel;
private:
	const AImage *img;
	Pixel  buf[IMAGE_MAX_FRAME*(IMAGE_MAX_FRAME+1)];
	int		sz; //frame size is square, not larger than IMAGE_MAX_FRAME
	          // data placed as array size of 'sz*sz', the remaing space not used.
	int x_cent,y_cent;//frame center in the original image coordinates.
	//int shift_to_leftup;
	int last_line_offset;//offset from start buffer to last line first byte
	int dx,dy;
	int dx2,dy2;
	int height;
	int width;
public:
	YImageFrame() {	img=0; sz=0; }
	YImageFrame(const AImage *image,int size) {	Init(image,size); }
	~YImageFrame() {}
	void Init(const AImage *image,int size)
	{
		assert(image);
		assert(size>0 && size<=IMAGE_MAX_FRAME);
		img=image;
		sz=size;
		width=img->Width();
 		height=img->Height();
		dx=sz/2;//offset from left up to frame center;
		dy=sz/2;
		dx2=sz-dx-1;
		dy2=sz-dy-1;
		//shift_to_leftup=-(dy*sz+dx);
		last_line_offset=sz*sz;
	}


	inline const Pixel* getLeftUp() const {	return buf+sz; }
	inline const Pixel* operator[](int y) const {	return buf+(y+1)*sz;}
	int GetFrameSize() const {	return sz; };
	inline int X() const { return x_cent; }
	inline int Y() const { return y_cent; }

	void SetTo(int x,int y)
	{
		x_cent=x;
		y_cent=y;
		x-=dx;
		y-=dy+1;
		Pixel *b=buf;
		for(int i=0; i<(sz+1); i++)
		{
			int yy;
			if(y<0)
				yy=-y-1; //(-1) -   
			else if(y>=height)
				yy=height-(y-height+2);
			else
				yy=y;
			readSlow(x,yy,b);//,sz);
			y++;
			b+=sz;
		}
	}
	int NumberFastMove() const { return height-1-(y_cent+dy2);}
	int GetLeftRightBorderWidth() const { return (sz+4)/2; }
	inline void MoveDown()
	{	//     !
		++y_cent;
		::memmove(buf, buf+sz, sizeof(Pixel)*last_line_offset);
		img->read(x_cent-dx,y_cent+dy2,buf+last_line_offset,sz);
	}

	inline void MoveDownSlow()
	{	//   ,    !
		++y_cent;
		::memmove(buf, buf+sz, sizeof(Pixel)*last_line_offset);
		int y=y_cent+dy2;
		if(y<height)
			readSlow(x_cent-dx,y,buf+last_line_offset);
		else
		{
			y=sz-(y-height+1)*2+1;//  ... -    !!
			//    ??
			::memcpy(buf+last_line_offset,buf+sz*y,sizeof(Pixel)*sz);
		}
	}
	void readSlow(int x,int y,Pixel16 *buf) const
	{
		if(x<0)
		{
			int ll=-x;
			buf+=ll;
			img->read(0,y,buf,sz-ll);
			Pixel16 *b=buf;
			while(ll--)
				*(--b)=*(buf++);
		}
		else
		{
			int s=width-x;//  ... -    !!!
			int rr=sz-s;
			if(rr>0)
			{
				img->read(x,y,buf,s);
				buf+=s;
				Pixel16 *b=buf;
				while(rr--)
					*(b++)=*(--buf);
			}
			else
				img->read(x,y,buf,sz);
		}
	}
};
/** @} */ // end of AbstractImage

//////////////////////////////////////////////////////////////////////
#endif // _AIMAGE_H
