
#ifndef _IMPL_IMTRANSF_H
#define _IMPL_IMTRANSF_H

#include <assert.h>
#include <algorithm>

#include "ymem.h"

class ImageTransformCellImpl;

class TransformCell
{
public:
	TransformCell( ImageTransformCellImpl& transform, int x, int y);

	const YPoint<double>* getCellPICS() const;
	const YPoint<double>* getCellICS() const;
	const YPoint<double>* getCellRCS() const;
	const YPoint<double>* getCellPRCS() const;
private:
	void fill(int x, int y);

	ImageTransformCellImpl& transform_;

	YPoint<double> pics_[4];
	YPoint<double> ics_[4];
	YPoint<double> rcs_[4];
	YPoint<double> prcs_[4];
};

struct DMatrixElem
{
	DMatrixElem* next;
	YPoint<int> p;
};

class DistributionMatrix
{
public:
	DistributionMatrix( int width = 0, int height = 0, size_t capacity = 0)
	{ clear(); setSize(width,height,capacity); }
	void clear() { matrix_.clear(); allocator_.DeInit(); }
	void setSize( int width, int height, size_t capacity)
	{
		w_ = width; 
		h_ = height;
		matrix_.resize(w_*h_,0);
		if(capacity)
			allocator_.Init( sizeof(DMatrixElem), capacity);
	}
	void add( const YPoint<int>& cell, const YPoint<int>& p)
	{
		if( cell.x < 0 || cell.x >= w_ || cell.y < 0 || cell.y >= h_ )
			return;
		DMatrixElem* pElem = static_cast<DMatrixElem*>(allocator_.malloc());
		DMatrixElem** ppMElem = at(cell);
		pElem->next = *ppMElem;
		pElem->p = p;
		*ppMElem = pElem;
	}
	std::vector<const DMatrixElem*> collectRaduis(
		const YPoint<int>&prp, int r, int* pn = 0) const;
	bool empty( const YPoint<int>& p) const
	{ return 0 == *at(p); }
private:
	DMatrixElem** at( const YPoint<int>& p)
	{ 
		assert( p.x >= 0 && p.x < w_ && p.y >= 0 && p.y < h_ );
		return &matrix_[p.y*w_+p.x];
	}
	DMatrixElem* const * at( const YPoint<int>& p) const
	{ 
		assert( p.x >= 0 && p.x < w_ && p.y >= 0 && p.y < h_ );
		return &matrix_[p.y*w_+p.x];
	}
	inline void collectCell(
		std::vector<const DMatrixElem*>& elems, const DMatrixElem* elem) const;
private:
	std::vector<DMatrixElem*> matrix_;
	YSmallAllocLtd allocator_;
	int w_;
	int h_;
};

class CoverageMatrix
{
public:
	CoverageMatrix( int width = 0, int height = 0)
	{ clear(); setSize(width,height); }
	void clear() { matrix_.clear(); zoomMatrix_.clear(); }
	void setSize( int width, int height)
	{
		w_ = width; 
		h_ = height;
		matrix_.resize(w_*h_, YPoint<int>(-1,-1));
		zoomMatrix_.resize(w_*h_,-1);
	}
	void set( const YPoint<int>& cell, const YPoint<int>& value, double zoomValue)
	{
		if( cell.x < 0 || cell.x >= w_ || cell.y < 0 || cell.y >= h_ )
			return;
		if( getZoom(cell) != -1 && getZoom(cell) > zoomValue )
			return;
		matrix_[cell.y*w_+cell.x] = value;
		zoomMatrix_[cell.y*w_+cell.x] = zoomValue;
	}
	void set( const YRect<int>& area, const YPoint<int>& value, double zoomValue)
	{
		int bottom = area.bottom();
		int right = area.right();
		for( int y = area.top(); y <= bottom; ++y)
			for( int x = area.left(); x <= right; ++x)
				set( YPoint<int>(x,y), value, zoomValue);
	}
	void set( 
		const YPoint<double>* cell,
		const YRect<int>& area, 
		const YPoint<int>& value,
		double zoomValue);
	const YPoint<int>& get( const YPoint<int>& cell) const
	{
		return matrix_[cell.y*w_+cell.x];
	}
	double getZoom( const YPoint<int>& cell) const
	{
		return zoomMatrix_[cell.y*w_+cell.x];
	}
	bool empty( const YPoint<int>& cell) const
	{
//		return matrix_[cell.y*w_+cell.x].x == -1;
		return zoomMatrix_[cell.y*w_+cell.x] == -1;
	}
private:
	std::vector<YPoint<int> > matrix_;
	std::vector<double> zoomMatrix_;
	int w_;
	int h_;
};

class ProcessMatrix
{
public:
	ProcessMatrix( int width = 0, int height = 0)
		: matrix_(0) { clear(); setSize(width,height); }
	~ProcessMatrix() { clear(); }
	void clear() { delete[] matrix_; matrix_ = 0; }
	void setSize( int width, int height)
	{
		w_ = width; 
		h_ = height;
		if(matrix_)
			clear();
		matrix_ = new bool[w_*h_];
		std::fill(&matrix_[0],&matrix_[w_*h_],false);
	}
	void set( const YPoint<int>& cell, bool value)
	{
		if( cell.x < 0 || cell.x >= w_ || cell.y < 0 || cell.y >= h_ )
			return;
		matrix_[cell.y*w_+cell.x] = value;
	}
	void set( const YRect<int>& area, bool value)
	{
		int bottom = area.bottom();
		int right = area.right();
		for( int y = area.top(); y <= bottom; ++y)
			for( int x = area.left(); x <= right; ++x)
				set( YPoint<int>(x,y), value);
	}
	const bool& at( const YPoint<int>& cell) const
	{
		return matrix_[cell.y*w_+cell.x];
	}
	bool& at( const YPoint<int>& cell)
	{
		return matrix_[cell.y*w_+cell.x];
	}
private:
	bool* matrix_;
	int w_;
	int h_;
};


typedef std::vector<YPoint<double> > ITPolygon;

class ImageTransformImpl
{
public:
	ImageTransformImpl();
	virtual ~ImageTransformImpl() {}

	virtual bool run( 
		ImageProxyVector* outputVector, 
		const YRect<int>& outputFragment,
		ProgressCounter* indicator) = 0;
	virtual bool calculateOutputData( ProgressCounter* indicator) = 0;

	void setBackgroundColor( bool doCorrect, int value);
	void setErrorColor( bool doCorrect, int value);
	void setFlipHorizontal( bool value);
	void setFlipVertical( bool value);
	void setGridAdjustment( const YPoint<double>& gridNode);
	int width() const;
	int height() const;
	std::vector<int> bits() const;
	std::vector<AImage::YImageClass> imageClass() const;
	YPoint<double> base() const;
	bool init(
		const ImageProxyVector* vector,
		const YRect<int>& inputFragment,
		const CoordArray2D* coords,
		const ATransformFunction& transformFunction,
		double scale,
		ProgressCounter* indicator = 0,
		const ImageTransform::EdgeMethod& edgeMethod = ImageTransform::None);

	void checkInputParameters(
		const ImageProxyVector* vector,
		const YRect<int>& inputFragment,
		const CoordArray2D* coords,
		const ATransformFunction& transformFunction,
		double scale,
		const ImageTransform::EdgeMethod& edgeMethod) const;
	void acceptInputParameters(
		const ImageProxyVector* vector,
		const YRect<int>& inputFragment,
		const CoordArray2D* coords,
		const ATransformFunction& transformFunction,
		double scale,
		const ImageTransform::EdgeMethod& edgeMethod);
	void calculateOutputCoords();

	inline void fillOutputPixel( AImageProxy* out, int x, int y, int value);

	// coordinate transform functions
	YPoint<double> RCStoPRCS( const YPoint<double>& p) const;
	YPoint<int> RCStoPRCSr( const YPoint<double>& p) const;
	YPoint<double> PRCStoRCS( const YPoint<int>& p) const;
	YPoint<double> PRCStoRCS( const YPoint<double>& p) const;
	bool canICStoRCS( const YPoint<double>& p) const;
	bool canRCStoICS( const YPoint<double>& p) const;
	YPoint<double> ICStoRCS( const YPoint<double>& p) const;
	YPoint<double> RCStoICS( const YPoint<double>& p) const;
	YPoint<double> ICStoPICS(
		const YPoint<double>& ip,
		const YPoint<int>* pip4,
		const YPoint<double>* ip4) const;
	inline YPoint<double> PICStoICS( const YPoint<int>& p) const;
	inline bool canPICStoICS( const YPoint<int>& p) const;
public:
	// input parameters
	const ImageProxyVector* inputVector_;
	YRect<int> inputFragment_;
	const CoordArray2D* inputCoords_;
	std::auto_ptr<const ATransformFunction> transformFunction_;
	double scale_;
	int inputWidth_;
	int inputHeight_;
	int inputImageWidth_;
	int inputImageHeight_;
	// output data
	size_t outputWidth_;
	size_t outputHeight_;
	YRect<double> outputArea_;
	ImageProxyVector* outputVector_;
	YRect<int> outputFragment_;
	CoordArray2D outputCoords_;
	// algorithm behavior
	bool flipH_;
	bool flipV_;
	int backgroundColor_;
	bool edgeH_;
	bool edgeV_;
	int errorColor_;
	bool enableGridAdjustment_;
	YPoint<double> gridNode_;
//	std::auto_ptr<AImageMemRgb8> debugImg;
private:
	ImageTransformImpl( const ImageTransformImpl& );
	ImageTransformImpl& operator=( const ImageTransformImpl& );
};

class ImageTransformCellImpl: public ImageTransformImpl
{
public:
	ImageTransformCellImpl() {}
	virtual ~ImageTransformCellImpl() {}

	bool run( 
		ImageProxyVector* outputVector, 
		const YRect<int>& outputFragment,
		ProgressCounter* indicator);
	bool calculateOutputData( ProgressCounter* indicator);
public:
	ProcessMatrix processed_;
};

class ImageTransformMatrixImpl: public ImageTransformImpl
{
	enum
	{
		MaxFindRadius = 20
	};
public:
	ImageTransformMatrixImpl();
	virtual ~ImageTransformMatrixImpl() {}

	bool run( 
		ImageProxyVector* outputVector, 
		const YRect<int>& outputFragment,
		ProgressCounter* indicator);

	void calculateInputPolygon();
	bool calculateOutputData( ProgressCounter* indicator);

	YRect<int> makeCellRect( const YPoint<double>* pOutputCoordsPoint);
	void makeCellPoly(
		const YPoint<double>* pOutputCoordsPoint, YPoint<double>* poly4 );
	virtual void initMatrices();
	virtual void fillMatricesCell(
		const YPoint<double>* it, YPoint<double>* resit, const YPoint<int>& p);
	bool hasError( const YPoint<int>& p) const;

	void processPixel( const YPoint<int>& prp);
	double computeZoom( const YPoint<double>* pOutputCoordsPoint) const;
	virtual double estimateZoom( const YPoint<int>& prp);
	virtual bool isFilledPixel( const YPoint<int>& prp);
	virtual bool getAnchor( const YPoint<int>& prp, YPoint<int>& pip);
//	YPoint<int> findNearest( 
//		const YPoint<double>& ip,
//		const YPoint<int>& pipAnchor) const;
	bool findNearest(
		const YPoint<int>& prp,
		const YPoint<double>& ip,
		YPoint<int>& pip,
		int rMax) const;
	bool findCell(
		const YPoint<double>& ip,
		const YPoint<int>& pipNearest,
		YPoint<int>* pip4,
		YPoint<double>* ip4 ) const;
	bool findCellAtRadius(
		const YPoint<double>& ip,
		YPoint<int>* pip4,
		YPoint<double>* ip4,
		int radius) const;
	bool shiftCell(
		int shiftX, int shiftY,
		YPoint<int>* pip4,
		YPoint<double>* ip4 ) const;
	bool interpolateAverage( const YPoint<int>& prp, size_t radius = 0);
	void fillPixelsWithAverages(
		const YPoint<int>& prp,
		std::vector<const DMatrixElem*>::const_iterator begin,
		std::vector<const DMatrixElem*>::const_iterator end,
		int size);
	void interpolateBicubical( 
		const YPoint<int>& prp,
//		const YPoint<int> basePip,
		const YPoint<double> pip);
public:
	// implementation data
	DistributionMatrix distributionMatrix_;
	ITPolygon polygon_;
};

class ImageTransformMatrixFastImpl: public ImageTransformMatrixImpl
{
public:
	virtual void fillMatricesCell(
		const YPoint<double>* it, YPoint<double>* resit, const YPoint<int>& p);
	virtual bool isFilledPixel( const YPoint<int>& prp);
	virtual double estimateZoom( const YPoint<int>& prp);
	virtual bool getAnchor( const YPoint<int>& prp, YPoint<int>& pip);
	virtual void initMatrices();
public:
	CoverageMatrix coverageMatrix_;
};

bool isBadCell( const YPoint<double>* cell );

#endif // _IMPL_IMTRANSF_H
