
#include <math.h>
#include <vector>
#include <iostream>
#include <assert.h>

#include "ygeometr.h"
#include "imtransf.h"
#include "_imtransf.h"

#define IT_EXCEPTION(msg) \
	("Exception in " __FILE__ " at line "  " : " #msg)

// class ImageTransformImpl

ImageTransformImpl::ImageTransformImpl()
: inputCoords_(0), 
  transformFunction_(0), scale_(0),
  outputWidth_(0), outputHeight_(0),
  outputArea_(0,0,0,0),
  flipH_(false),flipV_(false),backgroundColor_(-1),
  inputImageWidth_(0),inputImageHeight_(0),
  edgeH_(0), edgeV_(0), errorColor_(-1),
  enableGridAdjustment_(false)
{
}

void ImageTransformImpl::setBackgroundColor( bool doCorrect, int value)
{
	backgroundColor_ = -1;
	if(!doCorrect)
		return;
	if(value<0)
		return;
	backgroundColor_ = value;
}

void ImageTransformImpl::setErrorColor( bool doCorrect, int value)
{
	errorColor_ = -1;
	if(!doCorrect)
		return;
	if(value<0)
		return;
	errorColor_ = value;
}

void ImageTransformImpl::setFlipHorizontal( bool value)
{
	flipH_ = value;
}

void ImageTransformImpl::setFlipVertical( bool value)
{
	flipV_ = value;
}

void ImageTransformImpl::setGridAdjustment( 
    const YPoint<double>& gridNode)
{
    enableGridAdjustment_ = true;
    gridNode_ = gridNode;
}

int ImageTransformImpl::width() const
{
	return outputWidth_;
}

int ImageTransformImpl::height() const
{
	return outputHeight_;
}

std::vector<int> ImageTransformImpl::bits() const
{
	if(!inputVector_)
		throw IT_EXCEPTION("ImageTransform is not initialized.");
	return inputVector_->bits();
}

std::vector<AImage::YImageClass> ImageTransformImpl::imageClass() const
{
	if(!inputVector_)
		throw IT_EXCEPTION("ImageTransform is not initialized.");
	return inputVector_->imageClass();
}

YPoint<double> ImageTransformImpl::base() const
{
	if(!inputVector_)
		throw IT_EXCEPTION("ImageTransform is not initialized.");
	return PRCStoRCS( YPoint<int>(0,0) );
}

bool ImageTransformImpl::init(
	const ImageProxyVector* vector,
	const YRect<int>& inputFragment,
	const CoordArray2D* coords,
	const ATransformFunction& transformFunction,
	double scale,
	ProgressCounter* indicator,
	const ImageTransform::EdgeMethod& edgeMethod)
{
	/*
	:
	 -    ;
	 -    ;
	 -  ,      ();
	 -     ;
	 -      
	 -     ;
	 -        ();
	*/
	checkInputParameters(
		vector,inputFragment,coords,transformFunction,scale,edgeMethod);
	acceptInputParameters(
		vector,inputFragment,coords,transformFunction,scale,edgeMethod);
//	calculateInputPolygon();
	calculateOutputCoords();
	return calculateOutputData(indicator);
}

void ImageTransformImpl::checkInputParameters(
	const ImageProxyVector* vector,
	const YRect<int>& inputFragment,
	const CoordArray2D* coords,
	const ATransformFunction& transformFunction,
	double scale,
	const ImageTransform::EdgeMethod& edgeMethod) const
{
	if(!vector)
		throw IT_EXCEPTION("No input images");
	if(!coords)
		throw IT_EXCEPTION("No input coords");
	if( scale <= 0 )
		throw IT_EXCEPTION("Scale is invalid");
//	if( vector->width() != coords->width()
//	 || vector->height() != coords->height() )
//		throw IT_EXCEPTION("Image and coords size doesn't match.");
	if( inputFragment.width() != coords->width()
	 || inputFragment.height() != coords->height()
	 || inputFragment.right() > vector->width()
	 || inputFragment.bottom() > vector->height()
	 || inputFragment.top() < 0
	 || inputFragment.left() < 0 )
		throw IT_EXCEPTION("Unacceptable inputFragment.");
}

void ImageTransformImpl::acceptInputParameters(
	const ImageProxyVector* vector,
	const YRect<int>& inputFragment,
	const CoordArray2D* coords,
	const ATransformFunction& transformFunction,
	double scale,
	const ImageTransform::EdgeMethod& edgeMethod)
{
	inputVector_ = vector;
	inputFragment_ = inputFragment;
	inputCoords_ = coords;
	inputWidth_ = coords->width();
	inputHeight_  = coords->height();
	inputImageWidth_ = inputVector_->width();
	inputImageHeight_ = inputVector_->height();
	transformFunction_ = 
		std::auto_ptr<const ATransformFunction>( transformFunction.clone() );
	scale_ = scale;
	// edgeMethod is ignored by now
}

void ImageTransformImpl::calculateOutputCoords()
{
	int w = inputCoords_->width();
	int h = inputCoords_->height();

	const YPoint<double>* it = inputCoords_->begin();
	const YPoint<double>* end = inputCoords_->end();
	outputCoords_.clear();
	outputCoords_.setSize(inputCoords_->width(),inputCoords_->height());
	YPoint<double>* resit = outputCoords_.begin();
	const YPoint<double>* resend = outputCoords_.end();
	while(!canICStoRCS(*it) && it != end)
		++it, ++resit;
	if( it != end )
	{
		*resit = ICStoRCS(*it);
		outputArea_ = YRect<double>(*resit,*resit);
	}
	for( ; it != end; ++it, ++resit)
	{
		if(canICStoRCS(*it))
		{
			*resit = ICStoRCS(*it);
			outputArea_.expand(*resit);
		}
	}

	if(enableGridAdjustment_)
	{
		outputArea_.left() = gridNode_.x + scale_*floor((outputArea_.left()-gridNode_.x)/scale_);
		outputArea_.bottom() = gridNode_.y + scale_*ceil((outputArea_.bottom()-gridNode_.y)/scale_);
	}
	YPoint<double> rtRCS(outputArea_.right()-scale_/2,outputArea_.top()+scale_/2);
	YPoint<int> rtPRCS = RCStoPRCSr(rtRCS);
	outputWidth_ = static_cast<size_t>(rtPRCS.x+1);
	outputHeight_ = static_cast<size_t>(rtPRCS.y+1);
	//outputWidth_ = static_cast<size_t>(int(outputArea_.right()/scale_)-int(outputArea_.left()/scale_)+1);
	//outputHeight_ = static_cast<size_t>(int(outputArea_.bottom()/scale_)-int(outputArea_.top()/scale_)+1);
}

//inline
void ImageTransformImpl::fillOutputPixel( AImageProxy* out, int x, int y, int value)
{
	if(backgroundColor_ != -1 && value == backgroundColor_)
		value = (value==0)?(value+1):(value-1);
	out->set(x,y,value);
}

YPoint<double> ImageTransformImpl::RCStoPRCS( const YPoint<double>& p) const
{
	double x = (p.x - outputArea_.left())/scale_ - 0.5;
	double y = (outputArea_.bottom() - p.y)/scale_ - 0.5;
	return YPoint<double>(x,y);
}

YPoint<int> ImageTransformImpl::RCStoPRCSr( const YPoint<double>& p) const
{
	double x = (p.x - outputArea_.left())/scale_ - 0.5;
	double y = (outputArea_.bottom() - p.y)/scale_ - 0.5;
	return YPoint<int>(int(x+SIGN(0.5,x)),int(y+SIGN(0.5,y)));
//	return YPoint<int>((x>0)?int(x):int(x-1),(y>0)?int(y):int(y-1));
}

YPoint<double> ImageTransformImpl::PRCStoRCS( const YPoint<int>& p) const
{
	double x = outputArea_.left() + p.x*scale_ + scale_/2;
	double y = outputArea_.bottom() - p.y*scale_ - scale_/2;
	return YPoint<double>(x,y);
}

YPoint<double> ImageTransformImpl::PRCStoRCS( const YPoint<double>& p) const
{
	double x = outputArea_.left() + p.x*scale_ + scale_/2;
	double y = outputArea_.bottom() - p.y*scale_ - scale_/2;
	return YPoint<double>(x,y);
}

YPoint<double> ImageTransformImpl::ICStoRCS( const YPoint<double>& p) const
{
	return transformFunction_->forward(p);
}

YPoint<double> ImageTransformImpl::RCStoICS( const YPoint<double>& p) const
{
	return transformFunction_->reverse(p);
}

bool ImageTransformImpl::canICStoRCS( const YPoint<double>& p) const
{
	return transformFunction_->canForward(p);
}

bool ImageTransformImpl::canRCStoICS( const YPoint<double>& p) const
{
	return transformFunction_->canReverse(p);
}

YPoint<double> ImageTransformImpl::PICStoICS( const YPoint<int>& p) const
{
	return inputCoords_->at(p.x,p.y);
}
/*
YPoint<double> ImageTransformImpl::ICStoPICS(
	const YPoint<double>& ip,
	const YPoint<int>* pip4,
	const YPoint<double>* ip4) const
{
	double x = ip.x - 0.5;
	double y = inputCoords_->height() - ip.y - 0.5;
	return YPoint<double>(x,y);
}
*/

void parallelDxDy(
	const YPoint<double>& ip,
	const YPoint<double>& ipO,
	const YPoint<double>& ipX,
	const YPoint<double>& ipY,
	double& dx, double& dy)
{
	double ppx = ::PointOnAxisProjectionLength( 
		ipO.x, ipO.y,
		ipX.y - ipO.y + ipO.x, ipO.x - ipX.x + ipO.y,
		ip.x, ip.y);
	double p01x = ::PointOnAxisProjectionLength( 
		ipO.x, ipO.y,
		ipX.y - ipO.y + ipO.x, ipO.x - ipX.x + ipO.y,
		ipY.x, ipY.y);
	double ppy = ::PointOnAxisProjectionLength( 
		ipO.x, ipO.y,
		ipY.y - ipO.y + ipO.x, ipO.x - ipY.x + ipO.y,
		ip.x, ip.y);
	double p10y = ::PointOnAxisProjectionLength( 
		ipO.x, ipO.y,
		ipY.y - ipO.y + ipO.x, ipO.x - ipY.x + ipO.y,
		ipX.x, ipX.y);

	dx = ABS(ppy / p10y);
	dy = ABS(ppx / p01x);
}


YPoint<double> ImageTransformImpl::ICStoPICS(
	const YPoint<double>& ip,
	const YPoint<int>* pip4,
	const YPoint<double>* ip4) const
{
	int m = 0;
	for( int i = 0; i < 4; ++i)
		if( ip.distance(ip4[i]) < ip.distance(ip4[m]) )
			m = i;

	double dx;
	double dy;
	::parallelDxDy(ip,ip4[m],ip4[3-m],ip4[(5-m)%4],dx,dy);

	dx = (m==0 || m==1)?dx:(-dx);
	dy = (m==1 || m==2)?dy:(-dy);

	return YPoint<double>(pip4[m].x + dx, pip4[m].y + dy);
}

bool ImageTransformImpl::canPICStoICS( const YPoint<int>& p) const
{
	return p.x >= 0
		&& p.x < inputWidth_
		&& p.y >= 0
		&& p.y < inputHeight_;
}

// class ImageTransformCellImpl

bool ImageTransformCellImpl::run( 
	ImageProxyVector* outputVector, 
	const YRect<int>& outputFragment,
	ProgressCounter* indicator)
{
	if(!outputVector)
		throw IT_EXCEPTION("Output images are absent.");
	if( outputVector->countProxy() != inputVector_->countProxy() )
		throw IT_EXCEPTION("Output and input images number is different.");

	outputVector_ = outputVector;
	outputFragment_ = outputFragment;

	YRect<int> outArea = YRect<int>(0,0,outputWidth_, outputHeight_);
	outArea = outArea.intersection(
		YRect<int>(
			-outputFragment.left(),
			-outputFragment.top(),
			outputVector->width()-outputFragment.left(),
			outputVector->height()-outputFragment.top()));

	if(indicator) indicator->start(
		indicator->russian()?" ":"Image transformation",
		0,outputHeight_);

	// main loop
	for( int y = 0; y < inputHeight_-1; ++y)
		for( int x = 0; x < inputWidth_-1; ++x)
		{
			TransformCell cell(*this,x,y);
//			if( cell.estimateZoom() >= 1.6 )
//				interpolateAverage(cell);
//			else
//				interpolateBicubical(cell);
		}

	if(indicator)
		indicator->finish();
	return true;
}

bool ImageTransformCellImpl::calculateOutputData( ProgressCounter* indicator)
{
	processed_.setSize(outputWidth_,outputHeight_);
	return true;
}

// class ImageTransformMatrixImpl

ImageTransformMatrixImpl::ImageTransformMatrixImpl()
: polygon_(0)
{
}

bool ImageTransformMatrixImpl::run( 
	ImageProxyVector* outputVector, 
	const YRect<int>& outputFragment,
	ProgressCounter* indicator)
{
	/*
	:
	*/
	if(!outputVector)
		throw IT_EXCEPTION("Output images are absent.");
	if( outputVector->countProxy() != inputVector_->countProxy() )
		throw IT_EXCEPTION("Output and input images number is different.");
//	if( outputFragment.right() > outputVector->width()
//	 || outputFragment.bottom() > outputVector->height()
//	 || outputFragment.top() < 0
//	 || outputFragment.left() < 0 )
//		throw IT_EXCEPTION("Unacceptable outputFragment.");

	outputVector_ = outputVector;
	outputFragment_ = outputFragment;

	YRect<int> outArea = YRect<int>(0,0,outputWidth_, outputHeight_);
	outArea = outArea.intersection(
		YRect<int>(
			-outputFragment.left(),
			-outputFragment.top(),
			outputVector->width()-outputFragment.left(),
			outputVector->height()-outputFragment.top()));

	if(indicator) indicator->start(
		indicator->russian()?" ":"Image transformation",
		0,outputHeight_);

	for( int y = outArea.top(); y < outArea.bottom(); ++y)
	{
		for( int x = outArea.left(); x < outArea.right(); ++x)
			processPixel( YPoint<int>(x,y) );
		if(indicator)
		{
			if(indicator->current(y))
			{
				indicator->finish();
				return false;
			}
		}
	}

	if(indicator)
		indicator->finish();
	return true;
}

void ImageTransformMatrixImpl::calculateInputPolygon()
{
	polygon_.clear();
	YPoint<int> p(0,0);
	int w = inputCoords_->width();
	int h = inputCoords_->height();
	for( ; p.x < w-1; ++p.x)
		polygon_.push_back( inputCoords_->at(p.x,p.y) );
	for( ; p.y < h-1; ++p.y)
		polygon_.push_back( inputCoords_->at(p.x,p.y) );
	for( ; p.x >= 1; --p.x)
		polygon_.push_back( inputCoords_->at(p.x,p.y) );
	for( ; p.y >= 1; --p.y)
		polygon_.push_back( inputCoords_->at(p.x,p.y) );
}

bool ImageTransformMatrixImpl::hasError( const YPoint<int>& p) const
{
	int x = p.x+inputFragment_.left();
	int y = p.y+inputFragment_.top();
	if( y < 1 || y >= inputVector_->height() - 2 
		|| x < 1 || x >= inputVector_->width() - 2 )
		return false;
//	std::vector<AImageProxy*>::const_iterator it = inputVector_->begin();
//	std::vector<AImageProxy*>::const_iterator end = inputVector_->end();
//	for( ; it != end; ++it 

	// ,    -  ,   
	//  AlcorSat       

	ImageProxyVector::const_iterator it = inputVector_->begin();
	ImageProxyVector::const_iterator end = inputVector_->end();
	bool error[16] = { true, true, true, true, true, true, true, true,
		true, true, true, true, true, true, true, true };
	for(;it != end; ++it)
	{
	    if( (*it)->get(x,y) != errorColor_ )
		error[0] = false;
	    if( (*it)->get(x+1,y) != errorColor_ )
		error[1] = false;
	    if( (*it)->get(x+1,y+1) != errorColor_ )
		error[2] = false;
	    if( (*it)->get(x,y+1) != errorColor_ )
		error[3] = false;

		if( (*it)->get(x-1,y-1) != errorColor_ )
		error[4] = false;
	    if( (*it)->get(x,y-1) != errorColor_ )
		error[5] = false;
	    if( (*it)->get(x+1,y-1) != errorColor_ )
		error[6] = false;
	    if( (*it)->get(x+2,y-1) != errorColor_ )
		error[7] = false;
	    if( (*it)->get(x+2,y) != errorColor_ )
		error[8] = false;
	    if( (*it)->get(x+2,y+1) != errorColor_ )
		error[9] = false;
	    if( (*it)->get(x+2,y+2) != errorColor_ )
		error[10] = false;
	    if( (*it)->get(x+1,y+2) != errorColor_ )
		error[11] = false;
	    if( (*it)->get(x,y+2) != errorColor_ )
		error[12] = false;
	    if( (*it)->get(x-1,y+2) != errorColor_ )
		error[13] = false;
	    if( (*it)->get(x-1,y+1) != errorColor_ )
		error[14] = false;
	    if( (*it)->get(x-1,y) != errorColor_ )
		error[15] = false;
	}
	return error[0] || error[1] || error[2] || error[3] || error[4] || error[5] || error[6] || error[7]
		|| error[8] || error[9] || error[10] || error[11] || error[12] || error[13] || error[14] || error[15];

	//if( (*inputVector_->begin())->get(x,y) == errorColor_
	//	|| (*inputVector_->begin())->get(x+1,y) == errorColor_
	//	|| (*inputVector_->begin())->get(x+1,y+1) == errorColor_
	//	|| (*inputVector_->begin())->get(x,y+1) == errorColor_ )
	//	return true;
	//return false;
}

bool ImageTransformMatrixImpl::calculateOutputData( ProgressCounter* indicator)
{
	/*
	 +    ;
	 +  ,      ();
	 +     ;
	 +      
	 */
	int w = inputCoords_->width();
	int h = inputCoords_->height();

	if(indicator)
		indicator->start("Noname",0,h);

	initMatrices();
	const YPoint<double> *it = inputCoords_->begin();
	YPoint<double> *resit = outputCoords_.begin();
	for(int y = 0; y < h; ++y)
	{
		for( int x = 0; x < w; ++x, ++resit, ++it)
		{
			//   .   , 
			//          
			//     ,   .
			//  ,        
			//      .
			if(errorColor_>=0 && hasError(YPoint<int>(x,y)))
				continue;
			fillMatricesCell(it,resit,YPoint<int>(x,y));
		}
		if( indicator && indicator->current(y) )
		{
			indicator->finish();
			return false;
		}
	}

	if(indicator)
		indicator->finish();
	return true;
}

void ImageTransformMatrixImpl::initMatrices()
{
	distributionMatrix_.clear();
	distributionMatrix_.setSize(outputWidth_,outputHeight_,inputCoords_->width()*inputCoords_->height());
}

void ImageTransformMatrixImpl::fillMatricesCell(
	const YPoint<double>* it, YPoint<double>* resit, const YPoint<int>& p)
{
	if(canICStoRCS(*it))
		distributionMatrix_.add( RCStoPRCSr(*resit), p);
}

double ImageTransformMatrixImpl::computeZoom( const YPoint<double>* pOutputCoordsPoint) const
{
	const YPoint<double>* p = pOutputCoordsPoint;
	YPoint<double> a = RCStoPRCS(*p);
	++p;
	YPoint<double> b = RCStoPRCS(*p);
	p += inputWidth_;
	YPoint<double> c = RCStoPRCS(*p);
	--p;
	YPoint<double> d = RCStoPRCS(*p);
	return 1.0/::QuadrangleArea(a,b,c,d);
}

//YRect<int> ImageTransformMatrixImpl::makeCellRect( 
//	const YPoint<double>* pOutputCoordsPoint)
//{
//	const YPoint<double>* p = pOutputCoordsPoint;
//	YRect<int> result( RCStoPRCSr(*p), RCStoPRCSr(*p) );
//	++p;
//	result.expand( RCStoPRCSr(*p) );
//	p += inputWidth_;
//	result.expand( RCStoPRCSr(*p) );
//	--p;
//	result.expand( RCStoPRCSr(*p) );
//	return result;
//}

YRect<int> ImageTransformMatrixImpl::makeCellRect( 
	const YPoint<double>* pOutputCoordsPoint)
{
	const YPoint<double>* p = pOutputCoordsPoint;
	YPoint<double> prcs( RCStoPRCS(*p) );
	int left = int(floor(prcs.x));
	int right = int(ceil(prcs.x));
	int top = int(floor(prcs.y));
	int bottom = int(ceil(prcs.y));
	++p;
	prcs = RCStoPRCS(*p);
	left = int(MIN(left,floor(prcs.x)));
	right = int(MAX(right,ceil(prcs.x)));
	top = int(MIN(top,floor(prcs.y)));
	bottom = int(MAX(bottom,ceil(prcs.y)));
	p += inputWidth_;
	prcs = RCStoPRCS(*p);
	left = int(MIN(left,floor(prcs.x)));
	right = int(MAX(right,ceil(prcs.x)));
	top = int(MIN(top,floor(prcs.y)));
	bottom = int(MAX(bottom,ceil(prcs.y)));
	--p;
	prcs = RCStoPRCS(*p);
	left = int(MIN(left,floor(prcs.x)));
	right = int(MAX(right,ceil(prcs.x)));
	top = int(MIN(top,floor(prcs.y)));
	bottom = int(MAX(bottom,ceil(prcs.y)));

	YRect<int> result( left, top, right, bottom );
	result = result.intersection( YRect<int>(0,0,outputWidth_-1,outputHeight_-1) );
	return result;
}

void ImageTransformMatrixImpl::makeCellPoly( 
	const YPoint<double>* pOutputCoordsPoint, YPoint<double>* poly4 )
{
	const YPoint<double>* p = pOutputCoordsPoint;
	poly4[0] = RCStoPRCS(*p);
	++p;
	poly4[1] = RCStoPRCS(*p);
	p += inputWidth_;
	poly4[2] = RCStoPRCS(*p);
	--p;
	poly4[3] = RCStoPRCS(*p);
}

void ImageTransformMatrixImpl::processPixel( const YPoint<int>& prp)
{
	if(!isFilledPixel(prp))
		return;
	if( estimateZoom(prp) <= 0.001 )
		return;
	if( estimateZoom(prp) >= 1.6 )
	{
		if( !distributionMatrix_.empty(prp) )
		{
			if(interpolateAverage(prp))
				return;
		}
		else
		{
			if(interpolateAverage(prp,1))
				return;
		}
		// fall back to bicubic - seems like the function has big distortions here
	}
	// interpolation method here (bicubical)
	YPoint<double> rp = PRCStoRCS(prp);
	if(!canRCStoICS(rp))
		return;

	YPoint<double> ip = RCStoICS(rp);

	YPoint<int> pipAnchor;
	getAnchor(prp,pipAnchor);

	YPoint<int> pip4[4];
	YPoint<double> ip4[4];

	if( !findCell( ip, pipAnchor, pip4, ip4 ) )
		return;

	YPoint<double> pip = ICStoPICS( ip, pip4, ip4);
	// bicubical interpolation
	interpolateBicubical(prp,pip);
}

bool ImageTransformMatrixImpl::isFilledPixel( const YPoint<int>& prp)
{
	return true;
}

bool ImageTransformMatrixImpl::getAnchor( const YPoint<int>& prp, YPoint<int>& pip)
{
	return findNearest(prp,RCStoICS(PRCStoRCS(prp)),pip,MaxFindRadius);
}

double ImageTransformMatrixImpl::estimateZoom( const YPoint<int>& prp)
{
	int n = 0;
	int sum = 0;
	sum += distributionMatrix_.collectRaduis(prp,0,&n).size();
	sum += distributionMatrix_.collectRaduis(prp,1,&n).size();
	return double(sum)/(n?n:1);
}

const int directionSequenceX[] =
{-1,+2, 0,-2, 0,+1,+1,-1 }; //, 0,+1,+1, 0, 0, 0, 0,-1,-1,-1,-1, 0, 0, 0, 0,+1,+1};
const int directionSequenceY[] =
{-1, 0,+2, 0,-1,-1,+1,+1 }; //,+1, 0, 0,-1,-1,-1,-1, 0, 0, 0, 0,+1,+1,+1,+1, 0, 0};
const size_t directionSequenceSize = sizeof(directionSequenceX)/sizeof(directionSequenceX[0]);

/*
YPoint<int> ImageTransformMatrixImpl::findNearest( 
	const YPoint<double>& ip,
	const YPoint<int>& pipAnchor) const
{

	YPoint<int> candidate = pipAnchor;
	YPoint<int> result;
	const int* dirextionShiftX;
	const int* dirextionShiftY;
	const int* directionSequenceEnd = &directionSequenceX[directionSequenceSize-1] + 1;
	double distance = ip.distance( PICStoICS(pipAnchor) );
	double distanceTmp;
	while(true)
	{
		dirextionShiftX = directionSequenceX;
		dirextionShiftY = directionSequenceY;
		result = candidate;
		for( size_t i = 0; i < directionSequenceSize; ++i, ++dirextionShiftX, ++dirextionShiftY )
		{
			candidate.x += *dirextionShiftX;
			candidate.y += *dirextionShiftY;
			if( canPICStoICS(candidate) )
			{
				distanceTmp = ip.distance( PICStoICS(candidate) );
                if( distance > distanceTmp )
				{ distance = distanceTmp; break; }
			}
		}
		if( directionSequenceEnd == dirextionShiftX )
			return result;
	}
}
*/
bool ImageTransformMatrixImpl::findNearest( 
	const YPoint<int>& prp,
	const YPoint<double>& ip,
	YPoint<int>& pip,
	int rMax) const
{
	std::vector<const DMatrixElem*> elems;
	for( int r = 0; r <= rMax; ++r)
	{
		elems = distributionMatrix_.collectRaduis(prp,r);
		if(!elems.empty())
			break;
	}
	if(elems.empty())
		return false;
	YPoint<int> pipN = elems[0]->p;
	YPoint<double> ipN = PICStoICS(pipN);
	std::vector<const DMatrixElem*>::iterator it = elems.begin();
	std::vector<const DMatrixElem*>::iterator end = elems.end();
	for( ; it != end; ++it )
	{
		YPoint<int> pip_tmp = (*it)->p;
		YPoint<double> ip_tmp = PICStoICS(pip_tmp);
		if( ip.distance(ipN) > ip.distance(ip_tmp) )
		{
			pipN = pip_tmp;
			ipN = ip_tmp;
		}
	}
	pip = pipN;
	return true;
}

bool ImageTransformMatrixImpl::shiftCell(
	int shiftX, int shiftY,
	YPoint<int>* pip4,
	YPoint<double>* ip4 ) const
{
	bool success = true;
	for( int i = 0; i < 4; ++i)
	{
		pip4[i].x += shiftX;
		pip4[i].y += shiftY;
		if(!canPICStoICS(pip4[i]))
			success = false;
		else
			ip4[i] = PICStoICS(pip4[i]);
	}
	return success;
}

bool ImageTransformMatrixImpl::findCellAtRadius(
	const YPoint<double>& ip,
	YPoint<int>* pip4,
	YPoint<double>* ip4,
	int radius) const
{
	int d = radius*2;
	if( shiftCell( +1, +1, pip4, ip4) && PolyFun::isPointInPolygon(ip,ip4,4) )
		return true;
	for( int i = d; i != 0; --i )
		if( shiftCell( -1, 0, pip4, ip4) && PolyFun::isPointInPolygon(ip,ip4,4) )
			return true;
	for( int i = d; i != 0; --i )
		if( shiftCell( 0, -1, pip4, ip4) && PolyFun::isPointInPolygon(ip,ip4,4) )
			return true;
	for( int i = d; i != 0; --i )
		if( shiftCell( 1, 0, pip4, ip4) && PolyFun::isPointInPolygon(ip,ip4,4) )
			return true;
	for( int i = d; i != 0; --i )
		if( shiftCell( 0, 1, pip4, ip4) && PolyFun::isPointInPolygon(ip,ip4,4) )
			return true;
	return false;
}

bool ImageTransformMatrixImpl::findCell(
	const YPoint<double>& ip,
	const YPoint<int>& pipNearest,
	YPoint<int>* pip4,
	YPoint<double>* ip4 ) const
{
	//pip4[0] = pipNearest;
	//pip4[1] = YPoint<int>(pipNearest.x,pipNearest.y-1);
	//pip4[2] = YPoint<int>(pipNearest.x+1,pipNearest.y-1);
	//pip4[3] = YPoint<int>(pipNearest.x+1,pipNearest.y);
	// -      pip[0]
	//     
	// pip[0]      
	pip4[0] = YPoint<int>(pipNearest.x,pipNearest.y+1);
	pip4[1] = YPoint<int>(pipNearest.x,pipNearest.y);
	pip4[2] = YPoint<int>(pipNearest.x+1,pipNearest.y);
	pip4[3] = YPoint<int>(pipNearest.x+1,pipNearest.y+1);
	shiftCell( 0, 0, pip4, ip4 );
//	if( shiftCell( 0, 0, pip4, ip4) && PolyFun::isPointInPolygon(ip,ip4,4) )
		return true;

//	for( int i = 1; i < 10; ++i )
//	if( findCellAtRadius(ip,pip4,ip4,i) )
//		return true;

//	return false;
}

int ITAverageProcessImage(
	const AImageProxy* in,
	std::vector<const DMatrixElem*>::const_iterator begin,
	std::vector<const DMatrixElem*>::const_iterator end,
	int size,
	int shiftX, int shiftY,
	int w, int h,
	bool fH, bool fV)
{
	double sum = 0;
	for( ; begin != end; ++begin)
	{
		int x = (*begin)->p.x + shiftX;
		if(fH)
			x = w - x - 1;
		int y = (*begin)->p.y + shiftY;
		if(fV)
			y = h - y - 1;
		sum += in->get(x,y);
	}

//	(*out)[resY][resX] = int((sum+size/2)/size);
	return int((sum+size/2)/size);
}

void ImageTransformMatrixImpl::fillPixelsWithAverages(
	const YPoint<int>& prp,
	std::vector<const DMatrixElem*>::const_iterator begin,
	std::vector<const DMatrixElem*>::const_iterator end,
	int size)
{
	ImageProxyVector::const_iterator it = inputVector_->begin();
	ImageProxyVector::const_iterator end_it = inputVector_->end();
	ImageProxyVector::iterator out = outputVector_->begin();
	for( ; it != end_it; ++it, ++out)
		fillOutputPixel(
			*out,
			prp.x+outputFragment_.left(),
			prp.y+outputFragment_.top(),
			::ITAverageProcessImage(
				*it,begin,end,size,
				inputFragment_.left(),inputFragment_.top(),
				inputImageWidth_,inputImageHeight_,
				flipH_,flipV_));
}

bool ImageTransformMatrixImpl::interpolateAverage( const YPoint<int>& prp, size_t radius)
{
	std::vector<const DMatrixElem*> elems = distributionMatrix_.collectRaduis(prp,0);
	for( size_t i = 1; i <= radius; ++i )
	{
		std::vector<const DMatrixElem*> elemsAtRadius =
			distributionMatrix_.collectRaduis(prp,i);
		elems.insert(elems.end(),elemsAtRadius.begin(),elemsAtRadius.end());
	}

	std::vector<const DMatrixElem*>::const_iterator begin =
		elems.begin();
	std::vector<const DMatrixElem*>::const_iterator end =
		elems.end();
	int size = elems.size();

	if(!size)
		return false;
	
	fillPixelsWithAverages(prp,begin,end,size);
	return true;
}

int ITBicubicalProcessImage(
	const AImageProxy* in,
	const double* P,
	const double* L,
	int x, int y,
	int dx, int dy)
{
	double res = 0;
	for( int i = 0; i < 4; i++)
	{
		double rtmp = 0;
		for( int j = 0; j < 4; j++)
			rtmp += in->get(x+j*dx,y+i*dy) * P[j];
		res += rtmp * L[i];
	}

//	(*out)[resY][resX] = IN_BOUNDS(int(res/36.0+0.5),0,max);
	return IN_BOUNDS(int(res/36.0+0.5),0,in->maxValue());
}	

void ImageTransformMatrixImpl::interpolateBicubical(
	const YPoint<int>& prp,
	const YPoint<double> pip)
{
	YPoint<int> basePip(int(pip.x),int(pip.y));
	if( basePip.x <= 0 )
		basePip.x = 1;
	if( basePip.x >= inputCoords_->width() - 2 )
		basePip.x = inputCoords_->width() - 3;
	if( basePip.y <= 0 )
		basePip.y = 1;
	if( basePip.y >= inputCoords_->height() - 2 )
		basePip.y = inputCoords_->height() - 3;

	double dx = pip.x - basePip.x + 1;
	double dy = pip.y - basePip.y + 1;
	
	double P[4];
	P[0] = ((-1.0*dx +  6.0)*dx - 11.0)*dx + 6.0;
	P[1] = (( 3.0*dx - 15.0)*dx + 18.0)*dx;
	P[2] = ((-3.0*dx + 12.0)*dx -  9.0)*dx;
	P[3] = ((     dx -  3.0)*dx +  2.0)*dx;

	double L[4];
	L[0] = ((-1.0*dy +  6.0)*dy - 11.0)*dy + 6.0;
	L[1] = (( 3.0*dy - 15.0)*dy + 18.0)*dy;
	L[2] = ((-3.0*dy + 12.0)*dy -  9.0)*dy;
	L[3] = ((     dy -  3.0)*dy +  2.0)*dy;

	int x = basePip.x-1+inputFragment_.left();
	int fx = 1;
	if(flipH_)
	{
		x = inputImageWidth_ - x - 1;
		fx = -1;
	}
	int y = basePip.y-1+inputFragment_.top();
	int fy = 1;
	if(flipV_)
	{
		y = inputImageHeight_ - y - 1;
		fy = -1;
	}

	ImageProxyVector::const_iterator it = inputVector_->begin();
	ImageProxyVector::const_iterator end = inputVector_->end();
	ImageProxyVector::iterator out = outputVector_->begin();
	for( ; it != end; ++it, ++out)
		fillOutputPixel(
			*out,
			prp.x+outputFragment_.left(),
			prp.y+outputFragment_.top(),
			::ITBicubicalProcessImage(
				*it,P,L,x,y,fx,fy));
}

// class ImageTransformMatrixFastImpl

void ImageTransformMatrixFastImpl::fillMatricesCell(
	const YPoint<double>* it, YPoint<double>* resit, const YPoint<int>& p)
{
	ImageTransformMatrixImpl::fillMatricesCell(it,resit,p);

	int w = inputCoords_->width();
	int h = inputCoords_->height();
	if( p.x >= w-1 || p.y >= h-1 )
		return;

	if( !canICStoRCS(*it)
	 || !canICStoRCS(*(it+1))
	 || !canICStoRCS(*(it+w))
	 || !canICStoRCS(*(it+w+1)) )
		return;

	double zoom = computeZoom(resit);
	YPoint<double> cell[4];
	makeCellPoly(resit,cell);
	if( ::isBadCell(cell) )
		return;

	YRect<int> cellArea = makeCellRect(resit);
	coverageMatrix_.set( cell, cellArea, p, zoom);
}

bool ImageTransformMatrixFastImpl::isFilledPixel( const YPoint<int>& prp)
{
	if(coverageMatrix_.empty(prp))
		return false;
	return true;
}

double ImageTransformMatrixFastImpl::estimateZoom( const YPoint<int>& prp)
{
	return coverageMatrix_.getZoom(prp);
}

bool ImageTransformMatrixFastImpl::getAnchor( const YPoint<int>& prp, YPoint<int>& pip)
{
	pip = coverageMatrix_.get(prp);
	return true;
}

void ImageTransformMatrixFastImpl::initMatrices()
{
	ImageTransformMatrixImpl::initMatrices();
	coverageMatrix_.clear();
	coverageMatrix_.setSize(outputWidth_,outputHeight_);
}


// class ImageTransform

//explicit
ImageTransform::ImageTransform( bool fastAlgorithm )
: pImpl_( fastAlgorithm?(new ImageTransformMatrixFastImpl):(new ImageTransformMatrixImpl) )
{
//: pImpl_( new ImageTransformCellImpl )
}

//virtual
ImageTransform::~ImageTransform()
{
	delete pImpl_;
}

void ImageTransform::setBackgroundColor( bool doCorrect, int value)
{
	pImpl_->setBackgroundColor(doCorrect,value);
}

void ImageTransform::setErrorColor( bool doCorrect, int value)
{
	pImpl_->setErrorColor(doCorrect,value);
}

void ImageTransform::setFlipHorizontal( bool value )
{
	pImpl_->setFlipHorizontal(value);
}

void ImageTransform::setFlipVertical( bool value )
{
	pImpl_->setFlipVertical(value);
}

void ImageTransform::setGridAdjustment( const YPoint<double>& gridNode)
{
	pImpl_->setGridAdjustment(gridNode);
}

/*
void ImageTransform::setHorizontalEdgesMethod( bool skip)
{
	pImpl_->edgeH_ = !skip;
}

void ImageTransform::setVerticalEdgesMethod( bool skip)
{
	pImpl_->edgeV_ = !skip;
}
*/
int ImageTransform::width() const
{
	return pImpl_->width();
}

int ImageTransform::height() const
{
	return pImpl_->height();
}

std::vector<int> ImageTransform::bits() const
{
	return pImpl_->bits();
}

std::vector<AImage::YImageClass> ImageTransform::imageClass() const
{
	return pImpl_->imageClass();
}

YPoint<double> ImageTransform::base() const
{
	return pImpl_->base();
}

bool ImageTransform::init(
	const ImageProxyVector* vector,
	const YRect<int>& inputFragment,
	const CoordArray2D* coords,
	const ATransformFunction& transformFunction,
	double scale,
	ProgressCounter* indicator,
	const EdgeMethod& edgeMethod)
{
	return pImpl_->init(
		vector,
		inputFragment,
		coords,
		transformFunction,
		scale,
		indicator,
		edgeMethod);
}

bool ImageTransform::init(
	const ImageProxyVector* vector,
	const CoordArray2D* coords,
	const ATransformFunction& transformFunction,
	double scale,
	ProgressCounter* indicator,
	const EdgeMethod& edgeMethod)
{
	return init(
		vector,
		YRect<int>(0,0,vector->width(),vector->height()),
		coords,
		transformFunction,
		scale,
		indicator,
		edgeMethod);
}

bool ImageTransform::run( 
	ImageProxyVector* outputVector, 
	const YRect<int>& outputFragment,
	ProgressCounter* indicator)
{
	return pImpl_->run(
		outputVector,
		outputFragment,
		indicator);
}

bool ImageTransform::run( ImageProxyVector* outputVector, ProgressCounter* indicator)
{
	return run(
		outputVector,
		YRect<int>(0,0,outputVector->width(),outputVector->height()),
		indicator);
}

// class DistributionMatrix

std::vector<const DMatrixElem*> DistributionMatrix::collectRaduis(
		const YPoint<int>&prp, int r, int* pn) const
{
	assert( prp.x >= 0 && prp.x < w_ && prp.y >= 0 && prp.y < h_ );
	std::vector<const DMatrixElem*> elems;
	YRect<int> edges;
	edges.top() = MAX( prp.y - r, 0 );
	edges.left() = MAX( prp.x - r, 0 );
	edges.bottom() = MIN( prp.y + r, h_ - 1 );
	edges.right() = MIN( prp.x + r, w_ - 1 );
	assert(
		   edges.BottomRight.x >= 0 
		&& edges.BottomRight.x < w_ 
		&& edges.BottomRight.y >= 0
		&& edges.BottomRight.y < h_ );
	assert(
		   edges.TopLeft.x >= 0 
		&& edges.TopLeft.x < w_ 
		&& edges.TopLeft.y >= 0
		&& edges.TopLeft.y < h_ );
	assert( prp.x >= 0 && prp.x < w_ && prp.y >= 0 && prp.y < h_ );
	int dx = edges.right() - edges.left();
	int dy = edges.bottom() - edges.top();
	DMatrixElem* const* pp = at( YPoint<int>(edges.left(),edges.top()) );
	// optimization
	if(pn)
	{
		if( dx*dy == 0 )
		{
			collectCell(elems,*pp);
			++(*pn);
			return elems;
		}
		for( int n = dx; n != 0; --n, ++pp)
		{
			collectCell(elems,*pp);
			++(*pn);
		}
		for( int n = dy; n != 0; --n, pp+=w_)
		{
			collectCell(elems,*pp);
			++(*pn);
		}
		for( int n = dx; n != 0; --n, --pp)
		{
			collectCell(elems,*pp);
			++(*pn);
		}
		for( int n = dy; n != 0; --n, pp-=w_)
		{
			collectCell(elems,*pp);
			++(*pn);
		}
	}
	else
	{
		if( dx*dy == 0 )
		{
			collectCell(elems,*pp);
			return elems;
		}
		for( int n = dx; n != 0; --n, ++pp)
			if(*pp) collectCell(elems,*pp);
		for( int n = dy; n != 0; --n, pp+=w_)
			if(*pp) collectCell(elems,*pp);
		for( int n = dx; n != 0; --n, --pp)
			if(*pp) collectCell(elems,*pp);
		for( int n = dy; n != 0; --n, pp-=w_)
			if(*pp) collectCell(elems,*pp);
	}
	return elems;
}

void DistributionMatrix::collectCell(
	std::vector<const DMatrixElem*>& elems, const DMatrixElem* elem) const
{
	const DMatrixElem* p = elem;
	while(p)
	{
		elems.push_back(p);
		p = p->next;
	}
}

// class CoverageMatrix

void CoverageMatrix::set( 
	const YPoint<double>* cell,
	const YRect<int>& area, 
	const YPoint<int>& value,
	double zoomValue)
{
	int bottom = area.bottom();
	int right = area.right();
	for( int y = area.top(); y <= bottom; ++y)
		for( int x = area.left(); x <= right; ++x)
			if( /*empty(YPoint<int>(x,y))
			 || */PolyFun::isPointInPolygon( YPoint<double>(x,y), cell, 4) )
				set( YPoint<int>(x,y), value, zoomValue);
}

// utility functions

void AImageSetColor( AImage* image, int r, int g, int b)
{
	// AImageMem8 case
	AImageMem8* image8 = dynamic_cast<AImageMem8*>(image);
	if( image8 )
		image8->SetTo(r);
	// AImageMem16 case
	AImageMem16* image16 = dynamic_cast<AImageMem16*>(image);
	if( image16 )
		image16->SetTo(r);
	// AImageMemRgb8 case
	AImageMemRgb8* imageRgb8 = dynamic_cast<AImageMemRgb8*>(image);
	if( imageRgb8 )
	{
		AImage::Rgb color = { r,g,b };
		imageRgb8->SetTo(color);
	}
}

void AImageSetNull( AImage* image)
{
	::AImageSetColor(image);
}

bool isBadCell( const YPoint<double>* cell )
{
	if( LineFun::areIntersecting( cell[0], cell[1], cell[2], cell[3] )
	 || LineFun::areIntersecting( cell[0], cell[3], cell[1], cell[2] ) )
		return true;
	else
		return false;
}

// class ImageProxyVector

ImageProxyVector::ImageProxyVector( 
	const ImageProxyVector& src)
	: bits_(src.bits_), imageClass_(src.imageClass_),
	width_(src.width_), height_(src.height_)
{
	const_iterator it = src.begin();
	const_iterator end = src.end();
	for(; it != end; ++it)
		proxies_.push_back( (*it)->clone() );
}

ImageProxyVector& ImageProxyVector::operator=( 
	const ImageProxyVector& src)
{
	if( this == &src )
		return *this;
	bits_ = src.bits_;
	imageClass_ = src.imageClass_;
	height_ = src.height_;
	width_ = src.width_;
	const_iterator it = src.begin();
	const_iterator end = src.end();
	for(; it != end; ++it)
		proxies_.push_back( (*it)->clone() );
	return *this;
}

//virtual
ImageProxyVector::~ImageProxyVector()
{
	iterator it = proxies_.begin();
	iterator end = proxies_.end();
	for( ; it != end; ++it)
		delete *it;
}

void ImageProxyVector::add( AImage* image)
{
	for( int i = 0; i < image->NumPlanes(); ++i)
	{
		if( AImageMem8* bit8 = dynamic_cast<AImageMem8*>(image->GetPlane(i)) )
			add( ImageProxy<AImageMem8>(bit8) );
		else if( AImageMem16* bit16 = dynamic_cast<AImageMem16*>(image->GetPlane(i)) )
			add( ImageProxy<AImageMem16>(bit16) );
	}
	bits_.push_back(image->Bits());
	imageClass_.push_back(image->ImageClass());
}

void ImageProxyVector::add( const AImage* image)
{
	for( int i = 0; i < image->NumPlanes(); ++i)
	{
		if( const AImageMem8* bit8 = 
			dynamic_cast<const AImageMem8*>(image->GetPlane(i)) )
			add( ImageProxyConst<AImageMem8>(bit8) );
		else if( const AImageMem16* bit16 = 
			dynamic_cast<const AImageMem16*>(image->GetPlane(i)) )
			add( ImageProxyConst<AImageMem16>(bit16) );
	}
	bits_.push_back(image->Bits());
	imageClass_.push_back(image->ImageClass());
}

void ImageProxyVector::add( const AImageProxy& imageProxy)
{
	if(bits_.empty())
	{
		width_ = imageProxy.width();
		height_ = imageProxy.height();
	}
	if( imageProxy.width() != width_
	 || imageProxy.height() != height_ )
		throw "ImageProxyVector::add(): Images should have same size.";
	proxies_.push_back(imageProxy.clone());
}

bool operator==( const CoordArray2D& lhs, const CoordArray2D& rhs)
{
	if( lhs.width() != rhs.width() || lhs.height() != rhs.height() )
		return false;
	return std::equal( lhs.begin(), lhs.end(), rhs.begin() );
}

bool operator!=( const CoordArray2D& lhs, const CoordArray2D& rhs)
{
	return !(lhs == rhs);
}

std::ostream& operator<<( std::ostream& os, const CoordArray2D& src)
{
	os << "CoordsArray2D ( " << 
		"w = " << src.width() << 
		", h = " << src.height() << " )";
	return os;
}

// class TransformCell

TransformCell::TransformCell( 
	ImageTransformCellImpl& transform,
	int x, int y)
	: transform_(transform)
{
	if( x < 0 || x >= transform.inputWidth_ - 1
		|| y < 0 || y >= transform.inputHeight_ - 1 )
		throw std::out_of_range("invalid point");
	fill(x,y);
}

const YPoint<double>* TransformCell::getCellPICS() const
{
	return pics_;
}

const YPoint<double>* TransformCell::getCellICS() const
{
	return ics_;
}

const YPoint<double>* TransformCell::getCellRCS() const
{
	return rcs_;
}

const YPoint<double>* TransformCell::getCellPRCS() const
{
	return prcs_;
}

void TransformCell::fill(int x0, int y0)
{
	int x = x0;
	int y = y0;
	int sx[] = { 0, 1, 0, -1 }; 
	int sy[] = { 0, 0, 1,  0 }; 
	for( int i = 0; i < 4; ++i )
	{
		x += sx[i];
		y += sy[i];
		pics_[i] = YPoint<double>(x,y);
		ics_[i] = transform_.PICStoICS(YPoint<int>(x,y));
		rcs_[i] = transform_.ICStoRCS(ics_[i]);
		prcs_[i] = transform_.RCStoPRCS(rcs_[i]);
	}
}

#if 0
namespace nr {

#define TINY 1.0e-20
void ludcmp(double **a, int n, int *indx, double *d)
{
	int i,imax,j,k;
	double big,dum,sum,temp;
	double *vv;

//	vv=vector(1,n);
	*d=1.0;
	for (i=1;i<=n;i++) {
		big=0.0;
		for (j=1;j<=n;j++)
			if ((temp=fabs(a[i][j])) > big) big=temp;
		if (big == 0.0) throw std::out_of_range("Singular matrix in routine ludcmp");
		vv[i]=1.0/big;
	}
	for (j=1;j<=n;j++) {
		for (i=1;i<j;i++) {
			sum=a[i][j];
			for (k=1;k<i;k++) sum -= a[i][k]*a[k][j];
			a[i][j]=sum;
		}
		big=0.0;
		for (i=j;i<=n;i++) {
			sum=a[i][j];
			for (k=1;k<j;k++)
				sum -= a[i][k]*a[k][j];
			a[i][j]=sum;
			if ( (dum=vv[i]*fabs(sum)) >= big) {
				big=dum;
				imax=i;
			}
		}
		if (j != imax) {
			for (k=1;k<=n;k++) {
				dum=a[imax][k];
				a[imax][k]=a[j][k];
				a[j][k]=dum;
			}
			*d = -(*d);
			vv[imax]=vv[j];
		}
		indx[j]=imax;
		if (a[j][j] == 0.0) a[j][j]=TINY;
		if (j != n) {
			dum=1.0/(a[j][j]);
			for (i=j+1;i<=n;i++) a[i][j] *= dum;
		}
	}
//	free_vector(vv,1,n);
}
#undef TINY

void lubksb(double **a, int n, int *indx, double *b)
{
	int i,ii=0,ip,j;
	double sum;

	for (i=1;i<=n;i++) {
		ip=indx[i];
		sum=b[ip];
		b[ip]=b[i];
		if (ii)
			for (j=ii;j<=i-1;j++) sum -= a[i][j]*b[j];
		else if (sum) ii=i;
		b[i]=sum;
	}
	for (i=n;i>=1;i--) {
		sum=b[i];
		for (j=i+1;j<=n;j++) sum -= a[i][j]*b[j];
		b[i]=sum/a[i][i];
	}
}

} // namespace nr
#endif //0