#ifndef _PLF_FILTERS_H
#define _PLF_FILTERS_H
/////////////////////////////////////////////////////////////////////////////
#include <algorithm>

#include "plf_kernels_1d.h"

#include "plf_filter_elem.h"
#include "plf_io_filter.h"
#include "plf_iterators.h"

namespace Vtl
{
/// @brief  Pipe Line Filtering (of images) library
namespace Plf 
{
/** @addtogroup ElementaryFiltersDoxGroup  Elementary filters
 *  @ingroup PipeLineFiltersDoxGroup  
 *
 *  @brief         
 *     (Pipe Line Filtering) 
 *
 *  @{
 */

	///////////////////////////////////////////////////////////////////////////
	// class Levels
	/// @brief     .
	///        
	///   .       ,
	///          .
	///     -      
	///     \b  \b  ,    
	///    ,     (   ).
	///      Levels   -   
	/// ,       ,   
	///    -      
	///       Vtl::Plf::F::Comb .
	///////////////////////////////////////////////////////////////////////////
	template<typename T,class P=Vtl::Plf::F11::DoNothing<T> >
	class Levels
		: public FilterElementary
	{
	public:
		InPort<T> in;
		OutPort<T,P> out;
		int tmp;
	public:
		Levels()
			: FilterElementary(1,1)
		{
		}
		Levels(std::string name,P post_proc=Vtl::Plf::F11::DoNothing<T>())
			: FilterElementary(1,1)
		{
			Init(name,post_proc);
		}
		virtual ~Levels()
		{
		}
		void Init(std::string name,P post_proc=Vtl::Plf::F11::DoNothing<T>())
		{
			setName(name);
			out.setPostProc(post_proc);
			AddPort(in,0,Extent(0,0,0,0));
			AddPort(out,0);
			tmp=0;
		}
		virtual void processLine() 
		{
			ImgSize ims=in.getImageSize();
			const int w=ims.width;
			out.beginUpdate();
				ImageRef<T> img=in.getImg(); 
				const T* input=img[0];
				OutPortNoPostProc<T>::BufType::Iter_Line output=out.begin();
				///Simply copy and allow modify for post processing in OutPort<,>
				for(int x=0; x<w; x++)
					*output++=*input++;
			in.ToNextLine(Y_pos()+1);
			out.endUpdate();
//			printf("-G---- %d\n",tmp++);
//			printf("-G---- %d %d %d %d\n",out.buffer.ext().left,out.buffer.ext().right,out.buffer.ext().before,out.buffer.ext().forward);
		}
	};
	///////////////////////////////////////////////////////////////////////////
	// class FilterSqrtOfSumSqr
	/// @brief    ,    
	///         , 
	///      (x,y).
	///
	///       -   ...
	///    
	///////////////////////////////////////////////////////////////////////////
	template<typename T,class F,class P=Vtl::Plf::F11::DoNothing<T> >
	class PixByPix2to1
		: public FilterElementary
	{
	public:
		InPort<T> in_1;
		InPort<T> in_2;
		OutPort<T,P> out;
		F functor;
		int tmp;
	public:
		PixByPix2to1()
			: FilterElementary(2,1)
		{
		}
		PixByPix2to1(std::string name,F fun,P post_proc=Vtl::Plf::F11::DoNothing<T>())
			: FilterElementary(2,1)
		{
			Init(name,fun,post_proc);
		}
		void Init(std::string name,F fun,P post_proc=Vtl::Plf::F11::DoNothing<T>())
		{
			setName(name);
			out.setPostProc(post_proc);
			functor=fun;
			AddPort(in_1,0,Extent(0,0,0,0));
			AddPort(in_2,1,Extent(0,0,0,0));
			AddPort(out,0);
			tmp=0;
		}
		virtual ~PixByPix2to1()
		{
		}
		virtual void processLine() 
		{

			ImgSize ims=in_1.getImageSize();
			const int w=ims.width;

			out.beginUpdate();

//			printf("PixByPix2to1: out.Y() | in_1.Y(), in_2.Y()=( %d | %d, %d )\n",out.Y_wait(),in_1.Y(),in_2.Y());
				ImageRef<T> img_1=in_1.getImg(); 
				ImageRef<T> img_2=in_2.getImg(); 
				const T* p1=img_1[0];
				const T* p2=img_2[0];
				OutPortNoPostProc<T>::BufType::Iter_Line output=out.begin();
				for(int x=0; x<w; x++)
					*output++=functor(*p1++,*p2++);
			in_1.ToNextLine(Y_pos()+1);
			in_2.ToNextLine(Y_pos()+1);
			out.endUpdate();
//			printf("-SQR-- %d\n",tmp++);
//			printf("-G---- %d %d %d %d\n",out.buffer.ext().left,out.buffer.ext().right,out.buffer.ext().before,out.buffer.ext().forward);
		}
	};
	///////////////////////////////////////////////////////////////////////////
	// class PixByPix3to1
	/// @brief        ,
	///          
	///    .    
	///  .  :
	/// @li  TrueColor    @see Ppf::C.
	/// @li      -, , DiZenzo,
	///              
	///      \f$ <g_x^2>, <g_x \cdot g_y>, <g_y^2>  \f$. 
	///     @see Ppf::F, Harris, Trace, MaxEigenVal, MinEigenVal, DiffEigenVal.
	/// 
	///    
	///////////////////////////////////////////////////////////////////////////
	template<typename T,class Func=Vtl::Plf::F::Harris<T>,class P=Vtl::Plf::F11::DoNothing<T> >
	class PixByPix3to1
	: public FilterElementary
	{
	public:
		InPort<T> in_xx;
		InPort<T> in_xy;
		InPort<T> in_yy;
		OutPort<T,P> out;
		Func func;
	public:
		PixByPix3to1()
			: FilterElementary(3,1)
		{
		}
		PixByPix3to1(std::string name,Func functor,P post_proc=Vtl::Plf::F11::DoNothing<T>())
			: FilterElementary(3,1)
		{
			Init(name,functor,post_proc);
		}
		virtual ~PixByPix3to1()
		{
		}
		void Init(std::string name,Func functor,P post_proc=Vtl::Plf::F11::DoNothing<T>())
		{
			setName(name);
			func=functor;
			out.setPostProc(post_proc);
			
			AddPort(in_xx,0,Extent(0,0,0,0));
			AddPort(in_xy,1,Extent(0,0,0,0));
			AddPort(in_yy,2,Extent(0,0,0,0));
			
			AddPort(out,0);
		}
		virtual void set_OutPortsImagesSizeAndInPortsExtents()
		{
			ImgSize ims_xx=in_xx.getImageSize();
			ImgSize ims_xy=in_xy.getImageSize();
			ImgSize ims_yy=in_yy.getImageSize();
			if(ims_xx!=ims_xy ||ims_xx!=ims_yy )
			{
				std::ostringstream os;
				os<< "Filter Vtl::Plf::CorrelationMatrix2x2 \"" << Name() <<"\" : input images are of different size."
					<< " in_xx : " << ims_xx << " in_xy : " << ims_xy << " in_yy : " << ims_yy;
				throw os.str();
			}
			FilterElementary::set_OutPortsImagesSizeAndInPortsExtents();
		}
		virtual void processLine() 
		{
			ImgSize ims=in_xx.getImageSize();
			const int w=ims.width;

			out.beginUpdate();
				ImageRef<T> img_xx=in_xx.getImg(); 
				ImageRef<T> img_xy=in_xy.getImg(); 
				ImageRef<T> img_yy=in_yy.getImg(); 

				const T* p_xx=img_xx[0];
				const T* p_xy=img_xy[0];
				const T* p_yy=img_yy[0];
				
				OutPortNoPostProc<T>::BufType::Iter_Line output=out.begin();
				
				for(int x=0; x<w; x++)
				{
					*output++ = func(*p_xx++,*p_xy++,*p_yy++);
				}
				in_xx.ToNextLine(Y_pos()+1);
				in_xy.ToNextLine(Y_pos()+1);
				in_yy.ToNextLine(Y_pos()+1);
			out.endUpdate();
		}
	};

	///////////////////////////////////////////////////////////////////////////
	// class PixByPix4to1
	/// @brief        ,
	/// 
	///    
	///////////////////////////////////////////////////////////////////////////
	template<typename T,class Func=Vtl::Plf::F::Harris<T>,class P=Vtl::Plf::F11::DoNothing<T> >
	class PixByPix4to1
	: public FilterElementary
	{
	public:
		InPort<T> in_1;
		InPort<T> in_2;
		InPort<T> in_3;
		InPort<T> in_4;
		OutPort<T,P> out;
		Func func;
	public:
		PixByPix4to1()
			: FilterElementary(4,1)
		{
		}
		PixByPix4to1(std::string name,Func functor,P post_proc=Vtl::Plf::F11::DoNothing<T>())
			: FilterElementary(4,1)
		{
			Init(name,functor,post_proc);
		}
		virtual ~PixByPix4to1()
		{
		}
		void Init(std::string name,Func functor,P post_proc=Vtl::Plf::F11::DoNothing<T>())
		{
			setName(name);
			func=functor;
			out.setPostProc(post_proc);
			AddPort(in_1,0,Extent(0,0,0,0));
			AddPort(in_2,1,Extent(0,0,0,0));
			AddPort(in_3,2,Extent(0,0,0,0));
			AddPort(in_4,3,Extent(0,0,0,0));
			AddPort(out,0);
		}
		virtual void set_OutPortsImagesSizeAndInPortsExtents()
		{
			FilterElementary::set_OutPortsImagesSizeAndInPortsExtents();
		}
		virtual void processLine() 
		{
			ImgSize ims=in_1.getImageSize();
			const int w=ims.width;

			out.beginUpdate();
				ImageRef<T> img_1=in_1.getImg(); 
				ImageRef<T> img_2=in_2.getImg(); 
				ImageRef<T> img_3=in_3.getImg(); 
				ImageRef<T> img_4=in_4.getImg(); 

				const T* p_1=img_1[0];
				const T* p_2=img_2[0];
				const T* p_3=img_3[0];
				const T* p_4=img_4[0];
				
				OutPortNoPostProc<T>::BufType::Iter_Line output=out.begin();
				
				for(int x=0; x<w; x++)
				{
					*output++ = func(*p_1++,*p_2++,*p_3++,*p_4++);
				}

				in_1.ToNextLine(Y_pos()+1);
				in_2.ToNextLine(Y_pos()+1);
				in_3.ToNextLine(Y_pos()+1);
				in_4.ToNextLine(Y_pos()+1);
			out.endUpdate();
		}
	};
	///////////////////////////////////////////////////////////////////////////
	// class PixByPix5to1
	/// @brief        ,
	/// 
	///    
	///////////////////////////////////////////////////////////////////////////
	template<typename T,class Func=Vtl::Plf::F::Harris<T>,class P=Vtl::Plf::F11::DoNothing<T> >
	class PixByPix5to1
	: public FilterElementary
	{
	public:
		InPort<T> in_1;
		InPort<T> in_2;
		InPort<T> in_3;
		InPort<T> in_4;
		InPort<T> in_5;
		OutPort<T,P> out;
		Func func;
	public:
		PixByPix5to1()
			: FilterElementary(5,1)
		{
		}
		PixByPix5to1(std::string name,Func functor,P post_proc=Vtl::Plf::F11::DoNothing<T>())
			: FilterElementary(5,1)
		{
			Init(name,functor,post_proc);
		}
		virtual ~PixByPix5to1()
		{
		}
		void Init(std::string name,Func functor,P post_proc=Vtl::Plf::F11::DoNothing<T>())
		{
			setName(name);
			func=functor;
			out.setPostProc(post_proc);
			AddPort(in_1,0,Extent(0,0,0,0));
			AddPort(in_2,1,Extent(0,0,0,0));
			AddPort(in_3,2,Extent(0,0,0,0));
			AddPort(in_4,3,Extent(0,0,0,0));
			AddPort(in_5,4,Extent(0,0,0,0));
			AddPort(out,0);
		}
		virtual void set_OutPortsImagesSizeAndInPortsExtents()
		{
			FilterElementary::set_OutPortsImagesSizeAndInPortsExtents();
		}
		virtual void processLine() 
		{
			ImgSize ims=in_1.getImageSize();
			const int w=ims.width;

			out.beginUpdate();
				ImageRef<T> img_1=in_1.getImg(); 
				ImageRef<T> img_2=in_2.getImg(); 
				ImageRef<T> img_3=in_3.getImg(); 
				ImageRef<T> img_4=in_4.getImg(); 
				ImageRef<T> img_5=in_5.getImg(); 

				const T* p_1=img_1[0];
				const T* p_2=img_2[0];
				const T* p_3=img_3[0];
				const T* p_4=img_4[0];
				const T* p_5=img_5[0];
				
				OutPortNoPostProc<T>::BufType::Iter_Line output=out.begin();
				
				for(int x=0; x<w; x++)
				{
					*output++ = func(*p_1++,*p_2++,*p_3++,*p_4++,*p_5++);
				}
				in_1.ToNextLine(Y_pos()+1);
				in_2.ToNextLine(Y_pos()+1);
				in_3.ToNextLine(Y_pos()+1);
				in_4.ToNextLine(Y_pos()+1);
				in_5.ToNextLine(Y_pos()+1);
			out.endUpdate();
		}
	};


	///////////////////////////////////////////////////////////////////////////
	// class PixByPix3to3
	/// @brief     3   3 ,
	///        
	///        .  
	///    .   
	///      , @see Vtl::Plf::C, @see Vtl::Plf::F::DiZenzo, 
	///    
	///////////////////////////////////////////////////////////////////////////
	template<typename T,class Func=Vtl::Plf::F::Harris<T>,class P1=Vtl::Plf::F11::DoNothing<T> ,class P2=Vtl::Plf::F11::DoNothing<T> ,class P3=Vtl::Plf::F11::DoNothing<T> >
	class PixByPix3to3
	: public FilterElementary
	{
	public:
		InPort<T> in_1;
		InPort<T> in_2;
		InPort<T> in_3;
		OutPort<T,P1> out_1;
		OutPort<T,P2> out_2;
		OutPort<T,P3> out_3;
		Func func;
	public:
		PixByPix3to3()
			: FilterElementary(3,3)
		{
		}
		PixByPix3to3(std::string name,Func functor
			,P1 post_proc_1=Vtl::Plf::F11::DoNothing<T>()
			,P2 post_proc_2=Vtl::Plf::F11::DoNothing<T>()
			,P3 post_proc_3=Vtl::Plf::F11::DoNothing<T>())
			: FilterElementary(3,3)
		{
			Init(name,functor,post_proc_1,post_proc_2,post_proc_3);
		}
		virtual ~PixByPix3to3()
		{
		}
		void Init(std::string name,Func functor
			,P1 post_proc_1=Vtl::Plf::F11::DoNothing<T>()
			,P2 post_proc_2=Vtl::Plf::F11::DoNothing<T>()
			,P3 post_proc_3=Vtl::Plf::F11::DoNothing<T>())
		{
			setName(name);
			func=functor;
			out_1.setPostProc(post_proc_1);
			out_2.setPostProc(post_proc_2);
			out_3.setPostProc(post_proc_3);
			
			AddPort(in_1,0,Extent(0,0,0,0));
			AddPort(in_2,1,Extent(0,0,0,0));
			AddPort(in_3,2,Extent(0,0,0,0));
			
			AddPort(out_1,0);
			AddPort(out_2,1);
			AddPort(out_3,2);
		}
		virtual void set_OutPortsImagesSizeAndInPortsExtents()
		{
			ImgSize ims_1=in_1.getImageSize();
			ImgSize ims_2=in_2.getImageSize();
			ImgSize ims_3=in_3.getImageSize();
			if(ims_1!=ims_2 ||ims_1!=ims_3 )
			{
				std::ostringstream os;
				os<< "Filter Vtl::Plf::PixByPix3to1 \"" << Name() <<"\" : input images are of different size."
					<< " in_1 : " << ims_1 << " in_2 : " << ims_2 << " in_3 : " << ims_3;
				throw os.str();
			}
			FilterElementary::set_OutPortsImagesSizeAndInPortsExtents();
		}
		virtual void processLine() 
		{
			ImgSize ims=in_1.getImageSize();
			const int w=ims.width;

			out_1.beginUpdate();
			out_2.beginUpdate();
			out_3.beginUpdate();
				ImageRef<T> img_1=in_1.getImg(); 
				ImageRef<T> img_2=in_2.getImg(); 
				ImageRef<T> img_3=in_3.getImg(); 

				const T* p_1=img_1[0];
				const T* p_2=img_2[0];
				const T* p_3=img_3[0];
				
				OutPortNoPostProc<T>::BufType::Iter_Line o_1=out_1.begin();
				OutPortNoPostProc<T>::BufType::Iter_Line o_2=out_2.begin();
				OutPortNoPostProc<T>::BufType::Iter_Line o_3=out_3.begin();
				
				for(int x=0; x<w; x++)
				{
					func(p_1[x],p_2[x],p_3[x],o_1[x],o_2[x],o_3[x]);
				}
				in_3.ToNextLine(Y_pos()+1);
				in_2.ToNextLine(Y_pos()+1);
				in_1.ToNextLine(Y_pos()+1);
			out_3.endUpdate();
			out_2.endUpdate();
			out_1.endUpdate();
		}
	};
	///////////////////////////////////////////////////////////////////////////
	// class PixByPix2to3
	/// @brief     \f$I_1\f$  \f$I_2\f$  3 
	///     : \n
	///          \f$I_{11}=I_1^2\f$\n
	///          \f$I_{12}=I_1*I_2\f$\n
	///          \f$I_{11}=I_1^2\f$\n
	/// ..   ,    
	///  . (     ,
	///         ,   
	///   @see class Average).
	///////////////////////////////////////////////////////////////////////////
	template<typename T, class P=Vtl::Plf::F::CorrMatrix<T,T&> >
	class PixByPix2to3
	: public FilterElementary
	{
		P functor;
	public:
		InPort<T> in_1;
		InPort<T> in_2;
		OutPort<T> out_11;
		OutPort<T> out_12;
		OutPort<T> out_22;
	public:
		PixByPix2to3()
			: FilterElementary(2,3)
		{
		}

		PixByPix2to3(std::string name)
			: FilterElementary(2,3)
		{
			Init(std::string name);
		}
		void Init(std::string name,P func=F23::CorrMatrix<T,T&>() )
		{
			setName(name);
			functor=func;
			
			AddPort(in_1,0,Extent(0,0,0,0));
			AddPort(in_2,1,Extent(0,0,0,0));
			AddPort(out_11,0);
			AddPort(out_12,1);
			AddPort(out_22,2);
		}
		virtual ~PixByPix2to3()
		{
		}
		virtual void set_OutPortsImagesSizeAndInPortsExtents()
		{
			ImgSize ims_1=in_1.getImageSize();
			ImgSize ims_2=in_2.getImageSize();
			if(ims_1!=ims_2)
			{
				std::ostringstream os;
				os<< "Filter Vtl::Plf::CorrelationMatrix2x2 \"" << Name() <<"\" : input images are of different size."
					<< " in_1 : " << ims_1 << " in_2 : " << ims_2 ;
				throw os.str();
			}
			Vtl::Plf::FilterElementary::set_OutPortsImagesSizeAndInPortsExtents();
		}
		virtual void processLine() 
		{
			ImgSize ims=in_1.getImageSize();
			const int w=ims.width;

			out_11.beginUpdate();
			out_12.beginUpdate();
			out_22.beginUpdate();
				ImageRef<T> img_1=in_1.getImg(); 
				ImageRef<T> img_2=in_2.getImg(); 
				const T* p1=img_1[0];
				const T* p2=img_2[0];
				OutPortNoPostProc<T>::BufType::Iter_Line o_11=out_11.begin();
				OutPortNoPostProc<T>::BufType::Iter_Line o_12=out_12.begin();
				OutPortNoPostProc<T>::BufType::Iter_Line o_22=out_22.begin();
				for(int x=0; x<w; x++)
				{
					functor(p1[x],p2[x],o_11[x],o_12[x],o_22[x]);
				}
				in_1.ToNextLine(Y_pos()+1);
				in_2.ToNextLine(Y_pos()+1);
			
			out_22.endUpdate();
			out_12.endUpdate();
			out_11.endUpdate();
		}
	};

	///////////////////////////////////////////////////////////////////////////
	// class EnvProc1to1
	/// @brief  1   1 .    (x,y)   
	///   ,       
	/// .     -   
	/// .   -   
	/// - "Non Maximum Supression"    (Vtl::Plf::F::Harris, Vtl::Plf::F::MinEigenVal)
	/// -    H-image
	///  .    . 
	///    :
	/// \code
	/// template<typename T,...> 
	/// class MyFunctor
	/// {
	/// 	int nls;
	/// 	//   
	/// public:
	/// 	/// !!! MUST be called by filter BEFORE using 'operator()', by default
	/// 	/// next line shist parameter set to value which probably most quick lead 
	/// 	/// a 'Invalid memory access' error (for bug search convenience).
	/// 	void set_IterNextLineShift(int value_nls) { nls=value_nls;}
	/// 	Extent ext() const {return Extent(/* -   */); }
	///     ///  ,      -  ,
	///     ///   -        .
	///     /// !     ,    
	///     ///     ,  ,  
	///     ///    ,     sizeof. 
	///     ///      ,    
	///     ///   FilterElementary,      
	///     ///  ,        
	///     ///   .
	/// 	MyFunctor() {nls=0x5FFFFFFF;}
	/// 	MyFunctor(/*,  */) {nls=0x5FFFFFFF;/*    */}
	/// 	T operator()(const T* input)
	/// 	{
	/// 		return ...;
	/// 	}
	/// };
	/// \endcode
	///      , - :
	/// \code
	///     #define HARRIS
	///     //#define KANADE_LUKAS
	/// 	const int bits=8; 
	///		//       0  maxPixelVal
	/// 	const float maxPixelVal=float((1<<bits)-1); 
	/// 	float sigma=1.5f;
	/// 	float env_size_in_sigmas=2.5f;
	/// 	float gamma=2.2f;
	/// 	Vtl::Plf::F::GammaCor<float> gam_cor(maxPixelVal,gamma);
	/// 	#if defined(HARRIS)
	/// 		Vtl::Plf::F::Harris<float> func(sigma,k); 
	/// 		Vtl::Plf::Detect<float,Vtl::Plf::F::Harris<float>,Vtl::Plf::F::GammaCor<float> > det("Harris",sigma,env_size_in_sigmas,func,gam_cor);
	///		#elif defined(KANADE_LUKAS)
	///			Vtl::Plf::F::MinEigenVal<float> func(sigma);
	///			Vtl::Plf::Detect<float,Vtl::Plf::F::MinEigenVal<float>,Vtl::Plf::F::GammaCor<float> > det(out_name,sigma,env_size_in_sigmas,func,gam_cor);
	///		#elif defined(...)
	///		   ...
	///		#edif
	/// 	Vtl::Plf::F::HarrisNMS<float> harris_nms;
	/// 	Vtl::Plf::EnvProc1to1<float,Vtl::Plf::F::HarrisNMS<float> > loc_max("loc_max",harris_nms);
	///     ...
	///     ...
	///     src.out.ConnectTo(&det.in());
	///     det.out().ConnectTo(&loc_max.in);
	///     loc_max.out.ConnectTo(&dst.in);
	/// \endcode
	///////////////////////////////////////////////////////////////////////////
	template<typename T,class Func=F::HarrisNMS<T>,class P=Vtl::Plf::F11::DoNothing<T> >
	class EnvProc1to1
	: public FilterElementary
	{
		Func func;
	public:
		InPort<T> in;
		OutPort<T,P> out;
	public:
		EnvProc1to1()
			: FilterElementary(1,1)
		{
		}
		EnvProc1to1(std::string name,Func functor,P post_proc=Vtl::Plf::F11::DoNothing<T>())
			: FilterElementary(1,1)
		{
			Init(name,functor,post_proc);
		}
		void Init(std::string name,Func functor,P post_proc=Vtl::Plf::F11::DoNothing<T>())
		{
			setName(name);
			out.setPostProc(post_proc);
			func=functor;
			AddPort(in,0,functor.ext());
			AddPort(out,0);
		}
		virtual ~EnvProc1to1()
		{
		}
		virtual void processLine() 
		{
			ImgSize ims=in.getImageSize();
			const int w=ims.width;
			out.beginUpdate();
			ImageRef<T> img=in.getImg(); 
			const T* input=img[0];
			int nls=img.nls();
			func.set_IterNextLineShift(nls);
			OutPort<T>::BufType::Iter_Line output=out.begin();
			for(int x=0; x<w; ++x)
			{
				*output++ = func(input+x);
			}
			in.ToNextLine(Y_pos()+1);
			out.endUpdate();
		}
	};
	///////////////////////////////////////////////////////////////////////////
	// class EnvProc3to1
	/// @brief  3   1 .    (x,y)   
	///   ,      (x,y)   
	///  in_val      (x,y)   in_gx,in_gy. 
	///     -   
	/// .   -   
	///  "Non Maximum Supression"    (Vtl::Plf::Canny). 
	///    :
	/// \code
	/// template<typename T,...> 
	/// class MyFunctor
	/// {
	/// 	int nls;
	/// 	//   
	/// public:
	/// 	/// !!! MUST be called by filter BEFORE using 'operator()', by default
	/// 	/// next line shist parameter set to value which probably most quick lead 
	/// 	/// a 'Invalid memory access' error (for bug search convenience).
	/// 	void set_IterNextLineShift(int value_nls) { nls=value_nls;}
	/// 	Extent ext() const {return Extent(/* -   */); }
	///     ///  ,      -  ,
	///     ///   -        .
	///     /// !     ,    
	///     ///     ,  ,  
	///     ///    ,     sizeof. 
	///     ///      ,    
	///     ///   FilterElementary,      
	///     ///  ,        
	///     ///   .
	/// 	MyFunctor() {nls=0x5FFFFFFF;}
	/// 	MyFunctor(/*,  */) {nls=0x5FFFFFFF;/*    */}
	/// 	T operator()(const T* modul,T gx,T gy)
	/// 	{
	/// 		return ...;
	/// 	}
	/// };
	/// \endcode
	///     NonMaximumSupression   
	///  ( CannyNMS, CannyNMS_2).  ,  gx,gy - 
	///   (. ),  modul -   ( 
	///  ).     \c    
	/// ,    gx,gy. \n
	///     .   Vtl::Plf::Canny.
	///////////////////////////////////////////////////////////////////////////
	template<typename T,class Func=Fpie::CannyNMS<T>,class P=Vtl::Plf::F11::DoNothing<T> >
	class EnvProc3to1
	: public FilterElementary
	{
		Func func;
	public:
		InPort<T> in_val;
		InPort<T> in_gx;
		InPort<T> in_gy;
		OutPort<T,P> out;
	public:
		EnvProc3to1()
			: FilterElementary(3,1)
		{
		}
		EnvProc3to1(std::string name,Func functor=Fpie::CannyNMS<T>(),P post_proc=Vtl::Plf::F11::DoNothing<T>())
			: FilterElementary(3,1)
		{
			Init(name,functor,post_proc);
		}
		void Init(std::string name,Func functor=Fpie::CannyNMS<T>(),P post_proc=Vtl::Plf::F11::DoNothing<T>())
		{
			setName(name);
			out.setPostProc(post_proc);
			func=functor;
			AddPort(in_val,0,func.ext());
			AddPort(in_gx ,1,Extent(0));//Temporary! func.ext()  Extent(0) -    .
			AddPort(in_gy ,2,Extent(0));//Temporary! func.ext()  Extent(0) -    .
//			AddPort(in_gx ,1,func.ext());//Temporary! func.ext()  Extent(0) -    .
//			AddPort(in_gy ,2,func.ext());//Temporary! func.ext()  Extent(0) -    .
			AddPort(out,0);
		}
		virtual ~EnvProc3to1()
		{
		}
		virtual void processLine() 
		{
			ImgSize ims=in_val.getImageSize();
			const int w=ims.width;


			ImageRef<T> img_v=in_val.getImg(); 
			const T* v=img_v[0];
			const int nls=img_v.nls();
			ImageRef<T> img_x=in_gx.getImg(); 
			const T* i_x=img_x[0];//-img_x.nls();;
			ImageRef<T> img_y=in_gy.getImg(); 
			const T* i_y=img_y[0];//-img_y.nls();

			float sigma=1.0f;
			Vtl::Plf::F11::GammaCorMult<T> gam_mult(255,2.2f,sqrt(2.0f*3.1415926f)*sigma);

			out.beginUpdate();
			OutPort<T>::BufType::Iter_Line output=out.begin();

//			printf("NonMaxSuppression: out.Y() | in_val.Y(), in_gx.Y(), in_gy.Y()=( %d | %d, %d, %d )\n",out.Y_wait(),in_val.Y(),in_gx.Y(),in_gy.Y());

			func.set_IterNextLineShift(nls);
			for(int x=0; x<w; x++)
				*output++=func(v+x,i_x[x],i_y[x]);

			in_val.ToNextLine(Y_pos()+1);
			in_gx.ToNextLine(Y_pos()+1);
			in_gy.ToNextLine(Y_pos()+1);
			out.endUpdate();
		}
	};
	///////////////////////////////////////////////////////////////////////////
	// class EnvProc1to2
	/// @brief  1   2 a.    -  
	///      ,  -  , 
	///   H-image \ref Lit_HImg, \see Vtl::Plf::F:Himg2 .
	///
	///
	///
	///   ,  
	///    (x,y)   
	///   ,       
	/// .     -   
	/// .   -   
	/// - "Non Maximum Supression"    (Vtl::Plf::F::Harris, Vtl::Plf::F::MinEigenVal)
	/// -    H-image
	///  .    . 
	///    :
	/// \code
	/// template<typename T,...> 
	/// class MyFunctor
	/// {
	/// 	int nls;
	/// 	//   
	/// public:
	/// 	/// !!! MUST be called by filter BEFORE using 'operator()', by default
	/// 	/// next line shist parameter set to value which probably most quick lead 
	/// 	/// a 'Invalid memory access' error (for bug search convenience).
	/// 	void set_IterNextLineShift(int value_nls) { nls=value_nls;}
	/// 	Extent ext() const {return Extent(/* -   */); }
	///     ///  ,      -  ,
	///     ///   -        .
	///     /// !     ,    
	///     ///     ,  ,  
	///     ///    ,     sizeof. 
	///     ///      ,    
	///     ///   FilterElementary,      
	///     ///  ,        
	///     ///   .
	/// 	MyFunctor() {nls=0x5FFFFFFF;}
	/// 	MyFunctor(/*,  */) {nls=0x5FFFFFFF;/*    */}
	/// 	T operator()(const T* input)
	/// 	{
	/// 		return ...;
	/// 	}
	/// };
	/// \endcode
	///      , - :
	/// \code
	///     #define HARRIS
	///     //#define KANADE_LUKAS
	/// 	const int bits=8; 
	///		//       0  maxPixelVal
	/// 	const float maxPixelVal=float((1<<bits)-1); 
	/// 	float sigma=1.5f;
	/// 	float env_size_in_sigmas=2.5f;
	/// 	float gamma=2.2f;
	/// 	Vtl::Plf::F::GammaCor<float> gam_cor(maxPixelVal,gamma);
	/// 	#if defined(HARRIS)
	/// 		Vtl::Plf::F::Harris<float> func(sigma,k); 
	/// 		Vtl::Plf::Detect<float,Vtl::Plf::F::Harris<float>,Vtl::Plf::F::GammaCor<float> > det("Harris",sigma,env_size_in_sigmas,func,gam_cor);
	///		#elif defined(KANADE_LUKAS)
	///			Vtl::Plf::F::MinEigenVal<float> func(sigma);
	///			Vtl::Plf::Detect<float,Vtl::Plf::F::MinEigenVal<float>,Vtl::Plf::F::GammaCor<float> > det(out_name,sigma,env_size_in_sigmas,func,gam_cor);
	///		#elif defined(...)
	///		   ...
	///		#edif
	/// 	Vtl::Plf::F::HarrisNMS<float> harris_nms;
	/// 	Vtl::Plf::EnvProc1to1<float,Vtl::Plf::F::HarrisNMS<float> > loc_max("loc_max",harris_nms);
	///     ...
	///     ...
	///     src.out.ConnectTo(&det.in());
	///     det.out().ConnectTo(&loc_max.in);
	///     loc_max.out.ConnectTo(&dst.in);
	/// \endcode
	///////////////////////////////////////////////////////////////////////////
	template<typename T,class Func=F::Himg2<T>,class P1=Vtl::Plf::F11::DoNothing<T>,class P2=Vtl::Plf::F11::DoNothing<T> >
	class EnvProc1to2
	: public FilterElementary
	{
		Func func;
	public:
		InPort<T> in;
		OutPort<T,P1> out_x;
		OutPort<T,P2> out_y;
	public:
		EnvProc1to2()
			: FilterElementary(1,2)
		{
		}
		EnvProc1to2(std::string name,Func functor,P1 post_proc_x=Vtl::Plf::F11::DoNothing<T>(),P2 post_proc_y=Vtl::Plf::F11::DoNothing<T>())
			: FilterElementary(1,2)
		{
			Init(name,functor,post_proc_x,post_proc_y);
		}
		void Init(std::string name,Func functor=Fpie::Himg2<T>(),P1 post_proc_x=Vtl::Plf::F11::DoNothing<T>(),P2 post_proc_y=Vtl::Plf::F11::DoNothing<T>())
		{
			setName(name);
			out_x.setPostProc(post_proc_x);
			out_y.setPostProc(post_proc_y);
			func=functor;
			AddPort(in,0,functor.ext());
			AddPort(out_x,0);
			AddPort(out_y,1);
		}
		virtual ~EnvProc1to2()
		{
		}
		virtual void processLine() 
		{
			ImgSize ims=in.getImageSize();
			const int w=ims.width;
			out_x.beginUpdate();
			out_y.beginUpdate();
			ImageRef<T> img=in.getImg(); 
			const T* input=img[0];
			int nls=img.nls();
			func.set_IterNextLineShift(nls);
			OutPort<T>::BufType::Iter_Line o_x=out_x.begin();
			OutPort<T>::BufType::Iter_Line o_y=out_y.begin();
			for(int x=0; x<w; ++x)
			{
				func(input+x,*o_x++,*o_y++);
			}
			in.ToNextLine(Y_pos()+1);
			out_y.endUpdate();
			out_x.endUpdate();
		}
	};
	
	///////////////////////////////////////////////////////////////////////////
	// class ConvolveHorz
	/// @brief          
	///    
	///////////////////////////////////////////////////////////////////////////
	template<typename T,class P=Vtl::Plf::F11::DoNothing<T> >
	class ConvolveHorz
		: public FilterElementary
	{
	public:
		InPort<T> in;
		OutPort<T,P> out;
		const Kernel::Base<T,Kernel::KT_Conv> *ker;
		int tmp;
	public:
		ConvolveHorz()
			: FilterElementary(1,1)
		{
		}
		ConvolveHorz(std::string name,const Kernel::Base<T,Kernel::KT_Conv> *kernel,P post_proc=Vtl::Plf::F11::DoNothing<T>())
			: FilterElementary(1,1)
		{
			Init(name,kernel,post_proc);
		}
		~ConvolveHorz()
		{
		}
		void Init(std::string name,const Kernel::Base<T,Kernel::KT_Conv> *kernel,P post_proc=Vtl::Plf::F11::DoNothing<T>())
		{
			setName(name);
			out.setPostProc(post_proc);
			ker=kernel;
			Extent ext(ker->left(),ker->right(),0,0);
			AddPort(in,0,ext);
			AddPort(out,0);
			tmp=0;
		}
		virtual void processLine() 
		{
			ImgSize ims=in.getImageSize();
			const int w=ims.width;

			out.beginUpdate();
//			printf("HConv: out.Y() | in.Y() = ( %d | %d ) \n",out.Y_wait(),in.Y());

			ImageRef<T> img=in.getImg(); 
			const T* input=img[0]-ker->left();
			OutPortNoPostProc<T>::BufType::Iter_Line output=out.begin();
			Vtl::ImAlg::Conv::ConvolveNoExt(input,input+img.w(),output,*ker);

			in.ToNextLine(Y_pos()+1);
			out.endUpdate();
//			printf("-H---- %d\n",tmp++);
		}
	};

	///////////////////////////////////////////////////////////////////////////
	// class ConvolveHorz
	/// @brief        
	/// \b  \b     (   .  
	///    
	///////////////////////////////////////////////////////////////////////////
	template<typename T,class P=Vtl::Plf::F11::DoNothing<T> >
	class ConvolveVert
		: public FilterElementary
	{
	public:
		InPort<T> in;
		OutPort<T,P> out;
		const Kernel::Base<T,Kernel::KT_Conv> *ker;
		int tmp;
	public:
		ConvolveVert()
			: FilterElementary(1,1)
		{
		}
		ConvolveVert(std::string name,const Kernel::Base<T,Kernel::KT_Conv> *kernel,P post_proc=Vtl::Plf::F11::DoNothing<T>())
			: FilterElementary(1,1)
		{
			Init(name,kernel,post_proc);
		}
		~ConvolveVert()
		{
		}
		void Init(std::string name,const Kernel::Base<T,Kernel::KT_Conv> *kernel,P post_proc=Vtl::Plf::F11::DoNothing<T>())
		{
			setName(name);
			out.setPostProc(post_proc);
			ker=kernel;
			Extent ext(0,0,ker->left(),ker->right());
			AddPort(in,0,ext);
			AddPort(out,0);
//			printf("ConvolveVert: ext= %d   %d\n",ext.before,ext.forward);
			tmp=0;
		}
		virtual void processLine() 
		{
			ImgSize ims=in.getImageSize();
			const int w=ims.width;

			out.beginUpdate();

			ImageRef<T> img=in.getImg(); 
			LineIterator<T> input_begin(img);
			LineIterator<T> input_end(img,true);
			OutPortNoPostProc<T>::BufType::Iter_Line output=out.begin();
			Vtl::ImAlg::Conv::ConvolveNoExt(input_begin,input_end,output,*ker);

			//in.setProcessed_YES();
			in.ToNextLine(Y_pos()+1);
			out.endUpdate();
//			printf("-V---- %d\n",tmp++);
		}
	};

	///////////////////////////////////////////////////////////////////////////
	// class FilterResampleHorz
	/// @brief    (/)   
	/// 
	///////////////////////////////////////////////////////////////////////////
	template<typename T,class P=Vtl::Plf::F11::DoNothing<T> >
	class FilterResampleHorz
		: public FilterElementary
	{
	public:
		InPort<T> in;
		OutPort<T,P> out;
		int w_new;
		int w_old;
		int tmp;
	public:
		FilterResampleHorz()
			: FilterElementary(1,1),w_new(0)
		{
		}
		FilterResampleHorz(std::string name,int new_width,P post_proc=Vtl::Plf::F11::DoNothing<T>())
			: FilterElementary(1,1)
		{
			Init(name,new_width,post_proc);
		}
		virtual ~FilterResampleHorz()
		{
		}
		
		void Init(std::string name,int new_width,P post_proc=Vtl::Plf::F11::DoNothing<T>())
		{
			setName(name);
			out.setPostProc(post_proc);
			w_new=new_width;
			AddPort(in,0,Extent(0,0,0,0));
			AddPort(out,0);
			tmp=0;
		}
		virtual void set_OutPortsImagesSizeAndInPortsExtents()
		{
			ImgSize ims=in.getImageSize();
			w_old=ims.width;
			ims.width=w_new;
			out.set_bufferImgSize(ims);
			// Extension -    ,    .
		}

		virtual void processLine() 
		{
			out.beginUpdate();
		
			ImageRef<T> img=in.getImg(); 
			const T* input=img[0];
			int w_in=w_old;
			int w_out=w_new;
			OutPortNoPostProc<T>::BufType::Iter_Line output=out.begin();
			//if(w_out>w_in)
			{
				double k=(w_in-1.)/(w_out-1.);
				for(int x_out=0; x_out<w_out ; ++x_out)
				{
					double x=x_out*k;
					int x_in=(int)x;
					x-=x_in;
					*output++=T(input[x_in]*(1-x)+input[x_in+1]*x);
				}
//				OutPortNoPostProc<T>::BufType::Iter_Line output=out.begin_line_insert();
//				for(int x=0; x<w_new; x++)
//					*output++=functor(*input++);
			}
			in.ToNextLine(Y_pos()+1);
			out.endUpdate();
//			printf("-RsH-- %d\n",tmp++);
		}
	};
	///////////////////////////////////////////////////////////////////////////
	// class FilterResampleVert
	/// @brief    (/)   .
	/// 
	///////////////////////////////////////////////////////////////////////////
	template<typename T,class P=Vtl::Plf::F11::DoNothing<T> >
	class FilterResampleVert
		: public FilterElementary
	{
	public:
		InPort<T> in;
		OutPort<T,P> out;
		int h_new;
		int h_old;
		double cof;
		int y;
	public:
		FilterResampleVert()
			: FilterElementary(1,1),h_new(0)
		{
			y=0;
		}
		FilterResampleVert(std::string name,int new_height,P post_proc=Vtl::Plf::F11::DoNothing<T>())
			: FilterElementary(1,1)
		{
			Init(name,new_height,post_proc);
		}
		virtual ~FilterResampleVert()
		{
		}
		void Init(std::string name,int new_height,P post_proc=Vtl::Plf::F11::DoNothing<T>())
		{
			setName(name);
			out.setPostProc(post_proc);
			h_new=new_height;
			AddPort(in,0,Extent(0,0,0,1));
			AddPort(out,0);
			y=0;
		}
		virtual void set_OutPortsImagesSizeAndInPortsExtents()
		{
			ImgSize ims=in.getImageSize();
			h_old=ims.height;
			ims.height=h_new;
			out.set_bufferImgSize(ims);
			cof=((h_old-1.)/(h_new-1.));
			// Extension -    ,    .
		}

		virtual void processLine() 
		{
			out.beginUpdate();
		
			ImageRef<T> img=in.getImg(); 
			const T* input=img[0];
			int w=img.w();
			OutPortNoPostProc<T>::BufType::Iter_Line output=out.begin();

			double k=cof*y;
			int y_in=int(k);
			k-=y_in;
			//if(h_new>h_old)
			{
				for(int x=0; x<w ; ++x)
				{
					*output++=T(input[x]*(1-k)+input[x+img.nls()]*k);
				}

//				OutPortNoPostProc<T>::BufType::Iter_Line output=out.begin_line_insert();
//				for(int x=0; x<w_new; x++)
//					*output++=functor(*input++);
			}
			++y;
			int y_in_next=int(cof*y);
			if( ( y_in_next > y_in ) ) //|| ( y >= h_old ) )
			{
				//in.ToNextLine(y_in+1);
				in.ToNextLine(y_in_next);
			}
			out.endUpdate();
//			printf("-RsV-- %d\n",y);
//			printf("-RsV-- %d %d %d %d\n",out.buffer.ext().left,out.buffer.ext().right,out.buffer.ext().before,out.buffer.ext().forward);
		}
	};

/** @} */ // end of ElementaryFiltersDoxGroup  
} //namespace Plf 
} //namespace Vtl
/////////////////////////////////////////////////////////////////////////////
#endif //_PLF_FILTERS_H
