#ifndef _PLF_FUNCTORS_H
#define _PLF_FUNCTORS_H
/////////////////////////////////////////////////////////////////////////////
#include "alexandresku.h"
#include "ytypes.h"
#include "plf_strip_buffer_support.h"
/////////////////////////////////////////////////////////////////////////////
namespace Vtl
{
// @brief     plf_filters.h,   
namespace Plf 
{
/** @addtogroup ElementaryFiltersDoxGroup  Elementary filters
 *  @ingroup PipeLineFiltersDoxGroup  
 *
 *  @brief         
 *     (Pipe Line Filtering) 
 *
 *  @{
 */
	///////////////////////////////////////////////////////////////////////////////
	/// @brief    template<typename T,typename T1> T operator()(T1 value)
	///
	///   FilterLevels,   Vtl::Plf   , 
	///     Vtl::ImAlg::Tools     .
	/// 
	namespace F11
	{
		template<typename T> T sqr(T v) {return v*v;}
		///////////////////////////////////////////////////////////////////////////
		/// \brief functor DoNothing - 
		///
		///   FilterLevels     .  
		///     ,  OutPort, 
		///    - ,     
		///  ,    \b  ,  
		///    ,        
		///         .
		///       FilterLevels   
		///       .
		/// do nothing
		///////////////////////////////////////////////////////////////////////////
		template<typename T> 
		struct DoNothing
		{
			T operator()(T val) {return val;}
		};
		///////////////////////////////////////////////////////////////////////////
		/// \brief functor Sqrt -      T.
		///////////////////////////////////////////////////////////////////////////
		template<typename T> 
		struct Convert
		{
			template<typename T1> 
			T operator()(T1 val) {return T(val);}
		};
		///////////////////////////////////////////////////////////////////////////
		/// \brief functor Sqrt -      T.
		///////////////////////////////////////////////////////////////////////////
		struct Convert2ByteBounded
		{
			template<typename T1> 
			uint8_t operator()(T1 val) {return uint8_t( (val>255) ? 255 : ((val<0)?0:val) ); }
		};

		///////////////////////////////////////////////////////////////////////////
		/// \brief functor Threshold 
		///
		///   FilterLevels     .  
		///     ,  OutPort, 
		///    - ,     
		///  ,    \b  ,  
		///    ,        
		///         .
		///       FilterLevels   
		///       .
		/// do nothing
		///////////////////////////////////////////////////////////////////////////
		template<typename T,typename Comp> 
		class Threshold
		{
			T tr;
			Comp c;
			T r;
		public:
			Threshold(T threshold=128,Comp comp=std::less<T>(),T replace=T(0)) 
				: tr(threshold),c(comp),r(replace)
			{}
			T operator()(T val) 
			{
				return c(val,tr)?r:val;
			}
		};
		///////////////////////////////////////////////////////////////////////////
		/// \brief functor Comb -       .
		///
		///   FilterLevels     .  
		///     ,  OutPort, 
		///    - ,     
		///  ,    \b  ,  
		///    ,        
		///         .
		///       FilterLevels   
		///       .
		/// do nothing
		///////////////////////////////////////////////////////////////////////////
		template<typename T,class F1,class F2> 
		struct Comb
		{
			F1 first_;
			F2 second_;
			Comb() {}
			Comb(F1 first,F2 second) :first_(first),second_(second) {}
			template<typename T1> 
			T operator()(T1 val) {return second_(first_(val));}
		};
		/*
		template<typename T,class F1,class F2> 
		struct Comb
		{
			F1 first_;
			F2 second_;
			Comb() {}
			Comb(F1 first,F2 second) :first_(first),second_(second) {}
			T operator()(T val) {return second_(first_(val));}
		};
		*/
		///////////////////////////////////////////////////////////////////////////
		/// \brief functor Negative -   ,   
		///    0  maxval,    .
		///
		///   FilterLevels     .  
		///     ,  OutPort, 
		///    - ,     
		///  ,    \b  ,  
		///    ,        
		///         .
		///       FilterLevels   
		///       .
		/// do nothing
		///////////////////////////////////////////////////////////////////////////
		template<typename T>
		class Negative
		{
			T maxval;
		public:
			Negative() {}
			Negative(T max_val):maxval(max_val) {}
			T operator()(T val) const {return maxval-val;}
		};
		///////////////////////////////////////////////////////////////////////////
		/// \brief functor Bias -      .
		///
		///   FilterLevels     .  
		///     ,  OutPort, 
		///    - ,     
		///  ,    \b  ,  
		///    ,        
		///         .
		///       FilterLevels   
		///       .
		/// do nothing
		///////////////////////////////////////////////////////////////////////////
		template<typename T>
		class Bias
		{
			T bias;
		public:
			Bias() {}
			Bias(T add_to_pixel):bias(add_to_pixel) {}
			T operator()(T val) const {return val+bias;}
		};
		///////////////////////////////////////////////////////////////////////////
		/// \brief functor Mult -      .
		///
		///   FilterLevels     .  
		///     ,  OutPort, 
		///    - ,     
		///  ,    \b  ,  
		///    ,        
		///         .
		///       FilterLevels   
		///       .
		/// do nothing
		///////////////////////////////////////////////////////////////////////////
		template<typename T>
		class Mult
		{
			T mult;
		public:
			Mult() {}
			Mult(T mult_each_pixel):mult(mult_each_pixel) {}
			T operator()(T val) const {return val*mult;}
		};
		///////////////////////////////////////////////////////////////////////////
		/// \brief functor Mult -           T.
		///////////////////////////////////////////////////////////////////////////
		template<typename T> 
		struct ConvertMult
		{
			ConvertMult(T mult_) : mult(mult_) {}
			template<typename T1> 
			T operator()(T1 val) {return T(mult*val);}
		private:
			T mult;
		};
		///////////////////////////////////////////////////////////////////////////
		/// \brief functor Lin -       a*x+b.
		///
		///   FilterLevels     .  
		///     ,  OutPort, 
		///    - ,     
		///  ,    \b  ,  
		///    ,        
		///         .
		///       FilterLevels   
		///       .
		/// do nothing
		///////////////////////////////////////////////////////////////////////////
		template<typename T>
		class Lin
		{
			T _a;
			T _b;
		public:
			Lin(){}
			Lin(T a,T b):_a(a),_b(b) {}
			T operator()(T val) const {return _a*val+_b;}
		};
		///////////////////////////////////////////////////////////////////////////
		/// \brief functor Bounds -      min_val -:- max_val.
		///
		///   FilterLevels     .  
		///     ,  OutPort, 
		///    - ,     
		///  ,    \b  ,  
		///    ,        
		///         .
		///       FilterLevels   
		///       .
		/// do nothing
		///////////////////////////////////////////////////////////////////////////
		template<typename T>
		class ToBounds
		{
			T n;
			T x;
		public:
			ToBounds(T min_val=0,T max_val=255):n(min_val),x(max_val) {}
			T operator()(T val) const {return Vtl::ToBounds(val,n,x);}
		};
		///////////////////////////////////////////////////////////////////////////
		/// \brief functor Abs -    ()  .
		///////////////////////////////////////////////////////////////////////////
		template<typename T> 
		struct Abs
		{
			template<typename T1> 
			T operator()(T1 val) {return T(val<T1(0)?-val:val);}
		};

		///////////////////////////////////////////////////////////////////////////
		/// \brief functor Sqrt -   .
		///////////////////////////////////////////////////////////////////////////
		template<typename T>
		struct Sqrt
		{
			T operator() (T val) { return sqrt(val); }	
		};
		///////////////////////////////////////////////////////////////////////////
		/// \brief functor Sqr -     \f$v\f$   \f$v^2\f$.
		///
		///   FilterLevels     .  
		///     ,  OutPort, 
		///    - ,     
		///  ,    \b  ,  
		///    ,        
		///         .
		///       FilterLevels   
		///       .
		/// do nothing
		///////////////////////////////////////////////////////////////////////////
		template<typename T> 
		struct Sqr
		{
			template<typename T1> 
			T operator()(T1 val) {return T(val*val);}
		};
		/*
		template<typename T>
		class Sqr
		{
		public:
			Sqr(){}
			T operator()(T val) const {return val*val;}
		};
		*/
		///////////////////////////////////////////////////////////////////////////
		/// \brief functor GammaCor -   , ..   
		///   V   \f$ M*(V/M)^(1/gamma) \f$ , .  M -  
		///   ,      
		///  M=(1<<bits)-1.
		///
		///   FilterLevels     .  
		///     ,  OutPort, 
		///    - ,     
		///  ,    \b  ,  
		///    ,        
		///         .
		///       FilterLevels   
		///       .
		/// do nothing
		///////////////////////////////////////////////////////////////////////////
		template<typename T>
		class GammaCor
		{
			T maxval;
			double gamma;
		public:
			GammaCor() {}
			GammaCor(T max_val,double _gamma):maxval(max_val),gamma(1./std::max(_gamma,1.0e-8)) {}
			T operator()(T val) const {return T(maxval*pow(double(val/maxval),gamma));}
		};
		///////////////////////////////////////////////////////////////////////////
		/// \brief functor Log -           .
		///////////////////////////////////////////////////////////////////////////
		template<typename T> 
		class Log
		{
			double threshold;
		public:
			Log(double threshold_) : threshold(threshold_) {}
			template<typename T1> 
			T operator()(T1 val) {return T(std::log(std::max(threshold,double(val)))); }
		};
		///////////////////////////////////////////////////////////////////////////
		/// \brief functor GammaCorMult -   
		///  Mult,  GammaCor,    .
		///
		///   FilterLevels     .  
		///     ,  OutPort, 
		///    - ,     
		///  ,    \b  ,  
		///    ,        
		///         .
		///       FilterLevels   
		///       .
		/// do nothing
		///////////////////////////////////////////////////////////////////////////
		template<typename T>
		class GammaCorMult
		{
			T maxval;
			double gamma;
			double mult;
		public:
			GammaCorMult(){}
			GammaCorMult(T max_val,double _gamma,double mult_each_pixel)
				: maxval(max_val)
				  ,gamma(1./std::max(_gamma,1.0e-8))
				  ,mult(mult_each_pixel/max_val) {}
			T operator()(T val) const {return T(maxval*pow(double(val*mult),gamma));}
		};
	}//namespace F11
	///
	/// @brief    template<typename T,typename T1> T operator()(T1 value1,T1 value2)
	///
	///   FilterPixByPix2to1, 
	///     Vtl::ImAlg::Tools     .
	namespace F21
	{
		///////////////////////////////////////////////////////////////////////////
		/// \brief  SqrtOfSumSqr   PixByPix2to1.
		///        
		/// \f$\sqrt{I_1(x,y)^2+I_2(x,y)^2}\f$
		///////////////////////////////////////////////////////////////////////////
		template<typename T>
		class SqrtOfSumSqr
		{
		public:
			SqrtOfSumSqr() {}
			T operator()(T v1,T v2) const {return sqrt(v1*v1+v2*v2);}
		};


		template<typename T> 
		struct Sum
		{
			template<typename T1> 
			T operator()(T1 val1,T1 val2) {return T(val1+val2);}
		};
		template<typename T> 
		struct Dif
		{
			template<typename T1> 
			T operator()(T1 val1,T1 val2) {return T(val1-val2);}
		};
		template<typename T> 
		struct AbsDif
		{
			template<typename T1> 
			T operator()(T1 val1,T1 val2) {return T(std::abs(val1-val2));}
			T operator()(uint8_t val1,uint8_t val2) {return T(std::abs(int(val1)-int(val2)));}
			T operator()(uint16_t val1,uint16_t val2) {return T(std::abs(int(val1)-int(val2)));}
			T operator()(uint32_t val1,uint32_t val2) {return T(std::abs(int(val1)-int(val2)));}
		};
		///////////////////////////////////////////////////////////////////////////
		/// \brief  Prod   PixByPix2to1.
		///        
		/// \f$ I_1(x,y)*I_2(x,y) \f$
		///////////////////////////////////////////////////////////////////////////
		template<typename T> 
		struct Prod
		{
			template<typename T1> 
			T operator()(T1 val1,T1 val2) {return T(val1*val2);}
		};
		///////////////////////////////////////////////////////////////////////////
		/// \brief  Sharpen   PixByPix2to1.
		///     -      
		///    
		/// \f$\sqrt{I_{sqr_mean}(x,y)-I_{mean}(x,y)^2}\f$ , .. 
		///    \f$\sqrt{<I(x,y)^2>-<I(x,y)>^2}\f$ 
		///////////////////////////////////////////////////////////////////////////
		template<typename T>
		class Sharpen
		{
			float alpha;
		public:
			Sharpen(float amp=2.0f):alpha(amp) {}
			T operator()(T val,T mean) const {return mean+alpha*(val-mean);}
		};
		///////////////////////////////////////////////////////////////////////////
		/// \brief  sqrtOfSumSqr   PixByPix2to1.
		///     -      
		///    
		/// \f$\sqrt{I_{sqr_mean}(x,y)-I_{mean}(x,y)^2}\f$ , .. 
		///    \f$\sqrt{<I(x,y)^2>-<I(x,y)>^2}\f$ 
		///////////////////////////////////////////////////////////////////////////
		template<typename T>
		class Disp
		{
		public:
			Disp() {}
			T operator()(T mean,T sqr_mean) const {return sqrt(sqr_mean-mean*mean);}
		};
	}//namespace F21
	///////////////////////////////////////////////////////////////////////////////////////////////////
	/// @brief    template<typename T,typename T1> T operator()(T1 value1,T1 value2,T1 value3)
	///
	///   FilterPixByPix3to1, 
	///     Vtl::ImAlg::Tools     .
	namespace F31
	{
		/*!
		* \brief  Sum3  PixByPix3to1.
		*   \f$ result = v_{1},\ v_{v2},\ v_{3}\f$ .
		*
		*******************************************************************************************/
		template<typename T>
		class Sum3
		{
		public:
			Sum3() 
			{
			}
			T operator()(T v1,T v2,T v3) const 
			{
				return v1+v2+v3;;
			}
		};
		template<typename T> 
		struct HessianDet
		{
			template<typename T1> 
			T operator()(T1 Dxx,T1 Dxy,T1 Dyy) {return T(Dxx*Dyy-Dxy*Dxy);}
		};
		template<typename T> 
		struct HessianDetSqrt
		{
			template<typename T1> 
			T operator()(T1 Dxx,T1 Dxy,T1 Dyy) {T1 d=Dxx*Dyy-Dxy*Dxy; return T(d>=0?sqrt(d):-sqrt(d));}
			//T operator()(T1 Dxx,T1 Dxy,T1 Dyy) {return T(std::abs(Dxx*Dyy-Dxy*Dxy));}
		};

		/*!
		* \brief  Harris  PixByPix3to1.
		*    \f$o_{11},\ o_{12},\ o_{22}\f$    
		*     
		* \f$M=\biggl\Vert\begin{array}{cc} <g_x^2>  &  <g_x g_y> \\ <g_x g_y>  & <g_y^2>  \end{array}\biggr\Vert=
		* \biggl\Vert\begin{array}{cc} o_{11} &  o_{12} \\ o_{12}  & o_{22}  \end{array}\biggr\Vert\f$\n 
		*	,    
		*  ,     -: 
		* \f$ Harris=det(M)-k \cdot trace(M) = o_{11}o_{22}-o_{12}^2 - k \cdot (o_{11}+o_{22})^2 \f$,
		*  k=0.04.           .\n
		* \b :        \f$ Harris \f$, 
		*  \f$ 6.3 \cdot \sigma \cdot \root4\of {Harris} \f$ .     
		*  ,  \f$ Harris \sim g_i^4\f$,   \f$  \sigma \f$ - ,  
		* \f$  g_x,g_y \sim \frac{1}{\sigma} \f$, ..    
		*        (    ). 
		*  6.3 - .   
		*  ,        
		*    ( [0-255] -> [0-255]. \n
		* \b :  \f$  \sigma \f$    (multiscale)  
		* ,     \n
		* Y. Dufournaud, C. Schmid, and R. Horaud, "Matching images with different resolutions", 
		* Proceedings of IEEE Conference on Computer Vision and Pattern Recognition (CVPR00), 
		* 2000 (Hilton Head Island, SC, USA).  V. 1,   P. 612-618. ISBN: 0-7695-0662-3.
		* http://citeseer.ist.psu.edu/dufournaud00matching.html \n
		*      : \n
		* C. G. Harris and M. Stephens, "A combined corner and edge detector," 
		* In Proc. 4th Alvey Vision Conf., Manchester, pages 147-151, 1988.
		*
		*******************************************************************************************/
		template<typename T>
		class Harris
		{
			T k;
			T mult;
		public:
			Harris(T sigma=1.0f,T coef=0.04f) : mult(6.3f*sigma),k(coef) 
			{
				//mult=sqrt(2.0f*3.1415926f)*sigma;
			}
			T operator()(T v11,T v12,T v22) const 
			{
				T h=v11*v22-v12*v12-k*sqr(v11+v22);
				return mult*sqrt(sqrt(h));
			}
		};
		/*!
		* \brief  Trace  PixByPix3to1.   ,  \c  
		*    @see Vtl::Plf::F::Harris
		*
		*******************************************************************************************/
		template<typename T>
		class Trace 
		{
			T mult;
		public:
			Trace(T sigma=1.0f) : mult(3.0f*sigma)
			{
				//mult=sqrt(2.0f*3.1415926f)*sigma;
			}
			T operator()(T v11,T v12,T v22) const 
			{
				return mult*sqrt(v11+v22);
			}
		};
		/*!
		* \brief  MaxEigenVal  PixByPix3to1.
		*    \f$o_{11},\ o_{12},\ o_{22}\f$    
		*     
		* \f$M=\biggl\Vert\begin{array}{cc} <g_x^2>  &  <g_x g_y> \\ <g_x g_y>  & <g_y^2>  \end{array}\biggr\Vert=
		* \biggl\Vert\begin{array}{cc} o_{11} &  o_{12} \\ o_{12}  & o_{22}  \end{array}\biggr\Vert\f$\n 
		*	,    
		*  ,     (  DiZenzo): 
		* \f$ result=max(\lambda_1,\lambda_2)\f$,\n
		*  \f$ det\biggl\Vert \begin{array}{cc} o_{11}-\lambda & o_{12} \\ o_{12} & o_{22}-\lambda_2 \end{array} \biggl\Vert=0 \f$,
		* .. \f$ \lambda^2-\lambda(o_{11}+o_{22})+o_{11}o_{22}-o_{12}^2=0 \f$,
		*  \f$ D/4=((o_{11}+o_{22})/2)^2-(o_{11}o_{22}-o_{12}^2)=((o_{11}-o_{22})/2)^2+o_{12}^2 \f$ ,
		*  \f$ \lambda_1=\frac{o_{11}+o_{22}}{2}+\sqrt{\frac{D}{4}}\f$ ,
		* \f$ \lambda_2=\frac{o_{11}+o_{22}}{2}-\sqrt{\frac{D}{4}}\f$ 
		*******************************************************************************************/
		template<typename T>
		class MaxEigenVal
		{
			T mult;
		public:
			MaxEigenVal(T sigma=1.0f) : mult(2.9f*sigma)
			{
			}
			T operator()(T v11,T v12,T v22) const 
			{
				T a=(v11-v22)/2;
				T b=(v11+v22)/2;
				T D = sqrt(sqr(a) + sqr(v12));
				return mult*sqrt(b+D);
				//T lambda_1= b+D;
				//T lambda_2= b-D;
				//return lambda_2;
			}
		};
		/*!
		* \brief  MinEigenVal  PixByPix3to1.
		*    \f$o_{11},\ o_{12},\ o_{22}\f$    
		*     
		* \f$M=\biggl\Vert\begin{array}{cc} <g_x^2>  &  <g_x g_y> \\ <g_x g_y>  & <g_y^2>  \end{array}\biggr\Vert=
		* \biggl\Vert\begin{array}{cc} o_{11} &  o_{12} \\ o_{12}  & o_{22}  \end{array}\biggr\Vert\f$\n 
		*	,    
		*  ,     (  -): 
		* \f$ KL=min(\lambda_1,\lambda_2)\f$,\n
		*  \f$ det\biggl\Vert \begin{array}{cc} o_{11}-\lambda & o_{12} \\ o_{12} & o_{22}-\lambda_2 \end{array} \biggl\Vert=0 \f$,
		* .. \f$ \lambda^2-\lambda(o_{11}+o_{22})+o_{11}o_{22}-o_{12}^2=0 \f$,
		*  \f$ D/4=((o_{11}+o_{22})/2)^2-(o_{11}o_{22}-o_{12}^2)=((o_{11}-o_{22})/2)^2+o_{12}^2 \f$,
		*  \f$ \lambda_1=\frac{o_{11}+o_{22}}{2}+\sqrt{\frac{D}{4}}\f$ ,
		* \f$ \lambda_2=\frac{o_{11}+o_{22}}{2}-\sqrt{\frac{D}{4}}\f$ 
		*******************************************************************************************/
		template<typename T>
		class MinEigenVal
		{
			T mult;
		public:
			MinEigenVal(T sigma=1.0f) : mult(6.3f*sigma)
			{
			}
			T operator()(T v11,T v12,T v22) const 
			{
				T a=(v11-v22)/2;
				T b=(v11+v22)/2;
				T D = sqrt(sqr(a) + sqr(v12));
				return mult*sqrt(b-D);
				//T lambda_1= b+D;
				//T lambda_2= b-D;
				//return lambda_2;
			}
		};
		/*!
		* \brief  DiffEigenVal  PixByPix3to1.
		*    \f$o_{11},\ o_{12},\ o_{22}\f$    
		*     
		* \f$ M=\biggl\Vert\begin{array}{cc} <g_x^2>  &  <g_x g_y> \\ <g_x g_y>  & <g_y^2>  \end{array}\biggr\Vert=
		* \biggl\Vert\begin{array}{cc} o_{11} &  o_{12} \\ o_{12}  & o_{22}  \end{array}\biggr\Vert\f$ \n 
		*	,    
		*  ,     (    ()): 
		* \f$ result=\lambda_1-\lambda_2 \f$,\n
		*  \f$ det\biggl\Vert \begin{array}{cc} o_{11}-\lambda & o_{12} \\ o_{12} & o_{22}-\lambda_2 \end{array} \biggl\Vert=0 \f$ ,
		* .. \f$ \lambda^2-\lambda(o_{11}+o_{22})+o_{11}o_{22}-o_{12}^2=0 \f$,
		*  \f$ D/4=((o_{11}+o_{22})/2)^2-(o_{11}o_{22}-o_{12}^2)=((o_{11}-o_{22})/2)^2+o_{12}^2 \f$ ,
		*  \f$ \lambda_1=\frac{o_{11}+o_{22}}{2}+\sqrt{\frac{D}{4}}\f$ ,
		* \f$ \lambda_2=\frac{o_{11}+o_{22}}{2}-\sqrt{\frac{D}{4}}\f$ 
		*******************************************************************************************/
		template<typename T>
		class DiffEigenVal
		{
			T mult;
		public:
			DiffEigenVal(T sigma=1.0f) : mult(2.9f*sigma)
			{
			}
			T operator()(T v11,T v12,T v22) const 
			{
				T a=(v11-v22)/2;
				//T b=(v11+v22)/2;
				T D = sqrt(sqr(a) + sqr(v12));
				return mult*sqrt(2*D);
				//T lambda_1= b+D;
				//T lambda_2= b-D;
				//return lambda_2;
			}
		};
		/*!
		* \brief  StatDiff  PixByPix3to1.
		*    \f$o_{11},\ o_{12},\ o_{22}\f$   
		*  ,      ,    
		*      ,
		* \ref Lit_Pratt
		*******************************************************************************************/
		template<typename T>
		class StatDiff
		{
			float A;
			float sigma_d;
			float m_d;
			float alpha;
		public:
			StatDiff(float amplification=6.0f,float sigma_desired=85.0f,float mean_desired=128.0f,float phone_to_edges=0.8f) 
				: A(amplification),sigma_d(sigma_desired),m_d(mean_desired),alpha(phone_to_edges)
			{
			}
			T operator()(T f,T f_mean,T f_sigma) const 
			{
				return (f-f_mean)*(A*sigma_d/(A*f_sigma+sigma_d))+alpha*m_d+(1-alpha)*f_mean;
			}
		};
		/*!
		* \brief  DeNoise  PixByPix3to1.
		*    \f$o_{11},\ o_{12},\ o_{22}\f$   
		*   ,      ,    
		*  :\n
		* \f$result=\biggl\lbrace\begin{array}{cc} \langle f\rangle  & ,  \quad if | f-\langle f\rangle | >k*\sigma_f  \\ f & , \quad else  \end{array}\f$
		* 
		*******************************************************************************************/
		template<typename T>
		class DeNoise
		{
			float k;
		public:
			DeNoise(float in_k_sigmas=2.0f) 
				: k(in_k_sigmas)
			{
			}
			T operator()(T f,T f_mean,T f_sigma) const 
			{
				return T((abs(f-f_mean)>k*f_sigma )?f_mean:f);
			}
		};
	}//namespace F31

	///////////////////////////////////////////////////////////////////////////////////////////////////
	/// @brief    template<typename T,typename T1> T operator()(T1 value1,T1 value2,T1 value3,T1 value4,T1 value5)
	///
	///   FilterPixByPix5to1, 
	///     Vtl::ImAlg::Tools     .
	namespace F51
	{
		/*!
		* \brief  HarrisLaplace  PixByPix5to1.
		*
		*    \f$o_{11},\ o_{12},\ o_{22}\f$    
		*******************************************************************************************/
		template<typename T>
		class HarrisLaplace
		{
			T k;
			T mult;
		public:
			HarrisLaplace(T sigma=1.0f,T coef=0.04f) : mult(6.3f*sigma),k(coef) 
			{
				//mult=sqrt(2.0f*3.1415926f)*sigma;
			}
			T operator()(T v11,T v12,T v22,T d2x,T d2y) const 
			{
				T h=v11*v22-v12*v12-k*sqr(v11+v22);
				T harris=mult*sqrt(sqrt(h));
				if(harris>30)
					return harris;
				else
					return (d2x+d2y)/2;
			}
		};
	}//namespace F51
	///////////////////////////////////////////////////////////////////////////////////////////////////
	/// @brief    template<typename T,typename Iterator> void operator()(T v1,T v2,Iterator o11,Iterator o12,Iterator o22) const;
	///
	///   Iterator     Iterator arg; *arg=return_value; 
	///   FilterPixByPix2to3, 
	///     Vtl::ImAlg::Tools     .
	namespace F23
	{
		///////////////////////////////////////////////////////////////////////////
		/// \brief  CorrMatrix  PixByPix2to3.
		/// @brief    \f$v_1\f$ e \f$v_2\f$   3 
		///     : \n
		/// \li \f$ o_{11}=v_1^2\f$\n
		/// \li \f$o_{12}=v_1 v_2\f$\n
		/// \li \f$o_{11}=v_1^2\f$\n
		/// ..    
		/// \f$Corr=\biggl\Vert\begin{array}{cc} v_x^2  &  v_x v_y \\ v_x v_y  & v_y^2  \end{array}\biggr\Vert
		/// =\biggl\Vert\begin{array}{cc} o_{11} &  o_{12} \\ o_{12}  & o_{22}  \end{array}\biggr\Vert\f$\n 
		///	,    
		///  . \n
		///      ,
		///         ,   
		///   @see class Average.
		///////////////////////////////////////////////////////////////////////////
		template<typename T,typename Iter=T&>
		class CorrMatrix
		{
		public:
			CorrMatrix() {}
			void operator()(T v1,T v2,Iter o11,Iter o12,Iter o22) const {o11=v1*v1; o12=v1*v2; o22=v2*v2;}
		};
		template<typename T,typename Iter=T&>
		class CorrMatrixAdd
		{
		public:
			CorrMatrixAdd() {}
			void operator()(T v1,T v2,Iter o11,Iter o12,Iter o22) const {o11+=v1*v1; o12+=v1*v2; o22+=v2*v2;}
		};

	}//namespace F23
	///////////////////////////////////////////////////////////////////////////////////////////////////
	/// @brief    template<typename T,typename Iterator> void operator()(T v11,T v12,T v22,Iterator out_1,Iterator out_2,Iterator out_3) const;
	///
	///   Iterator     Iterator arg; *arg=return_value; 
	///   FilterPixByPix3to3, 
	///     Vtl::ImAlg::Tools     .
	namespace F33
	{

		/*!
		* \brief  DiZenzo  PixByPix3to3.
		*    \f$o_{11},\ o_{12},\ o_{22}\f$    
		*     
		* \f$ M=\biggl\Vert\begin{array}{cc} <g_x^2>  &  <g_x g_y> \\ <g_x g_y>  & <g_y^2>  \end{array}\biggr\Vert=
		* \biggl\Vert\begin{array}{cc} o_{11} &  o_{12} \\ o_{12}  & o_{22}  \end{array}\biggr\Vert\f$ \n 
		*	,    
		*  ,     ( )  
		*    (    ). 
		*     .,   2  3 - x  y   .. . 
		* 
		*******************************************************************************************/
		template<typename T,typename Iter=T&>
		class DiZenzo
		{
		public:
			DiZenzo() 
			{
			}
			void operator()(T v11,T v12,T v22,Iter lamda_1,Iter eigv_x,Iter eigv_y) const 
			{
				T a=(v11-v22)/2;
				T b=(v11+v22)/2;
				T D = sqrt(sqr(a) + sqr(v12));
				//T lambda_1 ^2 = b+D;
				//T lambda_2 ^2 = b-D;
				a=b+D;
				// lamda_1=a;  -  16.10.2010 : /3 ..  3         sqrt() -    . ,   .
				lamda_1=sqrt(T(a/3));
				v11=v11-a;
				if(a>0)
				{
					if(v11 > v12)
					{
						b=v11;
						a=-v12;
					}
					else
					{
						a=-v12;
						b=v11;
					}
					//D=sqrt(a*a+b*b);
					eigv_x=a;///D;
					eigv_y=b;///D;
				}
				else
				{
					eigv_x=1;
					eigv_y=0;
				}
			}
		};
	}//namespace F33
	///////////////////////////////////////////////////////////////////////////
	/// @brief ,           ((F)unctor receiving (P)ointer to (I)mage data and using local (E)nvironment data). 
	///            (Plf).
	///
	namespace Fpie
	{
		
		/*!
		* \brief  HarrisNMS  EnvProc1to1.
		*    .  8  , 
		*             
		*    ,    -   0.
		*******************************************************************************************/
		template<typename T> 
		class HarrisNMS
		{
			int nls;
		public:
			/// !!! MUST be called by filter BEFORE using 'operator()', by default
			/// next line shist parameter set to value which probably most quick lead 
			/// a 'Invalid memory access' error (for bug search convenience).
			void set_IterNextLineShift(int modul_nls) { nls=modul_nls;}
			Extent ext() const {return Extent(1,1,1,1); }
			HarrisNMS() {nls=0x5FFFFFFF;}
			T operator()(const T* input)
			{
				T v=input[0];
				const T* inp1=input-1;
				const T* inp0=inp1-nls;
				const T* inp2=inp1+nls;
				if( v <= inp1[0] || v <= inp1[2] 
					|| v <= inp0[0] || v <= inp0[1] || v <= inp0[2]
					|| v <= inp2[0] || v <= inp2[1] || v <= inp2[2]
					)
					return 0;
				return v;			
			}
		};
		/*!
		* \brief  HarrisNMS_Smooth  EnvProc1to1  HarrisNMS, 
		*   1/10        33.
		*    .  8  , 
		*             
		*    ,    -   0.
		*******************************************************************************************/
		template<typename T> 
		class HarrisNMS_Smooth
		{
			int nls;
		public:
			/// !!! MUST be called by filter BEFORE using 'operator()', by default
			/// next line shist parameter set to value which probably most quick lead 
			/// a 'Invalid memory access' error (for bug search convenience).
			void set_IterNextLineShift(int modul_nls) { nls=modul_nls;}
			Extent ext() const {return Extent(1,1,1,1); }
			HarrisNMS_Smooth() {nls=0x5FFFFFFF;}
			T operator()(const T* input)
			{
				T v=input[0];
				const T* inp1=input-1;
				const T* inp0=inp1-nls;
				const T* inp2=inp1+nls;
				if( v <= inp1[0] || v <= inp1[2] 
					|| v <= inp0[0] || v <= inp0[1] || v <= inp0[2]
					|| v <= inp2[0] || v <= inp2[1] || v <= inp2[2]
					)
					return 0;
				return (2*v+inp1[0]+inp1[2]+ inp0[0]+inp0[1]+inp0[2]+ inp2[0]+inp2[1]+inp2[2])/T(10);			
			}
		};
		/*!
		* \brief  Himg2  EnvProc1to2.
		*   -    H-image \ref Lit_HImg.
		*******************************************************************************************/
		template<typename T> 
		class Himg2
		{
			int r;
			int nls;
			double mult;
		public:
			/// !!! MUST be called by filter BEFORE using 'operator()', by default
			/// next line shist parameter set to value which probably most quick lead 
			/// a 'Invalid memory access' error (for bug search convenience).
			void set_IterNextLineShift(int modul_nls) { nls=modul_nls;}
			Extent ext() const {return Extent(r); }
			Himg2(int radius_of_square=5):r(ABS(radius_of_square)) 
			{
				nls=0x5FFFFFFF;
				mult=1./(Vtl::Plf::F11::sqr(2*r+1)-1);
			}
			void operator()(const T* input,T& output_x,T& output_y)
			{
				T v0=input[0];
				const T* p=input-r*nls;
				double tx=0;
				double ty=0;
				for(int y=-r; y<=r ;++y,p+=nls)
				{
					for(int x=-r; x<=r ;++x)
					{
						T t=p[x]-v0;
						if(x!=0 || y!=0)
						{
							double d=sqrt(double(x*x+y*y)); 
							tx+=t*x/d;
							ty+=t*y/d;
						}
					}
				}
				output_x=T(mult*tx);
				output_y=T(mult*ty);
			}
		};
		/*!
		* \brief  Himg  EnvProc1to1.
		*    H-image \ref Lit_HImg.
		*******************************************************************************************/
		template<typename T> 
		class Himg
		{
			Himg2<T> hi;
		public:
			/// !!! MUST be called by filter BEFORE using 'operator()', by default
			/// next line shist parameter set to value which probably most quick lead 
			/// a 'Invalid memory access' error (for bug search convenience).
			void set_IterNextLineShift(int modul_nls) { hi.set_IterNextLineShift(modul_nls);}
			Extent ext() const {return hi.ext(); }
			Himg(int radius_of_square=5):hi(radius_of_square) {}
			T operator()(const T* input)
			{
				T v0=input[0];
				T gx;
				T gy;
				hi(input,gx,gy);
				return sqrt(gx*gx+gy*gy);			
			}
		};
		/*!
		* \brief  BiLateralSmooth  EnvProc1to1.
		* BiLateral - .  !
		*******************************************************************************************/
		template<typename T> 
		class BiLateralSmooth
		{
			int r;
			float sigma_22;
			int nls;
			//double mult;
			float eps_22;
		public:
			/// !!! MUST be called by filter BEFORE using 'operator()', by default
			/// next line shist parameter set to value which probably most quick lead 
			/// a 'Invalid memory access' error (for bug search convenience).
			void set_IterNextLineShift(int modul_nls) { nls=modul_nls;}
			Extent ext() const {return Extent(r); }
			BiLateralSmooth(float epsilon=20.0f,float sigma=2.0f,float env_sigmas=2.0f)
				:r(int(sigma*env_sigmas+0.5)),sigma_22(2*sigma*sigma),eps_22(2*epsilon*epsilon) 
			{
				nls=0x5FFFFFFF;
				//mult=1./(Vtl::Plf::F::sqr(2*r+1)-1);
			}
			T operator()(const T* input)
			{
				T v0=input[0];
				const T* p=input-r*nls;
				double sum=0;
				double sum_k=0;
				for(int y=-r; y<=r ;++y,p+=nls)
				{
					for(int x=-r; x<=r ;++x)
					{
						T k=fun(x,y,p[x],v0);
						sum_k+=k;
						sum+=p[x]*k;
					}
				}
				return T(sum/sum_k);
			}
		private:
			T fun(int x,int y,T fx,T fy)
			{
				return exp(-(x*x+y*y)/sigma_22-sqr(fx-fy)/eps_22);
			}
		};
		///////////////////////////////////////////////////////////////////////////
		// class CannyNMS
		/// @brief   EnvProc3to1,    
		///      (Non Maximum Suppression). 
		///
		///   33.  ,   (gx,gy)
		///    (v2-     v1-    
		///  )  8 .      - 
		///    8  .     
		///    .      v1  v2.
		///     v>v1 && v>v2 -    ,  
		///   ,  -  0 -     .\n
		/// threshold - , ,   modul[0]<threshold    
		///  . threshold     0,  
		///    0.\n
		///    
		///////////////////////////////////////////////////////////////////////////
		template<typename T> 
		class CannyNMS
		{
			T thr;
			int nls;
		public:
			/// !!! MUST be called by filter BEFORE using 'operator()', by default
			/// next line shist parameter set to value which probably most quick lead 
			/// a 'Invalid memory access' error (for bug search convenience).
			void set_IterNextLineShift(int modul_nls) { nls=modul_nls;}
			Extent ext() const {return Extent(1,1,1,1); }
			CannyNMS() {nls=0x5FFFFFFF;}
			CannyNMS(T threshold):thr(threshold) {nls=0x5FFFFFFF;}
			T operator()(const T* modul,T gx,T gy)
			{
				if(modul[0]>=thr)
				{
					T v1;
					T v2;
					if(ABS(gx) < ABS(gy))
					{//     .   -    
						T alpha=gx/gy;
						if(alpha<0)
						{// (gx< gy) && ((gx<0 gy >0 )|| (gx>0 gy < 0 ) ) 
							//    , '+' -  , 
							// alpha=0 -  (|) , alpha=1 -  (/)
							//
							//    +-+=+ 
							//    | | |
							//    +-+-+ 
							//    | | |
							//    +=+-+ 
							//
							v1=(1+alpha)*modul[-nls]-alpha*modul[-nls+1];
							v2=(1+alpha)*modul[+nls]-alpha*modul[+nls-1];
						}
						else
						{// (gx< gy) && ( (gx>=0 gy >0) || ( gx<=0 gy < 0 ) ) 
							//    , '+' -  , 
							// alpha=0 -  (|) , alpha=1 -  (\)
							//
							//    +=+-+ 
							//    | | |
							//    +-+-+ 
							//    | | |
							//    +-+=+ 
							//
							v1=(1-alpha)*modul[-nls]+alpha*modul[-nls-1];
							v2=(1-alpha)*modul[+nls]+alpha*modul[+nls+1];
						}
					}
					else
					{//     .  -    
						T alpha=gy/gx;
						if(alpha<0)
						{// (gx >= gy) && ( (gx>0 , gy<=0) || ( gx <0 gy >= 0 ) ) 
							//   (#)  , '+' -  , 
							// alpha=0 -  (-) , alpha=1 -  (/)
							//
							//
							//    +-+-+ 
							//    | | #
							//    +-+-+ 
							//    # | |
							//    +-+-+ 
							//
							v1=(1+alpha)*modul[-1]-alpha*modul[-1+nls];
							v2=(1+alpha)*modul[+1]-alpha*modul[+1-nls];
						}
						else
						{// (gx >= gy) && ( (gx>0 , gy>=0) || ( gx <0 gy <= 0 ) ) 
							//   (#)  , '+' -  , 
							// alpha=0 -  (-) , alpha=1 -  (\)
							//
							//    +-+-+ 
							//    # | |
							//    +-+-+ 
							//    | | #
							//    +-+-+ 
							//
							v1=(1-alpha)*modul[-1]+alpha*modul[-1-nls];
							v2=(1-alpha)*modul[+1]+alpha*modul[+1+nls];
						}
					}
					T v0=modul[0];
					if( (v0 >= v1) && (v0 >= v2) )
						return v0;
				}
				return 0;
			}
		};
		///////////////////////////////////////////////////////////////////////////
		// class CannyNMS_2
		/// @brief   EnvProc3to1,    
		///      (Non Maximum Suppression). 
		///
		///   55.    33.
		///  ,   (gx,gy)
		///    (v2-     v1-    
		///  )  8 .      - 
		///    8  .     
		///    .      v1  v2.
		///     v>v1 && v>v2 -    ,  
		///   ,  -  0 -     .
		///       v1  v2    v,     
		///      (   55)    
		/// . \n
		/// threshold - , ,   modul[0]<threshold    
		///  . threshold     0,  
		///    0.\n
		/// epsilon - (),       .\n
		///    
		///       CannyNMS  ,  .
		///  -           
		///   2     CannyNMS,   1 .
		///////////////////////////////////////////////////////////////////////////
		template<typename T> 
		class CannyNMS_2
		{
			T thr;
			T eps;
			int nls;
		public:
			/// !!! MUST be called by filter BEFORE using 'operator()', by default
			/// next line shist parameter set to value which probably most quick lead 
			/// a 'Invalid memory access' error (for bug search convenience).
			void set_IterNextLineShift(int modul_nls) { nls=modul_nls;}
			Extent ext() const {return Extent(2,2,2,2); }
			CannyNMS_2() {nls=0x5FFFFFFF;} 
			CannyNMS_2(T threshold,T epsilon=0.1f):thr(threshold),eps(epsilon) {nls=0x5FFFFFFF;}
			T operator()(const T* modul,T gx,T gy)
			{
				if(modul[0]>=thr)
				{
					T v1;
					T v2;
					T v3;
					T v4;
					if(ABS(gx) < ABS(gy))
					{//     .   -    
						T alpha=gx/gy;
						if(alpha<0)
						{// (gx< gy) && ((gx<0 gy >0 )|| (gx>0 gy < 0 ) ) 
							//    , '+' -  , 
							// alpha=0 -  (|) , alpha=1 -  (/)
							//
							//    +-+=+ 
							//    | | |
							//    +-+-+ 
							//    | | |
							//    +=+-+ 
							//
							v1=(1+alpha)*modul[-nls]-alpha*modul[-nls+1];
							v2=(1+alpha)*modul[+nls]-alpha*modul[+nls-1];
							if(alpha>-0.5)
							{
								v3=(1+alpha)*modul[-2*nls]-alpha*modul[-2*nls+1];
								v4=(1+alpha)*modul[+2*nls]-alpha*modul[+2*nls-1];
							}
							else
							{
								v3=(1+alpha)*modul[-2*nls+1]-alpha*modul[-2*nls+2];
								v4=(1+alpha)*modul[+2*nls-1]-alpha*modul[+2*nls-2];
							}
						}
						else
						{// (gx< gy) && ( (gx>=0 gy >0) || ( gx<=0 gy < 0 ) ) 
							//    , '+' -  , 
							// alpha=0 -  (|) , alpha=1 -  (\)
							//
							//    +=+-+ 
							//    | | |
							//    +-+-+ 
							//    | | |
							//    +-+=+ 
							//
							v1=(1-alpha)*modul[-nls]+alpha*modul[-nls-1];
							v2=(1-alpha)*modul[+nls]+alpha*modul[+nls+1];
							if(alpha<0.5)
							{
								v3=(1-alpha)*modul[-2*nls]+alpha*modul[-2*nls-1];
								v4=(1-alpha)*modul[+2*nls]+alpha*modul[+2*nls+1];
							}
							else
							{
								v3=(1-alpha)*modul[-2*nls-1]+alpha*modul[-2*nls-2];
								v4=(1-alpha)*modul[+2*nls+1]+alpha*modul[+2*nls+2];
							}
						}
					}
					else
					{//     .  -    
						T alpha=gy/gx;
						if(alpha<0)
						{// (gx >= gy) && ( (gx>0 , gy<=0) || ( gx <0 gy >= 0 ) ) 
							//   (#)  , '+' -  , 
							// alpha=0 -  (-) , alpha=1 -  (/)
							//
							//
							//    +-+-+ 
							//    | | #
							//    +-+-+ 
							//    # | |
							//    +-+-+ 
							//
							v1=(1+alpha)*modul[-1]-alpha*modul[-1+nls];
							v2=(1+alpha)*modul[+1]-alpha*modul[+1-nls];
							if(alpha>-0.5)
							{
								v3=(1+alpha)*modul[-2]-alpha*modul[-2+nls];
								v4=(1+alpha)*modul[+2]-alpha*modul[+2-nls];
							}
							else
							{
								v3=(1+alpha)*modul[-2+nls]-alpha*modul[-2+2*nls];
								v4=(1+alpha)*modul[+2-nls]-alpha*modul[+2-2*nls];
							}
						}
						else
						{// (gx >= gy) && ( (gx>0 , gy>=0) || ( gx <0 gy <= 0 ) ) 
							//   (#)  , '+' -  , 
							// alpha=0 -  (-) , alpha=1 -  (\)
							//
							//    +-+-+ 
							//    # | |
							//    +-+-+ 
							//    | | #
							//    +-+-+ 
							//
							v1=(1-alpha)*modul[-1]+alpha*modul[-1-nls];
							v2=(1-alpha)*modul[+1]+alpha*modul[+1+nls];
							if(alpha<0.5)
							{
								v3=(1-alpha)*modul[-2]+alpha*modul[-2-nls];
								v4=(1-alpha)*modul[+2]+alpha*modul[+2+nls];
							}
							else
							{
								v3=(1-alpha)*modul[-2-nls]+alpha*modul[-2-2*nls];
								v4=(1-alpha)*modul[+2+nls]+alpha*modul[+2+2*nls];
							}
						}
					}
					T v0=modul[0];
					if(  
						( ((v0-v1)>eps) || ((ABS(v0-v1)<=2*eps) && ((v0-v3)>4*eps) ) )
						&&	( ((v0-v2)>eps) || ((ABS(v0-v2)<=2*eps) && ((v0-v4)>4*eps) ) )
						)
						return v0;
				}
				return 0;
			}
		};
	}//namespace F
/** @} */ // end of ElementaryFiltersDoxGroup  
} //namespace Plf 
} //namespace Vtl

/////////////////////////////////////////////////////////////////////////////
#endif //_PLF_FUNCTORS_H 
