#ifndef _KERNELS_1D_H 
#define _KERNELS_1D_H 
/////////////////////////////////////////////////////////////////////////////
#include  "convolve.h" 
namespace Vtl
{
// @brief     plf_filters.h,   
namespace Plf 
{
/** 
 *  @brief   ,       
 *            //  
 *          .
 *
 *  @defgroup ConvolutionKernelsDoxGroup   Convolution kernels. 
 *  @ingroup PipeLineFiltersDoxGroup  
 *
 *  @{
 */

/** 
 *  @brief   ,       
 *            //  
 *          . \see Vtl::Plf::Kernel::F
 *
 *         :
 *            -   ( )    
 *             .      , 
 *                 \see Vtl::Plf::Kernel::F
 *            <b>   </b>     , 
 *          , ,    <b></b>.     
 *            -   <b>       
 *            </b>.        
 *                 (, 
 *              ).     Vtl::Plf::Kernel::F  
 *         ,       . 
 *         
 **********************************************************************************************/
namespace Kernel
{
	enum Type
	{
		KT_Conv, //!<    (    ,    ).
		KT_Corr,  //!<   ,  ,    , 
		         //!<   -      , ,    
				 //!<    .
		KT_Win   //!<     (  KT_Corr  ,  1   ).
	};
   /*!
	* 
	* \brief      // .  
	*     . !     
	*    ,      , ,    
	*   .       void _Init(int left,int right)
	*    ,   ,    \c  -    
	*    ,  ,       
	*   .   Base    left() b right()    
	*  "" , <b>    \c </b>, ..
	*       -   left  right,   
	*    \b right  \b left.
	* 
	**********************************************************************************/
	template<typename T,Kernel::Type kernel_type_=KT_Conv,bool double_kernel_=false>
	class Base
	{
	public:
		typedef T value_type;
		typedef T& reference_type;
		typedef const T& const_reference_type;
		typedef T* pointer_type;
		typedef const T* const_pointer_type;
		typedef T* iterator;
		typedef const T* const_iterator;
		enum { kernel_type=kernel_type_ };
		enum { double_kernel=double_kernel_?1:0 };
	private:
		std::vector<T> data;
		iterator ptr_left;
		iterator ptr_center;
		iterator ptr_after_right;
		int _size;
		int _left;
		int _right;
		std::string filtername;
	private:
		Base(); //   
		Base(const Base&); //  
		Base& operator=(const Base&); //  
	protected:
		Base(std::string name) : filtername(name) {} //    ,    
		//! \brief    left+1+right.  left=right=0,    1. 
		//! ""    ,  1   left+1+right.
		void _Init(int left,int right)
		{
			assert( left >=0);
			_left=left;
			_right=(right<0)?left:right;
			if(kernel_type==KT_Conv)
			{
				std::swap(_left,_right);
			}
			_size=left+right+1;
			data.resize(Vtl::Alg::mult_if<double_kernel>(_size,2));
			ptr_left=&(data[0]);
			ptr_center=&(data[left]);
			ptr_after_right=ptr_left+_size;
		}
		void _Init_Last_step()
		{
			_double_ker(Loki::Int2Type<double_kernel>() );
		}
	private:
		void _double_ker(Loki::Int2Type<1> ) {::memcpy(ptr_after_right,ptr_left,_size*sizeof(data[0]));}
		void _double_ker(Loki::Int2Type<0> ) {}
	public:
		Type type() const { return kernel_type;}
		virtual ~Base() {}
		///     
		const_iterator center() const {return ptr_center;}
		///    
		const_iterator begin() const {return (const_iterator)ptr_left;}
		///       (),   STL.
		const_iterator end() const {return ptr_after_right;}
		/// \brief  ,   -    , 0 -  
		///  ,    i: -left <= i <= right
		const T* Center() const {return ptr_center;}
		/// \brief  ,   -    , 0 -  
		///   ,    i: 0<= i < size
		const T* Left() const {return ptr_left;}
		/// \brief  ,   -    , 0 -  
		///    (    !),    i: -(size-1)<= i <= 0
		const T* Right() const {return ptr_left+_size-1;}
		///  ,   , size == left+1+right
		int size() const {return _size;}
		/// ! left>=0 
		int left() const {return _left;}
		int right() const {return _right;}
		///   ,    ,      
		///     ,     ,  
		///     ,       .  ,   
		///    (. Vtl::Plf::Kernel::FKern) -   ,   .
		std::string Name() const {return filtername;}
		/*! \brief   ,   Matlab.      
		 *       Matlab    load().
		 *
		 * :\n
		 * ++:\n
		 * \code
		 * template<class K>
		 * void save(const K* kernel)
		 * {
		 * 	std::string name=std::string("kernel_")+kernel->Name()+".txt";
		 * 	std::string s=kernel->SaveForMathLab();
		 * 	std::ofstream os;
		 * 	os.open(name.c_str(),std::ios_base::out | std::ios_base::trunc );
		 * 		os << s;
		 * 	os.flush();
		 * 	os.close();
		 * }
		 * int main(int argc, const char*argv[])
		 * {
		 *		typedef Vtl::Plf::Kernel::F::Blackman<float,0> FF;
		 *		FF func(19);
		 *		Vtl::Plf::Kernel::FKern<float,FF> ker(func);
		 *		save(&ker);
		 * }
		 * \endcode
		 * Matlab:\n
		 * \code
		 * >> a=load('kernel_Blackman.txt');
		 * >> x=a(:,1);
		 * >> y=a(:,2);
		 * >> plot(x,y);
		 * \endcode
		 *
		 ***************************************************************************/
		std::string SaveForMathLab() const
		{
			std::ostringstream os;
			int i=-left();
			for(const_iterator it=begin(); it!=end(); it++,i++)
				os<< i << " " <<*it <<std::endl;
			os<<std::endl;
			return os.str();
		}
	protected:
		iterator center_internal() {return ptr_center;}
		iterator begin_internal() {return ptr_left;}
		iterator end_internal() {return ptr_after_right;}
	protected:
		/*!
		*	@brief     (   
		*      ).    
		*         .  
		*       ,     
		*      <tt>    </tt>. 
		*      0   , .. 
		*     .     (KT_Win)     
		*    ,        
		*      normalize_window().
		*
		**************************************************************/
		template<unsigned N>	void normalize()
		{
			if( kernel_type==KT_Win )
				return normalize_window();
			const int cof=(Vtl::Alg::isOdd<N>::value)?-Vtl::Alg::Factorial<N>::value:Vtl::Alg::Factorial<N>::value;
			T sum=0;
			const_iterator d=begin();
			for(int i=0; i<size(); i++)
				sum+=Vtl::Alg::x_to_N<N>::value(i-left())*(*d++); 
			if(sum==0)
				throw "Kernel::Base::normalize(Vtl::Types::Int2Type<N>): sum==0 !";
			if( (kernel_type==KT_Conv) && (Vtl::Alg::isOdd<N>::value) )
				sum=-sum;
			multBy(cof/sum);
		}
		virtual void normalize_window()
		{
		}
	private:
		void multBy(T val)
		{
			for(iterator d=begin_internal(); d!=end(); ++d)
				*d=(*d)*val;
		}
	};
	/*
	 :
	:
	Functor f(params);
	KernelWin ker(f,size,center,metod_to_padd_extrapolate_if_need);
	ker -   size.
	*/


	/// ,  .
	template<typename T,Kernel::Type kernel_type_=KT_Conv,bool double_kernel_=false>
	class Table
		:public Base<T,kernel_type_,double_kernel_>
	{
	public:
		Table(const char *name) : Base<T,kernel_type_,double_kernel_>(name) {}
		Table(const char *name,const T* ker,int size,int left)
			: Base<T,kernel_type_,double_kernel_>(name)
		{	
			Init(ker,size,left);
		}
		virtual ~Table() {}
		/*!
		* \brief     <b>  </b>,     .
		* \param ker  -  ,     ().
		* \param size -  
		* \param left -      ,  ,   -1 0 1    "0"
		*               size=3, left =1.
		*
		*/
		void Init(const T* ker,int size,int left)
		{
			_Init(left,size-left-1);
			std::copy(ker,ker+size,begin_internal());
			_Init_Last_step();
		}
	};


	/**
	* \brief class FKern,        Vtl::Plf::Kern::F, (.  
	*   ). 
	*
	*      ,   
	* ,     Vtl::Plf::Kernel::F::Traits, Vtl::Plf::Kernel::F::Traits_common,
	*       ,     .
	* 
	*         , 
	*          ,  
	*             ,  . .
	*     (windowing)   (appodization)  , ,  
	*   ()   ,      1.
	* 
	* \htmlinclude kernels_1d_conv_gauss.htm
	*	
	* \htmlinclude kernels_1d_conv_gauss_1_deriv.htm
    * 
	* \htmlinclude KernelsNorm.htm 	
    * 
	**********************************************************************************/
	template<typename T,Kernel::Type kernel_type_=KT_Conv,bool double_kernel_=false>
	class FKern 
		:public Base<T,kernel_type_,double_kernel_>
	{
	public:
		FKern() : Base<T,kernel_type_,double_kernel_>("KernelNotInitializedByFunctor") {}
		template<class FF>
		FKern(FF functor)
			: Base<T,kernel_type_,double_kernel_>(FF::Name())
		{	
			Init(functor);
		}
		template<class FF>
		FKern(FF functor, size_t wnd_size,T center_pos=T(-1),T fill_by=T(0))
			: Base<T,kernel_type_,double_kernel_>(FF::Name())
		{	
			Init(functor,wnd_size,center_pos,fill_by);
		}
		virtual ~FKern() {}
		/*!
		* \brief     <b>  </b>,     .
		* \param functor -   .,       see Vtl::Plf::Kernel::FF
		*
		*/
		template<class FF>
		void Init(FF functor)
		{
			LOKI_STATIC_CHECK(((kernel_type_==Kernel::KT_Conv)||(kernel_type_==Kernel::KT_Corr)),This_Init_function_form_can_be_used_only_for_kernels_with_type_KT_Conv_or_KT_Corr);
//			func=functor;//  ?
			//Calculate kernel size and center
			_Init_conv( functor );
			_Init_Last_step();
		}
		/*!
		* \brief     <b> </b>,     .
		* \param functor    -  ,        see Vtl::ImAlg::WndFun
		* \param wnd_size   -   
		* \param center_pos -     0..wnd_size-1    (  ). 
		* \param fill_by    -     ,       (wnd_size-1)/2.0
		*                               ,
		*                       -  -    .
		*
		*/
		template<class FF>
		void Init(FF functor, size_t wnd_size,T center_pos,T fill_by=T(0))
		{
			LOKI_STATIC_CHECK((kernel_type_==Kernel::KT_Win),This_Init_function_form_can_be_used_only_for_kernels_with_type_KT_Win);
			//Calculate kernel size and center
			_Init_win( functor,wnd_size,center_pos,fill_by);
			_Init_Last_step();
		}
	private:
		template<class FF>
		void _Init_conv(FF functor)
		{
			//Calculate kernel size and center
			//: functor.left();functor.right();functor.center()   x  
			//   /    .
			//   left, right       (left+1+right) ,
			//     ,       .
			//   left, right   .
			int left=int(functor.center()-functor.left()+0.5);
			int right=int(functor.right()-functor.center()+0.5);
			if(FF::finit_support_left)
			{
				if((functor.center()-left)< functor.left())
					--left;
			}
			if(FF::finit_support_right)
			{
				if((functor.center()+right)> functor.right())
					--right;
			}
			//  ,  .
			FF::value_type vc=std::abs(functor(functor.center()));
			{
				FF::arg_type ofs=functor.center()-left; //          size=left+1+right;
				for(int i=0; i<left+right+1; ++i)
					vc=std::max(vc,std::abs(functor(i+ofs)));
			}
			const FF::value_type near_zero=vc*NumAlg::RelerrEps<FF::value_type>();
			if(std::abs(functor(functor.center()-left)) <= near_zero )
				--left;
			if(std::abs(functor(functor.center()+right)) <= near_zero)
				--right;
			FF::arg_type ofs=functor.center()-left; //          size=left+1+right;
			_Init(left,right);//size=left+right+1 !!!
			build(functor,ofs);
			//build(functor,functor.center);
		}
		template<class FF>
		void _Init_win(FF functor,size_t wnd_size,T center_pos,T fill_by)
		{
			//Calculate kernel size and center
			//: functor.left();functor.right();functor.center()   x  
			//   /    .
			//   left, right       (left+1+right) ,
			//     ,       .
			//   left, right   .

			FF::arg_type ofs=functor.center()-center_pos; //      center_pos (   0, center_pos   );
			int i_left=int(std::ceil(functor.left()-ofs));
			if(i_left>0)
			{
				//there is i_left entries to fill by default
			}
			int i_right=int(std::floor(functor.right()-ofs-0.5));
			if(i_right<((int(wnd_size)-1)))
			{
				//there is (wnd_size-i_right-1) entries to fill by default
			}
			//if(i_left<0 || i_right>=wnd_size-1 ) - error!!!
			_Init(0,int(wnd_size)-1);//size=left+right+1 !!!
			build(functor,ofs);
			//build(functor,functor.center);
		}
		/*!
		*	@brief   ( )   .
		*    ! .  template<unsigned N>	void normalize().  
		*       void _Init(int left,int right)
		*
		**************************************************************/
		template<class FF>
		void build(FF func,typename FF::arg_type ofs)
		{
			build(func,ofs,Loki::Int2Type<kernel_type_>() );
		}
		template<class FF>
		void build(FF func,typename FF::arg_type ofs, Loki::Int2Type<KT_Conv> )
		{
			iterator d=begin_internal();
			ofs=FF::arg_type(ofs+size()-1);
			for(int i=0; i<Base::size(); i++)
				*d++=func(ofs-i);//(size-1-i)-ofs)=(size-1-ofs)-i
			normalize<FF::conv_norm_alg>();
		}
		template<class FF>
		void build(FF func,typename FF::arg_type ofs, Loki::Int2Type<KT_Corr> )
		{
			iterator d=begin_internal();
			for(int i=0; i<Base::size(); i++)
				*d++=func(FF::arg_type(i+ofs));
			normalize<FF::conv_norm_alg>();
		}
		template<class FF>
		void build(FF func,typename FF::arg_type ofs, Loki::Int2Type<KT_Win> )
		{
			iterator d=begin_internal();
			for(int i=0; i<Base::size(); i++)
			{
				*d++=func(FF::arg_type(i+ofs));
			}
			//normalize<FF::conv_norm_alg>();
		}
	};
	/*!
	 * \brief  ,   in ,   "in place",
	 *      .        
	 *   bias,     mult,      
	 *    . 
	 * \see IntegrateWindowedData(), IntegrateKer().
	 **************************************************************************************/
	template<typename T,typename iterator_in,typename iterator_ker>
	void WindowData(iterator_in in,iterator_ker begin,iterator_ker end,T mult=1.0f,T bias=0.0f)
	{
		for(; begin!=end; ++in )
		{
			(*in)=mult*((*in)-bias)*(*begin++);
		}
	}
	/*!
	 *       . \see WindowData().
	 *
	 **************************************************************************************/
	template<typename T,typename iterator_in,typename iterator_ker>
	T IntegrateWindowedData(iterator_in in,iterator_ker begin,iterator_ker end)
	{
		T sum=0;
		for(; begin!=end; ++in )
		{
			sum+=(*in)*(*begin++);
		}
		return sum;
	}
	/*!
	 *    . \see WindowData().
	 *
	 **************************************************************************************/
	template<typename T,typename iterator_ker>
	T IntegrateKer(iterator_ker begin,iterator_ker end)
	{
		T sum=0;
		for(; begin!=end;  )
		{
			sum+=(*begin++);
		}
		return sum;
	}
	/*!
	 * \brief   .
	 * 
	 **********************************************************************************/
	template<typename T>
	class ImageWin
	{
		Vtl::Plf::Kernel::FKern<T,Vtl::Plf::Kernel::KT_Win,false> ker_x;
		Vtl::Plf::Kernel::FKern<T,Vtl::Plf::Kernel::KT_Win,false> ker_y;
	public:
		ImageWin()
		{
		}
		template<class Fx,class Fy,typename TT>
		void Init(Fx functor_x,Fy functor_y,size_t width,size_t height,TT center_x,TT center_y,T fill_by)
		{
			ker_x.Init(functor_x,width,center_x,fill_by);
			ker_y.Init(functor_y,height,center_y,fill_by);
		}
		void Run(T *p,int nls)
		{
			T int_w=0;
			T int_w_x_data=0;
			T *p_cur=p;
			for(Vtl::Plf::Kernel::Base<T,Vtl::Plf::Kernel::KT_Win,false>::const_iterator iy=ker_y.begin(); iy!=ker_y.end() ; ++iy)
			{
				int_w        += (*iy)*Vtl::Plf::Kernel::IntegrateKer<T>(ker_x.begin(),ker_x.end());
				int_w_x_data += (*iy)*Vtl::Plf::Kernel::IntegrateWindowedData<T>(p_cur,ker_x.begin(),ker_x.end());
				p_cur+=nls;
			}
			p_cur=p;
			T ofs=int_w_x_data/int_w;
			for(Vtl::Plf::Kernel::Base<T,Vtl::Plf::Kernel::KT_Win,false>::const_iterator iy=ker_y.begin(); iy!=ker_y.end() ; ++iy)
			{
				Vtl::Plf::Kernel::WindowData(p_cur,ker_x.begin(),ker_x.end(),*iy,ofs);
				p_cur+=nls;
			}
		}
	};



} //namespace Vtl::Plf::Kernel
/** @} */ // end of ConvolutionKernelsDoxGroup  
} //namespace Plf 
} //namespace Vtl

/////////////////////////////////////////////////////////////////////////////
#endif //_KERNELS_1D_H 
